from pathlib import Path
import os

import numpy as np
import pandas as pd
from loguru import logger
from tensorflow.keras.callbacks import Callback
import matplotlib.pyplot as plt
from scipy import stats

from worldcereal.classification.models import WorldCerealAutoencoder
from worldcereal.utils.scalers import scale_df
from worldcereal.utils import io
from worldcereal.utils import is_real_feature


def scale(data, ftname,
          scalemin=0,
          scalemax=1,
          clamp=False):

    return scalers.minmaxscaler(data,
                                ftname,
                                minscaled=scalemin,
                                maxscaled=scalemax,
                                clamp=clamp)


def sample_df(df, max_samples=2000, column='aez_zoneid'):

    df = df[df[column] != 0]
    df = df[df[column] != 9999]
    df = df[df[column] != -9999]

    unique_values = df[column].astype(int).unique()

    sampled_dfs = []

    for zone in unique_values:
        subset = df[df[column] == zone]
        if subset.shape[0] < max_samples:
            sampled_dfs.append(subset)
        else:
            sampled_dfs.append(subset.sample(n=max_samples))

    return pd.concat(sampled_dfs)


def performance_plotter(figdir, testdata,
                        limitsamples=None,
                        limitfeatures=None):
    os.makedirs(figdir, exist_ok=True)

    class performanceplottercallback(Callback):
        def on_epoch_end(self, epoch, logs=None, testdata=testdata):

            f, ax = plt.subplots(1, 1, figsize=(8, 8))

            if limitsamples:
                testdata = testdata[0:limitsamples, ...]

            reconstructions = self.model.predict(testdata)

            if limitfeatures:
                idx = np.random.choice(np.arange(testdata.shape[1]),
                                       size=limitfeatures,
                                       replace=False)
                testdata = testdata[:, idx]
                reconstructions = reconstructions[:, idx]

            # Calculate the point density
            logger.info('\nCalculating point density ...')
            xy = np.vstack([testdata.ravel(), reconstructions.ravel()])
            z = stats.gaussian_kde(xy)(xy)

            logger.info('Plotting points ...')
            ax.scatter(testdata.ravel(), reconstructions.ravel(),
                       c=z, s=50, edgecolor='')
            ax.set_xlim(-0.1, 1.1)
            ax.set_ylim(-0.1, 1.1)
            ax.set_xlabel('Original feature value', fontsize=14)
            ax.set_ylabel('Reconstructed feature value', fontsize=14)

            plt.tight_layout()
            plt.grid()
            plt.savefig(Path(figdir) / f'epoch_{epoch}.png')
            plt.close()

    return performanceplottercallback()


def main(spark, data_dir, outputdir, latent_dim,
         ft_selection_file=None,
         depth=3):

    if spark:
        logger.info('Setting up spark ...')
        from worldcereal.utils.spark import get_spark_context
        _ = get_spark_context()

    # Path where training data is stored
    data_dir = Path(data_dir)

    # Specify calibration and validation CSV files
    calfile = data_dir / 'CAL' / 'training_df_LC.csv'
    valfile = data_dir / 'VAL' / 'training_df_LC.csv'
    testfile = data_dir / 'TEST' / 'training_df_LC.csv'

    # Load the data in memory
    logger.info(f'Reading training CSV: {calfile}')
    trainingdata = pd.read_csv(calfile, index_col=0)
    logger.info(f'Reading validation CSV: {valfile}')
    validationdata = pd.read_csv(valfile, index_col=0)
    logger.info(f'Reading test CSV: {testfile}')
    testdata = pd.read_csv(testfile, index_col=0)

    trainingdata = sample_df(trainingdata,
                             max_samples=2000,
                             column='aez_zoneid')
    validationdata = sample_df(validationdata,
                               max_samples=570,
                               column='aez_zoneid')
    testdata = sample_df(testdata,
                         max_samples=285,
                         column='aez_zoneid')

    logger.info('-' * 50)
    logger.info(f'CAL samples retained: {trainingdata.shape[0]}')
    logger.info(f'VAL samples retained: {validationdata.shape[0]}')
    logger.info(f'TEST samples retained: {testdata.shape[0]}')
    logger.info('-' * 50)

    # Get the number and names of input features
    if ft_selection_file is None:
        # Use all features in df
        ftnames = [x for x in trainingdata.columns.tolist()
                   if is_real_feature(x)]
        features = len(ftnames)
    else:
        ftnames = []
        with open(ft_selection_file) as f:
            lines = f.readlines()
            for ft in lines:
                ftnames.append(ft.strip())
        features = len(ftnames)

    # Model name
    modelname = (f'WorldCereal_AutoEncoder_from{features}fts'
                 f'_to{latent_dim}latentfts')

    # Setup the autoencoder model
    logger.info('Initializing autoencoder model ...')
    ae_parameters = {'latent_dim': latent_dim, 'depth': depth}
    autoencoder = WorldCerealAutoencoder(feature_names=ftnames,
                                         parameters=ae_parameters,
                                         basedir=Path(outputdir))

    # Scale the input data
    logger.info('Scaling input data ...')
    trainingdatascaled = scale_df(trainingdata[ftnames], clamp=(-0.5, 1.5))
    validationdatascaled = scale_df(validationdata[ftnames], clamp=(-0.5, 1.5))
    testdatascaled = scale_df(testdata[ftnames], clamp=(-0.5, 1.5))

    # Remove NaN rows
    # inputscal = io.drop_nan(trainingdatascaled.values)
    # inputsval = io.drop_nan(validationdatascaled.values)
    # inputstest = io.drop_nan(testdatascaled.values)
    inputscal = trainingdatascaled.values
    inputsval = validationdatascaled.values
    inputstest = testdatascaled.values

    inputscal[np.isnan(inputscal)] = 0
    inputsval[np.isnan(inputsval)] = 0
    inputstest[np.isnan(inputstest)] = 0

    minvaluecal = np.percentile(inputscal.ravel(), 5)
    maxvaluecal = np.percentile(inputscal.ravel(), 95)
    minvalueval = np.percentile(inputsval.ravel(), 5)
    maxvalueval = np.percentile(inputsval.ravel(), 95)
    logger.info('Scaling done.')
    logger.info(f'Scaled P05 value CAL: {minvaluecal}')
    logger.info(f'Scaled P95 value CAL: {maxvaluecal}')
    logger.info(f'Scaled P05 value VAL: {minvalueval}')
    logger.info(f'Scaled P95 value VAL: {maxvalueval}')

    # Initialize performance plotter
    # which will be an additional callback during training
    performanceplottercallback = performance_plotter(
        Path(outputdir) / 'figs',
        inputstest,
        limitsamples=500,
        limitfeatures=None
    )

    # Train the autoencoder
    modelfile = Path(outputdir) / (modelname + '.h5')
    autoencoder.train(
        calibrationx=inputscal,
        calibrationy=inputscal,
        validationdata=(inputsval, inputsval),
        batch_size=128,
        epochs=100,
        earlystopping=True,
        tensorboard=False,
        csvlogger=False,
        customcallbacks=[]
    )

    logger.success('Training done!')

    # Save the model
    autoencoder.save(modelfile)

    logger.info('Now encoding the features ...')

    latentftnames = [f'latent_ft_{x+1}' for x in range(latent_dim)]
    nonftcolumns = [ft for ft in trainingdata.columns.tolist()
                    if not is_real_feature(ft)]

    for dataset in ['CAL', 'VAL', 'TEST']:
        encoded_path = Path(outputdir) / 'encodedfeatures' / dataset
        encoded_path.mkdir(parents=True, exist_ok=True)

        if dataset == 'CAL':

            data = pd.read_csv(calfile, index_col=0)
            datascaled = scale_df(data[ftnames], clamp=(-0.5, 1.5))
            inputs = datascaled.values
            metafts = [ft for ft in data.columns.tolist()
                       if not is_real_feature(ft)]

            inputs = trainingdatascaled.values
            origdf = trainingdata
        elif dataset == 'VAL':
            inputs = validationdatascaled.values
            origdf = validationdata
        elif dataset == 'TEST':
            inputs = testdatascaled.values
            origdf = testdata

        logger.info(f'Encoding features from {dataset} dataset ...')
        encoded = autoencoder.encode(inputs)
        logger.info('Encoding done!')

        logger.info('Putting in dataframe ...')
        encodeddf = pd.DataFrame(encoded, columns=latentftnames)

        logger.info('Adding original columns ...')
        encodeddf[nonftcolumns] = origdf[nonftcolumns]

        outfile = encoded_path / 'encoded_df.csv'
        logger.info(f'Writing to: {outfile}')
        encodeddf.to_csv(outfile)

    logger.success('All done!')


if __name__ == '__main__':

    # What is the base dir where the features are
    data_dir = ('/data/worldcereal/benchmark/'
                'CIB_V1/features/croplandextent_CIB')

    # How many condensed features to we want
    latent_features = 32

    # Output directory
    outputdir = ('/data/worldcereal/benchmark/CIB_V1/'
                 f'AutoEncoders/AutoEncoder130ftsTO{latent_features}hiddenfts')

    # Auto-encoder layer depth
    depth = 3

    # File with list of input features to use
    # Set to `None` to use all features in the dataframe
    ft_selection_file = ('/data/worldcereal/benchmark/CIB_V1'
                         '/features/croplandextent_CIB/CAL'
                         '/Selection/lists/all.txt')
    # ft_selection_file = None

    spark = True

    main(spark, data_dir, outputdir,
         latent_dim=latent_features,
         ft_selection_file=ft_selection_file,
         depth=depth)
