from loguru import logger
from pathlib import Path
import glob
from datetime import datetime
import pandas as pd
import requests
import json
import numpy as np
from typing import Dict, List
from dateutil.parser import parse
from satio.grid import tile_to_epsg
from satio.collections import (TerrascopeSigma0Collection,
                               TrainingCollection,
                               DiskCollection,
                               S2Collection)
from satio.geoloader import LandsatLoader, ParallelLoader
from worldcereal.fp import TSSIGMA0TiledFeaturesProcessor as TSS0FeatProc
from worldcereal.fp import L8ThermalFeaturesProcessor as L8FeatProc
from worldcereal.fp import WorldCerealSIGMA0FeaturesProcessor as WCS0FP
from worldcereal.fp import WorldCerealOpticalFeaturesProcessor as WCOPTICALFP
from worldcereal.fp import WorldCerealThermalFeaturesProcessor as WCTHERMALFP

_S2_BASEDIR = '/data/MTDA/TERRASCOPE_Sentinel2/NDVI_V2'
_TS_S2_BASEURL = ('https://services.terrascope.be/catalogue/'
                  'products?collection=urn%3Aeop%3AVITO%3A'
                  'TERRASCOPE_S2')
_S1_BASEDIR = '/data/MTDA/CGS_S1_GRD_SIGMA0_L1'
_TS_S1_BASEURL = ('https://services.terrascope.be/catalogue/'
                  'products?collection=urn%3Aeop%3AVITO%3A'
                  'CGS_S1_GRD_SIGMA0_L1')


class TerrascopeSigma0TiledCollection(TerrascopeSigma0Collection):

    @property
    def supported_resolutions(self):
        return [20]

    @property
    def loader(self):
        if self._loader is None:
            self._loader = ParallelLoader(fill_value=65535)
        return self._loader

    def features_processor(self,
                           settings: Dict,
                           rsi_meta: Dict = {},
                           features_meta: Dict = {},
                           ignore_def_features: bool = False) -> TSS0FeatProc:
        return TSS0FeatProc(self,
                            settings,
                            rsi_meta=rsi_meta,
                            features_meta=features_meta,
                            ignore_def_features=ignore_def_features)

    @staticmethod
    def get_terrascope_filenames(terrascope_product):
        filenames_list = glob.glob(str(Path(terrascope_product) / '*.tif'),
                                   recursive=True)
        filenames = {}
        resolution = 20

        for f in filenames_list:

            band = str(Path(f).stem).split('_')[-2].split('.')[0]

            filenames[band] = {resolution: f}

        return filenames

    @staticmethod
    def get_terrascope_band_filename(filenames, band, resolution=20):
        """Returns filename for band (TERRASCOPE product)
        closest to target resolution"""

        bands_20m = ['VV', 'VH']

        if band not in bands_20m:
            raise ValueError("Band {} not supported.".format(band))

        return filenames[band][resolution]


class WorldCerealSigma0TiledCollection(DiskCollection):

    sensor = 'S1'
    processing_level = 'SIGMA0'

    @property
    def supported_bands(self):
        return ['VV', 'VH']

    @property
    def supported_resolutions(self):
        return [20]

    def get_band_filenames(self, band, resolution):
        if self._filenames is None:
            self._filenames = self.df['path'].apply(
                lambda x: self.get_worldcereal_filenames(x))

        tif_filenames = self._filenames.apply(
            lambda x: self.get_worldcereal_band_filename(x, band, resolution))

        return tif_filenames.values.tolist()

    @classmethod
    def from_folder(cls, folder, s2grid=None):
        df = cls.build_products_df(folder)
        df = df.sort_values('date', ascending=True)
        collection = cls(df, s2grid=s2grid)
        return collection

    @property
    def loader(self):
        if self._loader is None:
            self._loader = ParallelLoader(fill_value=0)
        return self._loader

    def features_processor(self,
                           settings: Dict,
                           rsi_meta: Dict = {},
                           features_meta: Dict = {},
                           ignore_def_features: bool = False) -> WCS0FP:
        return WCS0FP(self,
                      settings,
                      rsi_meta=rsi_meta,
                      features_meta=features_meta,
                      ignore_def_features=ignore_def_features)

    @staticmethod
    def get_worldcereal_filenames(worldcereal_product):
        filenames_list = glob.glob(str(Path(worldcereal_product) / '*.tif'),
                                   recursive=True)
        filenames = {}
        resolution = 20

        for f in filenames_list:

            band = str(Path(f).stem).split('_')[-1].split('.')[-1]

            filenames[band] = {resolution: f}

        return filenames

    @staticmethod
    def get_worldcereal_band_filename(filenames, band, resolution=20):
        """Returns filename for band (WORLDCEREAL product)
        """

        bands_20m = ['VV', 'VH']

        if band not in bands_20m:
            raise ValueError("Band {} not supported.".format(band))

        return filenames[band][resolution]

    @classmethod
    def build_products_df(cls, folder):

        products = []
        zones = [Path(x).stem for x in glob.glob(str(Path(folder) / '*'))]
        for zone in zones:
            tilespart1 = [Path(x).stem for x in glob.glob(
                str(Path(folder) / zone / '*'))]
            for tilepart1 in tilespart1:
                tilespart2 = [Path(x).stem for x in glob.glob(
                    str(Path(folder) / zone / tilepart1 / '*'))]
                for tilepart2 in tilespart2:
                    tileproducts = [x for x in glob.glob(
                        str(Path(folder) / zone /
                            tilepart1 / tilepart2 / '*' / '*' / '*'))]

                    for tileproduct in tileproducts:
                        products.append(tileproduct)

        entries = [cls.sentinel1_entry(f) for f in products]
        if len(entries):
            df = pd.DataFrame(entries)
        else:
            df = pd.DataFrame([], columns=['date',
                                           'tile',
                                           'path',
                                           'level'])

        return df

    @staticmethod
    def sentinel1_entry(filename):
        """
        """
        basename = Path(filename).stem
        date = basename.split('_')[1]
        tile = Path(filename).stem.split('_')[-1]

        entry = dict(date=datetime.strptime(date, '%Y%m%dT%H%M%S'),
                     tile=tile,
                     level='SIGMA0',
                     path=filename)

        return entry


class WorldCerealOpticalTiledCollection(S2Collection):

    processing_level = 'SMAC'

    @ property
    def supported_bands(self):
        return ["B02", "B03", "B04", "B05", "B06",
                "B07", "B08", "B8A", "B11", "B12", "MASK"]

    def get_band_filenames(self, band, resolution):
        if self._filenames is None:
            self._filenames = self.df['path'].apply(
                lambda x: self.get_worldcereal_filenames(x))

        tif_filenames = self._filenames.apply(
            lambda x: self.get_worldcereal_band_filename(x, band, resolution))

        return tif_filenames.values.tolist()

    @classmethod
    def from_folder(cls, folder, s2grid=None):
        df = cls.build_products_df(folder)
        df = df.sort_values('date', ascending=True)
        collection = cls(df, s2grid=s2grid)
        return collection

    def features_processor(self,
                           settings: Dict,
                           rsi_meta: Dict = {},
                           features_meta: Dict = {},
                           ignore_def_features: bool = False) -> WCOPTICALFP:
        return WCOPTICALFP(self,
                           settings,
                           rsi_meta=rsi_meta,
                           features_meta=features_meta,
                           ignore_def_features=ignore_def_features)

    def load_timeseries(self,
                        *bands,
                        **kwargs):
        '''Need to override parent method to cope with missing
        bands in L8 data that can be part of the timeseries.

        and to correct currently faulty values
        '''
        from satio.timeseries import load_timeseries

        # FOR NOW OVERRIDE TO ONLY TAKE S2!
        logger.warning('L8 OPTICAL IS DEACTIVATED !!!')
        self.df = self.df[self.df.path.str.contains('MSIL2A')]

        # Which bands does L8 have too
        L8_available_bands = ["B02", "B03", "B04", "B08",
                              "B11", "B12", "MASK"]

        # Get the requested bands we can load from
        # both S2 and L8
        combined_to_load = [b for b in list(bands)
                            if b in L8_available_bands]

        if combined_to_load:
            # Load combined S2/L8 bands
            combined_ts = load_timeseries(self, *combined_to_load, **kwargs)
        else:
            combined_ts = None

        # Get the requested bands we can load
        # exclusively from S2
        S2_to_load = [b for b in list(bands)
                            if b not in L8_available_bands]

        if S2_to_load:
            # Load exclusive S2 bands
            S2_coll = self._clone()
            S2_coll.df = self.df[self.df.path.str.contains('MSIL2A')]
            S2_ts = load_timeseries(S2_coll, *S2_to_load, **kwargs)
        else:
            S2_ts = None

        if S2_ts is not None and combined_ts is not None:
            timeseries = combined_ts.merge(S2_ts, fill_missing_timestamps=True)
        elif combined_ts is not None:
            timeseries = combined_ts
        else:
            timeseries = S2_ts

        # Put faulty values to nodata
        timeseries.data[timeseries.data > 65000] = 0

        return timeseries

    @staticmethod
    def get_worldcereal_filenames(worldcereal_product):
        filenames_list = glob.glob(str(Path(worldcereal_product) / '*.tif'),
                                   recursive=True)
        filenames = {}
        bands_10m = ['B02', 'B03', 'B04', 'B08']
        bands_20m = ['B05', 'B06', 'B07', 'B8A', 'B11', 'B12',
                     'MASK']

        for f in filenames_list:

            band = str(Path(f).stem).split('_')[-1].split('.')[0]

            if band in bands_10m:
                resolution = 10
            elif band in bands_20m:
                resolution = 20

            filenames[band] = {resolution: f}

        return filenames

    @staticmethod
    def get_worldcereal_band_filename(filenames, band, resolution):
        bands_10m = ['B02', 'B03', 'B04', 'B08']
        bands_20m = ['B05', 'B06', 'B07', 'B8A', 'B11', 'B12',
                     'MASK']

        if band in bands_10m:
            resolution = 10
        elif band in bands_20m:
            resolution = 20
        else:
            raise ValueError("Band {} not supported.".format(band))

        return filenames[band][resolution]

    def filter_nodata(self,
                      resolution=20,
                      mask_th=0.05):
        """
        Load mask and filter collection based on the relative amount
        of valid pixels. Minimum is specified by `mask_th`.
        e.g. a frame is valid if the mean of the binary mask
        is above `mask_th`. Or in other words,
        keep frames with more than `mask_th * 100` % valid pixels.
        """
        scl_20 = self.load(['MASK'], resolution=resolution)['MASK']

        valid_ids = np.where(scl_20.mean(axis=(1, 2)) > mask_th)[0]

        return self._clone(df=self.df.iloc[valid_ids])

    @ classmethod
    def build_products_df(cls, folder):

        products = []
        zones = [Path(x).stem for x in glob.glob(str(Path(folder) / '*'))]
        for zone in zones:
            tilespart1 = [Path(x).stem for x in glob.glob(
                str(Path(folder) / zone / '*'))]
            for tilepart1 in tilespart1:
                tilespart2 = [Path(x).stem for x in glob.glob(
                    str(Path(folder) / zone / tilepart1 / '*'))]
                for tilepart2 in tilespart2:
                    tileproducts = [x for x in glob.glob(
                        str(Path(folder) / zone /
                            tilepart1 / tilepart2 / '*' / '*' / '*'))]

                    for tileproduct in tileproducts:
                        filenames = glob.glob(
                            str(Path(tileproduct) / '*.tif'))
                        bands = [Path(x).stem.split('_')[
                            -1].split('.')[0] for x in filenames]
                        if 'MASK' in bands:
                            products.append(tileproduct)
                        else:
                            logger.warning(
                                ('No MASK file found for: '
                                 f'`{tileproduct}` -> Ignoring product'))

        entries = [cls.sentinel2_entry(f) for f in products]
        if len(entries):
            df = pd.DataFrame(entries)
        else:
            df = pd.DataFrame([], columns=['date',
                                           'tile',
                                           'path',
                                           'level'])

        return df

    @ staticmethod
    def sentinel2_entry(filename):
        """
        """
        basename = Path(filename).stem
        date = basename.split('_')[2]
        tile = Path(filename).stem.split('_')[-1]

        entry = dict(date=pd.to_datetime(date),
                     tile=tile,
                     level='SMAC',
                     path=filename)

        return entry


class WorldCerealThermalTiledCollection(DiskCollection):

    sensor = 'L8'
    processing_level = 'L2SP'

    def __init__(self, df, s2grid=None):
        super().__init__(df, s2grid=s2grid)
        self.df = self.df[self.df.level == self.processing_level]

    @ property
    def supported_bands(self):
        return ['B10', 'QA_PIXEL']

    @ property
    def supported_resolutions(self):
        return [10, 30]

    @property
    def loader(self):
        if self._loader is None:
            self._loader = LandsatLoader()
        return self._loader

    @loader.setter
    def loader(self, value):
        self._loader = value

    @classmethod
    def from_folder(cls, folder, s2grid=None):
        df = cls.build_products_df(folder)
        df = df.sort_values('date', ascending=True)
        collection = cls(df, s2grid=s2grid)
        return collection

    def filter_nodata(self,
                      resolution=30,
                      mask_th=0.05):
        """
        filter collection based on the relative amount
        of valid pixels. Minimum is specified by `mask_th`.
        e.g. a frame is valid if the mean of the binary mask
        is above `mask_th`. Or in other words,
        keep frames with more than `mask_th * 100` % valid pixels.
        """
        qa_30 = self.load(['QA_PIXEL'], resolution=resolution)['QA_PIXEL']

        # Define the to-be-masked pixel_qa values
        shadow = 1 << 3
        cloud = 1 << 5
        cirrus = 1 << 9

        clear = ((qa_30.data & shadow == 0) &
                 (qa_30.data & cloud == 0) &
                 (qa_30.data & cirrus == 0))

        valid_ids = np.where(clear.mean(axis=(1, 2)) > mask_th)[0]

        return self._clone(df=self.df.iloc[valid_ids])

    def features_processor(self,
                           settings: Dict,
                           rsi_meta: Dict = {},
                           features_meta: Dict = {},
                           ignore_def_features: bool = False) -> WCTHERMALFP:
        return WCTHERMALFP(self,
                           settings,
                           rsi_meta=rsi_meta,
                           features_meta=features_meta,
                           ignore_def_features=ignore_def_features)

    def get_band_filenames(self, band, resolution):
        if self._filenames is None:
            self._filenames = self.df['path'].apply(
                lambda x: self.get_worldcereal_filenames(x))

        tif_filenames = self._filenames.apply(
            lambda x: self.get_worldcereal_band_filename(x, band, resolution))

        return tif_filenames.values.tolist()

    @staticmethod
    def get_worldcereal_filenames(worldcerealTherm_product):
        filenames_list = glob.glob(
            str(Path(worldcerealTherm_product) / '*.tif'), recursive=True)
        filenames = {}
        resolution = 30

        for f in filenames_list:

            band = '_'.join(str(Path(f).stem).split('_')[5:]).split('.')[0]

            filenames[band] = {resolution: f}

        return filenames

    @staticmethod
    def get_worldcereal_band_filename(filenames, band, resolution=30):
        """Returns filename for band
        closest to target resolution"""

        bands_30m = ['B10', 'QA_PIXEL']
        if band not in bands_30m:
            raise ValueError("Band {} not supported.".format(band))

        return filenames[band][resolution]

    @ classmethod
    def build_products_df(cls, folder):

        products = []
        zones = [Path(x).stem for x in glob.glob(str(Path(folder) / '*'))]
        for zone in zones:
            tilespart1 = [Path(x).stem for x in glob.glob(
                str(Path(folder) / zone / '*'))]
            for tilepart1 in tilespart1:
                tilespart2 = [Path(x).stem for x in glob.glob(
                    str(Path(folder) / zone / tilepart1 / '*'))]
                for tilepart2 in tilespart2:
                    tileproducts = [x for x in glob.glob(
                        str(Path(folder) / zone /
                            tilepart1 / tilepart2 / '*' / '*' / '*'))]

                    for tileproduct in tileproducts:
                        filenames = glob.glob(
                            str(Path(tileproduct) / '*.tif'))
                        bands = ['_'.join(str(Path(x).stem).split('_')[
                            5:]).split('.')[0]
                            for x in filenames]
                        if 'QA_PIXEL' in bands:
                            products.append(tileproduct)
                        else:
                            logger.warning(
                                ('No QA_PIXEL file found for: '
                                 f'`{tileproduct}` -> Ignoring product'))

        entries = [cls.landsat8_entry(f) for f in products]
        if len(entries):
            df = pd.DataFrame(entries)
        else:
            df = pd.DataFrame([], columns=['date',
                                           'tile',
                                           'path',
                                           'level'])

        return df

    @ staticmethod
    def landsat8_entry(filename):
        """
        """
        basename = Path(filename).stem
        date = basename.split('_')[2]
        tile = Path(filename).stem.split('_')[-1]

        entry = dict(date=pd.to_datetime(date),
                     tile=tile,
                     level='L2SP',
                     path=filename)

        return entry

    def load_timeseries(self, *bands, **kwargs):
        '''
        Override default method to deal
        with scaling
        '''

        L8SCALING = {
            'B10': {
                'scale': 0.00341802,
                'offset': 149.0,
                'nodata': 0
            }
        }

        from satio.timeseries import load_timeseries
        from satio.timeseries import Timeseries

        ts = load_timeseries(self, *bands, **kwargs)

        scaled_ts = None

        for band in bands:
            scaling = L8SCALING.get(band, None)
            if scaling is not None:
                ts.data = ts.data.astype(np.float32)
                ts_banddata = ts[band].data
                ts_banddata[ts_banddata == scaling['nodata']] = np.nan
                ts_banddata = ((ts_banddata * scaling['scale'])
                               + scaling['offset'])
            else:
                ts_banddata = ts[band].data

            ts_band = Timeseries(
                data=ts_banddata,
                timestamps=ts.timestamps,
                bands=[band],
                attrs=ts.attrs
            )
            scaled_ts = (ts_band if scaled_ts is None
                         else scaled_ts.merge(ts_band))

        return scaled_ts


class L8ThermalTiledCollection(DiskCollection):

    sensor = 'L8'
    processing_level = 'L2SP'

    def __init__(self, df, s2grid=None):
        super().__init__(df, s2grid=s2grid)
        self.df = self.df[self.df.level == self.processing_level]

    @ property
    def supported_bands(self):
        return ['ST-B10', 'PIXEL-QA']

    @ property
    def supported_resolutions(self):
        return [10]

    @ classmethod
    def from_folders(cls, *folders, s2grid=None):
        df = build_L8ThermalTiled_products_df(*folders)
        df = df.sort_values('date', ascending=True)
        collection = cls(df, s2grid=s2grid)
        return collection

    def filter_nodata(self,
                      resolution=10,
                      mask_th=0.05):
        """
        filter collection based on the relative amount
        of valid pixels. Minimum is specified by `mask_th`.
        e.g. a frame is valid if the mean of the binary mask
        is above `mask_th`. Or in other words,
        keep frames with more than `mask_th * 100` % valid pixels.
        """
        pixel_qa_10 = self.load(['PIXEL-QA'],
                                resolution=resolution)['PIXEL-QA']

        # Define the to-be-masked pixel_qa values
        cirrus = 1 << 2
        cloud = 1 << 3
        shadow = 1 << 4
        snow = 1 << 5

        clear = ((pixel_qa_10 & shadow == 0) &
                 (pixel_qa_10 & cloud == 0) &
                 (pixel_qa_10 & cirrus == 0) &
                 (pixel_qa_10 & snow == 0))

        valid_ids = np.where(clear.mean(axis=(1, 2)) > mask_th)[0]

        return self._clone(df=self.df.iloc[valid_ids])

    def features_processor(self,
                           settings: Dict,
                           rsi_meta: Dict = {},
                           features_meta: Dict = {},
                           ignore_def_features: bool = False) -> L8FeatProc:
        return L8FeatProc(self,
                          settings,
                          rsi_meta=rsi_meta,
                          features_meta=features_meta,
                          ignore_def_features=ignore_def_features)

    def get_band_filenames(self, band, resolution):
        if self._filenames is None:
            self._filenames = self.df['path'].apply(
                lambda x: self.get_filenames(x))

        tif_filenames = self._filenames.apply(
            lambda x: self.get_band_filename(x, band, resolution))

        return tif_filenames.values.tolist()

    @staticmethod
    def get_filenames(L8_tiled_product):
        filenames_list = glob.glob(str(Path(L8_tiled_product) / '*.tif'),
                                   recursive=True)
        filenames = {}
        resolution = 10

        for f in filenames_list:

            band = str(Path(f).stem).split('_')[-2]

            filenames[band] = {resolution: f}

        return filenames

    @staticmethod
    def get_band_filename(filenames, band, resolution=10):
        """Returns filename for band
        closest to target resolution"""

        bands_10m = ['ST-B10', 'PIXEL-QA']

        if band not in bands_10m:
            raise ValueError("Band {} not supported.".format(band))

        return filenames[band][resolution]


class L8ThermalTrainingCollection(TrainingCollection):

    sensor = 'L8'
    processing_level = 'L2-ST'

    def features_processor(self,
                           settings: Dict,
                           rsi_meta: Dict = {},
                           features_meta: Dict = {},
                           ignore_def_features: bool = False) -> L8FeatProc:
        return L8FeatProc(self,
                          settings,
                          rsi_meta=rsi_meta,
                          features_meta=features_meta,
                          ignore_def_features=ignore_def_features)

    def load(self,
             bands: List = None,
             resolution: int = None,
             location_id: int = None,
             mask_and_scale: bool = False) -> Dict:
        '''
        Override of parent class method to support native
        GEE-processed data
        '''

        if location_id is None:
            # get first location available
            location_id = self.location_ids[0]

        row = self.df[self.df.location_id == location_id].iloc[0]
        path = row['path']

        if not Path(path).is_dir():
            raise FileNotFoundError(f"{path} not found.")

        # Here we split up the worldcover and worldcereal approaches
        if self._dataformat == 'worldcereal':
            xarrs = self._load_worldcereal(row, bands,
                                           mask_and_scale=mask_and_scale)
        else:
            xarrs = self._load_satio(row, bands)

        # replace bytestrings by strings in timestamp if needed
        if isinstance(xarrs[bands[0]].timestamp.values[0], bytes):
            xarrs[bands[0]] = xarrs[bands[0]].assign_coords(
                timestamp=[s.decode("utf-8") for s in
                           xarrs[bands[0]].timestamp.values])

        return xarrs


class WorldCerealOpticalCREOCollection(S2Collection):

    processing_level = 'SMAC'

    @ property
    def supported_bands(self):
        return ["B02", "B03", "B04", "B05", "B06",
                "B07", "B08", "B8A", "B11", "B12", "MASK"]

    @classmethod
    def from_url(cls, url, s2grid=None):
        df = build_opticalcreo_products_df(url)
        return cls(df, s2grid=s2grid)

    @staticmethod
    def _get_band_tif(band, product):
        split = product.split('_')
        if split[1] == 'MSIL1C':
            split[1] = 'SMAC' if band != 'MASK' else 'SCL'
        elif split[1] == 'L1T':
            split[1] = 'SMAC' if band != 'MASK' else 'FMASK'
        else:
            raise ValueError(f'Unrecognized product: {product}')
        return "_".join(split) + f'_{band}.tif'

    def get_band_filenames(self, band, resolution=10):
        filenames = self.df.apply(lambda x:
                                  f"{x.path}/"
                                  f"{self._get_band_tif(band, x['product'])}",
                                  axis=1).values
        return filenames

    def features_processor(self,
                           settings: Dict,
                           rsi_meta: Dict = {},
                           features_meta: Dict = {},
                           ignore_def_features: bool = False) -> WCOPTICALFP:
        return WCOPTICALFP(self,
                           settings,
                           rsi_meta=rsi_meta,
                           features_meta=features_meta,
                           ignore_def_features=ignore_def_features)

    def filter_nodata(self,
                      resolution=20,
                      mask_th=0.05):
        """
        Load mask and filter collection based on the relative amount
        of valid pixels. Minimum is specified by `mask_th`.
        e.g. a frame is valid if the mean of the binary mask
        is above `mask_th`. Or in other words,
        keep frames with more than `mask_th * 100` % valid pixels.
        """
        scl_20 = self.load(['MASK'], resolution=resolution)['MASK']

        valid_ids = np.where(scl_20.mean(axis=(1, 2)) > mask_th)[0]

        return self._clone(df=self.df.iloc[valid_ids])


def get_S2_products(tile, startdate, enddate):
    logger.info(f'Getting S2 products for tile {tile} ...')

    # Make a query to the catalog to retrieve an example file
    response = requests.get((f'{_TS_S2_BASEURL}_TOC_V2&tileId={tile}'
                             '&cloudCover=[0,95]'
                             f'&start={startdate}&end={enddate}'
                             '&accessedFrom=MEP'))
    response = json.loads(response.text)

    if response['totalResults'] == 0:
        logger.warning('No Sentinel-2 products found.')
        return []

    products = []
    for acquisition in response['features']:
        for band in acquisition['properties']['links']['data']:
            if 'B08' in band['title']:
                products.append(str(Path(band['href'][7:]).parent))

    while len(products) != response['totalResults']:
        response = requests.get(
            response['properties']['links']['next'][0]['href'])
        response = json.loads(response.text)
        for acquisition in response['features']:
            for band in acquisition['properties']['links']['data']:
                if 'B08' in band['title']:
                    products.append(str(Path(band['href'][7:]).parent))
    return products


def get_S1_products(bbox, orbitpass, startdate, enddate):
    logger.info(f'Getting S1 products for bbox: {bbox}')
    logger.info(f'Considering orbit pass: {orbitpass}')

    if orbitpass not in ['ASCENDING', 'DESCENDING']:
        raise ValueError('Orbit pass should be one of ASCENDING/DESCENDING')

    # Make a query to the catalog to retrieve an example file
    response = requests.get((f'{_TS_S1_BASEURL}&bbox={bbox}'
                             f'&start={startdate}&end={enddate}'
                             f'&orbitDirection={orbitpass}'
                             '&accessedFrom=MEP'))
    response = json.loads(response.text)

    if response['totalResults'] == 0:
        logger.warning(f'No {orbitpass} Sentinel-1 products found.')
        return []

    products = []
    for acquisition in response['features']:
        for band in acquisition['properties']['links']['data']:
            if 'VV' in band['title']:
                products.append(str(Path(band['href'][7:]).parent))

    while len(products) != response['totalResults']:
        response = requests.get(
            response['properties']['links']['next'][0]['href'])
        response = json.loads(response.text)
        for acquisition in response['features']:
            for band in acquisition['properties']['links']['data']:
                if 'VV' in band['title']:
                    products.append(str(Path(band['href'][7:]).parent))

    return products


def get_s2_product_date(product):
    return pd.to_datetime(str(Path(product).stem).split('_')[1])


def get_s1_product_date(product):
    return pd.to_datetime(str(Path(product).stem).split('_')[5])


def create_products_df(products, tile, epsg, sensor='s2'):
    df = pd.DataFrame(columns=['product', 'date', 'path'],
                      index=None)

    if sensor == 's2':
        df['date'] = [get_s2_product_date(x) for x in products]
    elif sensor == 's1':
        df['date'] = [get_s1_product_date(x) for x in products]
    else:
        raise ValueError(f'Unrecognized sensor: {sensor}')

    df['product'] = [str(Path(product).stem) for product in products]
    df['path'] = products
    df['tile'] = tile
    df['epsg'] = epsg
    df = df.sort_values('date', axis=0)

    return df


def landsat8_entry(filename):
    """
    """
    basename = Path(filename).stem
    date = basename.split('_')[2]
    tile = Path(filename).parts[-2]

    entry = dict(date=datetime.strptime(date, '%Y%m%d'),
                 tile=tile,
                 level='L2SP',
                 path=filename)

    return entry


def _build_L8ThermalTiled_products_df(folder, tiles=None):

    products = []
    if tiles is None:
        tiles = [Path(x).stem for x in glob.glob(str(Path(folder) / '*'))]
    for tile in tiles:
        tileproducts = [Path(x).stem for x in glob.glob(
            str(Path(folder) / tile / '*'))]

        for tileproduct in tileproducts:
            filenames = glob.glob(str(Path(folder) / tile /
                                      tileproduct / '*.tif'))
            bands = [Path(x).stem.split('_')[-2] for x in filenames]
            if 'PIXEL-QA' in bands and 'ST-B10' in bands:
                products.append(str(Path(folder) / tile / tileproduct))

    entries = [landsat8_entry(f) for f in products]
    if len(entries):
        df = pd.DataFrame(entries)
    else:
        df = pd.DataFrame([], columns=['date',
                                       'tile',
                                       'path',
                                       'level'])

    return df


def build_L8ThermalTiled_products_df(*folders):
    dfs_list = [_build_L8ThermalTiled_products_df(folder) for
                folder in folders]
    return pd.concat(dfs_list, axis=0)


def _get_creo_optical_entry(fn):
    product = fn.split('/')[-1]
    split = product.split('_')
    tile = split[-1]
    entry = dict(
        product=product,
        date=parse(split[2]),
        path=fn,
        tile=tile,
        epsg=tile_to_epsg(tile),
        level='L2A'
    )

    return entry


def _get_bucket_products(url):
    prefix = 'WORLDCEREAL_PREPROC/OPTICAL/'
    resp = requests.get(url)
    body = resp.content.decode()
    body = body.split('\n')
    files_keys = filter(lambda x: prefix in x, body)
    products = list(set(map(lambda x: url + "/".join(x.split("/")[:-1]),
                            files_keys)))
    return products


def build_opticalcreo_products_df(url):
    products = _get_bucket_products(url)
    entries = list(map(_get_creo_optical_entry, products))
    return pd.DataFrame(entries)


if __name__ == '__main__':
    # url = ('https://cf2.cloudferro.com:8080/swift/v1'
    #        '/AUTH_b33f63f311844f2fbf62c5741ff0f734/world-cereal/')
    # coll = WorldCerealOpticalCREOCollection.from_url(url)

    coll = WorldCerealSigma0TiledCollection.from_folder(
        '/data/worldcereal/runs/largescaletest/WORLDCEREAL_PREPROC/SAR'
    )

    from satio.grid import get_blocks_gdf
    blocks = get_blocks_gdf(['30SVH'], coll._s2grid)

    b = blocks[blocks.block_id == 99]
    coll = coll.filter_bounds(b['bounds'].values[0], int(b.epsg)).filter_dates(
        '2018-11-10', '2019-11-10')

    # b = blocks.iloc[0]
    # coll = coll.filter_bounds(b.bounds, b.epsg).filter_dates(
    #     '2018-09-04', '2018-09-15')
    # f = coll.get_band_filenames('MASK')[0]
    # src = rasterio.open(f)
    # pass
    coll = WorldCerealSigma0TiledCollection.from_folder(
        '/data/worldcereal/runs/largescaletest/WORLDCEREAL_PREPROC/SAR'
    ).filter_bounds(
        bounds=(300000, 4189760, 310240, 4200000),
        epsg=32630)
    print(coll)
