import numpy as np
from datetime import timedelta
from scipy.signal import find_peaks
import sys
from satio.features import Features


################################################################################################################################
# The following suite of functions are related to the detection, visualization and selection of seasons

# The idea is to run these functions PRIOR to using the feature processors to compute features
################################################################################################################################

# helper function to find date nearest to certain date
def nearest_date(items, pivot):
    return min(items, key=lambda x: abs(x - pivot))


def detect_seasons(timeseries, max_seasons=5, amp_thr1=0.1, amp_thr2_factor=0.35, min_window=10, max_window=185):
    '''
    Computes peak, SOS and EOS of all seasons within a given timeseries

    :param timeseries: the Timeseries object from which to derive the seasons
    (should only contain one band!)
    :param max_seasons: maximum number of seasons to be detected
    :param amp_thr1: minimum threshold for amplitude of a season
    :param amp_thr2_factor: factor with which to multiply total amplitude of signal to derive second minimum amplitude threshold
    :param min_window: search window for finding start and end before/after peak - minimum distance from peak in days
    :param max_window: search window for finding start and end before/after peak - maximum distance from peak in days

    returns a dictionary with for each pixel SOS, MOS, EOS and nSeasons
    '''
    # TODO: also include seasons which have a valid start but which have not yet reached an end at the end of the timeseries
    # for this you need to adapt the code
    # whether or not you detect such a season could also be a feature in the classifier...

    data = np.squeeze(timeseries.data)
    time = timeseries.timestamps

    # prepare outputs
    nx = data.shape[1]
    ny = data.shape[2]
    seasons = {'nseasons': np.zeros((nx, ny), dtype='int32'),
               'sos': np.empty((max_seasons, nx, ny), dtype='int32'),
               'mos': np.empty((max_seasons, nx, ny), dtype='int32'),
               'eos': np.empty((max_seasons, nx, ny), dtype='int32')}

    # loop over each individual pixel to define start, peak and end of season(s)
    for i in range(nx):
        for j in range(ny):
            data_pix = np.squeeze(data[:, i, j])
            # find all local maxima
            # localmax_idx = argrelextrema(data_pix, np.greater)[0]
            localmax_idx = find_peaks(data_pix)[0]
            localmax = data_pix[localmax_idx]

            # sort local maxima according to VI amplitude
            sort_idx = localmax.argsort()
            localmax_idx_sorted = localmax_idx[sort_idx]

            # define outputs
            npeaks = localmax_idx_sorted.shape[0]
            valid = np.ones(npeaks, dtype='byte')
            start = np.zeros(npeaks, dtype='int32')
            end = np.zeros(npeaks, dtype='int32')

            # setting some threshold values
            totalrange = np.max(data_pix) - np.min(data_pix)
            amp_thr2 = amp_thr2_factor * totalrange

            # find for each peak the associated local minima and decide whether the peak is valid or not
            for p in range(npeaks):
                idx = localmax_idx_sorted[p]
                # define search window for SOS
                t_idx = time[idx]
                t_min = t_idx - timedelta(days=max_window)
                idx_min = np.where(time == nearest_date(time, t_min))[0][0]
                t_max = t_idx - timedelta(days=min_window)
                idx_max = np.where(time == nearest_date(time, t_max))[0][0]
                # adjust search window in case there is a valid peak within the window
                # find all intermediate VALID peaks
                val_peaks = localmax_idx_sorted.copy()
                val_peaks[valid == 0] = -1
                int_peaks_idx = localmax_idx_sorted[np.logical_and(val_peaks > idx_min, val_peaks < idx)]
                # if any, find the peak nearest to original peak and set t_min to that value
                if int_peaks_idx.shape[0] > 0:
                    idx_min = np.max(int_peaks_idx)
                    # if, by adjusting the window, idx_max < idx_min -> label peak as invalid
                    if idx_max < idx_min:
                        valid[p] = 0
                        continue

                # identify index of local minimum in search window
                win = data_pix[idx_min:idx_max+1]
                start[p] = np.where(win == np.amin(win))[0][0] + idx_min

                # check if amplitude conditions of the identified starting point are met
                amp_dif = data_pix[idx] - data_pix[start[p]]
                if not np.logical_and(amp_dif >= amp_thr1, amp_dif >= amp_thr2):
                    valid[p] = 0
                    continue

                # define search window for EOS
                t_min = t_idx + timedelta(days=min_window)
                idx_min = np.where(time == nearest_date(time, t_min))[0][0]
                t_max = t_idx + timedelta(days=max_window)
                idx_max = np.where(time == nearest_date(time, t_max))[0][0]
                # adjust search window in case there is a valid peak within the window
                # find all intermediate VALID peaks
                val_peaks = localmax_idx_sorted.copy()
                val_peaks[valid == 0] = -1
                int_peaks_idx = localmax_idx_sorted[np.logical_and(val_peaks > idx, val_peaks < idx_max)]
                # if any, find the peak nearest to original peak and set t_max to that value
                if int_peaks_idx.shape[0] > 0:
                    idx_max = np.min(int_peaks_idx)
                    # if, by adjusting the window, idx_max < idx_min -> label peak as invalid
                    if idx_max < idx_min:
                        valid[p] = 0
                        continue

                # identify index of local minimum in search window
                win = data_pix[idx_min:idx_max+1]
                end[p] = np.where(win == np.amin(win))[0][0] + idx_min

                # check if amplitude conditions of the identified end point are met
                amp_dif = data_pix[idx] - data_pix[end[p]]
                if not np.logical_and(amp_dif >= amp_thr1, amp_dif >= amp_thr2):
                    valid[p] = 0
                    continue

            # now delete invalid peaks
            peaks = localmax_idx_sorted[valid == 1]
            start = start[valid == 1]
            end = end[valid == 1]
            npeaks = peaks.shape[0]

            # if more than max_seasons seasons detected -> select the seasons with highest amplitudes
            if npeaks > max_seasons:
                toRemove = npeaks - max_seasons
                maxSeason = data_pix[peaks]
                baseSeason = np.average(np.stack((data_pix[start], data_pix[end]), axis=0), axis=0)
                amp = maxSeason - baseSeason
                idx_remove = np.zeros_like(amp)
                for r in range(toRemove):
                    idx_remove[np.where(amp == np.min(amp))[0][0]] = 1
                    amp[np.where(amp == np.min(amp))[0][0]] = np.max(amp)
                peaks = peaks[idx_remove != 1]
                start = start[idx_remove != 1]
                end = end[idx_remove != 1]
                npeaks = max_seasons

            # if less than max_seasons seasons detected -> add dummy seasons
            if peaks.shape[0] < max_seasons:
                toAdd = np.ones(max_seasons - peaks.shape[0], dtype='int32') * -1
                start = np.concatenate((start, toAdd))
                end = np.concatenate((end, toAdd))
                peaks = np.concatenate((peaks, toAdd))

            # transfer to output
            seasons['mos'][:, i, j] = peaks
            seasons['sos'][:, i, j] = start
            seasons['eos'][:, i, j] = end
            seasons['nseasons'][i, j] = npeaks

    return seasons


def visualize_seasons(timeseries, seasons, x, y):
    '''
    Plot seasons for a given timeseries

    :param timeseries: the Timeseries object from which the seasons were derived
    (should only contain one band!)
    :param seasons is output of function detect_seasons
    :x row index of pixel to be plotted
    :y column index of pixel to be plotted

    '''

    import matplotlib.pyplot as plt
    data = np.squeeze(timeseries.data)
    time = timeseries.timestamps

    # plot timeseries
    f, ax = plt.subplots()
    ts = np.squeeze(data[:, x, y])
    ax.plot(time, ts)

    # plot all seasons for particular pixel
    npeaks = seasons['nseasons'][x, y]
    for p in range(npeaks):
        ax.plot(time[seasons['sos'][p, x, y]], ts[seasons['sos'][p, x, y]], 'go')
        ax.plot(time[seasons['mos'][p, x, y]], ts[seasons['mos'][p, x, y]], 'k+')
        ax.plot(time[seasons['eos'][p, x, y]], ts[seasons['eos'][p, x, y]], 'ro')


def select_seasons(timeseries, seasons, mode, param):
    '''
    FUNCTION TO SELECT SOME SEASONS FROM THOSE IDENTIFIED BY detect_seasons

    :param timeseries: the Timeseries object from which the seasons were derived
    (should only contain one band!)
    :param seasons: original definition of seasons organized in dict (see detect_seasons function)
    :param mode: method used for the selection. Currently the following options are included
    - first -> selection of first x seasons (x defined by param)
    - last -> selection of last x seasons (x defined by param)
    - year -> select all seasons with their peak in a certain year (year defined by param)
    - amplitude -> select x seasons with highest/lowest amplitudes (x defined by param; positive means highest, negative means lowest)
    - length -> select x seasons with highest/lowest length of season (x defined by param; positive means highest, negative means lowest)
    - date -> select season with sos closest to a certain date (date defined by param)
    :param param: additional parameter to be used during selection, see list above

    Returns a dict of selected seasons
    '''
    data = np.squeeze(timeseries.data)
    time = timeseries.timestamps

    # prepare output
    nx = data.shape[1]
    ny = data.shape[2]
    nseas = seasons['sos'].shape[0]
    if (mode in ['first', 'last', 'amplitude', 'length']):
        nseas = param
    elif mode == 'date':
        nseas = 1
    sel_seasons = {'nseasons': np.zeros((nx, ny), dtype='int32'),
                   'sos': np.ones((nseas, nx, ny), dtype='int32') * -1,
                   'mos': np.ones((nseas, nx, ny), dtype='int32') * -1,
                   'eos': np.ones((nseas, nx, ny), dtype='int32') * -1}

    # loop over 2D patch of data
    for i in range(nx):
        for j in range(ny):
            # if no seasons detected -> don't do any selection
            npeaks = seasons['nseasons'][i, j]
            if npeaks == 0:
                continue

            # retrieve valid seasons
            start = seasons['sos'][0:npeaks, i, j]
            peaks = seasons['mos'][0:npeaks, i, j]
            end = seasons['eos'][0:npeaks, i, j]

            # if number of seasons requested is larger or equal than number of valid seasons:
            # select all seasons and fill with dummy seasons
            if (mode in ['first', 'last', 'amplitude', 'length']) and (npeaks <= abs(param)):
                ndif = abs(param) - npeaks
                if not isinstance(start, np.ndarray):
                    start_sel = np.array([start])
                    peaks_sel = np.array([peaks])
                    end_sel = np.array([end])
                else:
                    start_sel = start
                    peaks_sel = peaks
                    end_sel = end
                start_sel = np.concatenate((start, np.ones(ndif, dtype='int32') * -1))
                peaks_sel = np.concatenate((peaks, np.ones(ndif, dtype='int32') * -1))
                end_sel = np.concatenate((end, np.ones(ndif, dtype='int32') * -1))

            # selection of first x seasons (x defined by param)
            elif mode == 'first':
                start_sel = np.zeros(param, dtype='int32')
                peaks_sel = np.zeros(param, dtype='int32')
                end_sel = np.zeros(param, dtype='int32')
                t_start = time[start]
                nsel = 0
                while nsel < param:
                    idx = np.where(t_start == min(t_start))[0][0]
                    start_sel[nsel] = start[idx]
                    peaks_sel[nsel] = peaks[idx]
                    end_sel[nsel] = end[idx]
                    t_start = np.delete(t_start, idx)
                    nsel += 1

            # selection of last x seasons (x defined by param)
            elif mode == 'last':
                start_sel = np.zeros(param, dtype='int32')
                peaks_sel = np.zeros(param, dtype='int32')
                end_sel = np.zeros(param, dtype='int32')
                t_start = time[start]
                nsel = 0
                while nsel < param:
                    idx = np.where(t_start == max(t_start))[0][0]
                    start_sel[nsel] = start[idx]
                    peaks_sel[nsel] = peaks[idx]
                    end_sel[nsel] = end[idx]
                    t_start = np.delete(t_start, idx)
                    nsel += 1

            # select x seasons with highest/lowest amplitudes
            # (x defined by param; positive means highest, negative means lowest)
            elif mode == 'amplitude':
                d = np.squeeze(data[:, i, j])
                toRemove = npeaks - abs(param)
                maxSeason = d[peaks]
                baseSeason = np.average(np.stack(d[start], d[end], axis=0), axis=0)
                amp = maxSeason - baseSeason
                idx_remove = np.zeros_like(amp)
                for r in range(toRemove):
                    if param > 0:
                        idx_remove[np.where(amp == np.min(amp))[0][0]] = 1
                        amp[np.where(amp == np.min(amp))[0][0]] = np.max(amp)
                    else:
                        idx_remove[np.where(amp == np.max(amp))[0][0]] = 1
                        amp[np.where(amp == np.max(amp))[0][0]] = np.min(amp)
                peaks_sel = peaks[idx_remove != 1]
                start_sel = start[idx_remove != 1]
                end_sel = end[idx_remove != 1]

            # select x seasons with highest/lowest length
            # (x defined by param; positive means highest, negative means lowest)
            elif mode == 'length':
                toRemove = npeaks - abs(param)
                lseason = end - start
                idx_remove = np.zeros_like(lseason)
                for r in range(toRemove):
                    if param > 0:
                        idx_remove[np.where(lseason == np.min(lseason))[0][0]] = 1
                        lseason[np.where(lseason == np.min(lseason))[0][0]] = np.max(lseason)
                    else:
                        idx_remove[np.where(lseason == np.max(lseason))[0][0]] = 1
                        lseason[np.where(lseason == np.max(lseason))[0][0]] = np.min(lseason)
                peaks_sel = peaks[idx_remove != 1]
                start_sel = start[idx_remove != 1]
                end_sel = end[idx_remove != 1]

            # select all seasons with their peak in a certain year
            elif mode == 'year':
                peak_times = time[peaks].tolist()
                peak_years = [x.year for x in peak_times]
                test_year = [i for i, elem in enumerate(peak_years) if elem == param]
                # create selection making sure you end up with same amount of seasons as was originally in seasons
                if len(test_year) == 0:
                    peaks_sel = np.ones(nseas, dtype='int32') * -1
                    start_sel = np.ones(nseas, dtype='int32') * -1
                    end_sel = np.ones(nseas, dtype='int32') * -1
                else:
                    # check how many would be removed from original array
                    ndif = seasons['sos'].shape[0] - len(test_year)
                    if len(test_year) == 1:
                        start_sel = np.array([start[test_year]])
                        peaks_sel = np.array([peaks[test_year]])
                        end_sel = np.array([end[test_year]])
                    else:
                        start_sel = start[test_year]
                        peaks_sel = peaks[test_year]
                        end_sel = end[test_year]
                    peaks_sel = np.concatenate((peaks_sel, np.ones(ndif, dtype='int32') * -1))
                    start_sel = np.concatenate((start_sel, np.ones(ndif, dtype='int32') * -1))
                    end_sel = np.concatenate((end_sel, np.ones(ndif, dtype='int32') * -1))

            # select season with sos closest to a certain date
            elif mode == 'date':
                t_start = time[start]
                idx_sel = np.where(t_start == nearest_date(t_start, param))[0][0]
                peaks_sel = np.array([peaks[idx_sel]])
                start_sel = np.array([start[idx_sel]])
                end_sel = np.array([end[idx_sel]])

            else:
                sys.exit('Mode provided was not recognized.')

            # adjust values according to selection
            sel_seasons['sos'][:, i, j] = start_sel
            sel_seasons['mos'][:, i, j] = peaks_sel
            sel_seasons['eos'][:, i, j] = end_sel
            sel_seasons['nseasons'][i, j] = np.count_nonzero(peaks_sel != -1)

    return sel_seasons


################################################################################################################################
# The next functions compute features based on the identified seasons and are mainly meant to describe phenology over
# a longer timespan (e.g. 1 year or more) -> mainly useful to separate agricultural areas from non-agricultural land uses
################################################################################################################################

def pheno_amplitude_features(seasons, timeseries, listFeat=['aSeas_min', 'aSeas_med', 'aSeas_max']):
    '''
    Computes a range of features based on detected seasons and related to a certain VI or band.

    :param seasons: dicstionnary with season information (result of detect_seasons or select_seasons)
    :param timeseries: timeseries for which the features should be computed with dimensions (should be a Timeseries object)
    Note that features will be computed for each band in the timeseries.
    :param listFeat: list of features to be computed
    aSeas_med, aSeas_min and aSeas_max = median, min and max amplitude of seasons

    '''
    # prepare output
    features = []

    # loop over all bands in the input timeseries
    bands = timeseries.bands
    for b in bands:
        data = np.squeeze(timeseries.select_bands([b]).data)
        # compute amplitude of season for each season and pixel
        nx = seasons['nseasons'].shape[0]
        ny = seasons['nseasons'].shape[1]
        nz = seasons['sos'].shape[0]
        aSeas = np.zeros((nz, nx, ny))
        for i in range(nx):
            for j in range(ny):
                if seasons['nseasons'][i, j] == 0:
                    continue
                d = np.squeeze(data[:, i, j])
                nseas = seasons['nseasons'][i, j]
                maxSeason = d[seasons['mos'][0:nseas, i, j]]
                baseSeason = np.average(np.stack((d[seasons['sos'][0:nseas, i, j]],
                                                  d[seasons['eos'][0:nseas, i, j]]), axis=0), axis=0)
                aSeas[0:nseas, i, j] = maxSeason - baseSeason
        # convert all zeros to NaN's
        aSeas[aSeas == 0] = np.nan
        # compute amplitude features and add them to output
        if 'aSeas_min' in listFeat:
            features.append(Features(np.nanmin(aSeas, axis=0), [b + '_aSeas_min']))
        if 'aSeas_med' in listFeat:
            features.append(Features(np.nanmedian(aSeas, axis=0), [b + '_aSeas_med']))
        if 'aSeas_max' in listFeat:
            features.append(Features(np.nanmax(aSeas, axis=0), [b + '_aSeas_max']))

    return features


def pheno_los_features(seasons, time, listFeat=['lSeas_min', 'lSeas_med', 'lSeas_max']):
    '''
    Computes a range of features based on lengths of detected seasons.

    :param seasons: result of detect_seasons or select_seasons
    :param time: array of timestamps, matching "data"
    :param listFeat: list of features to be computed
    lSeas_med, lSeas_min and lSeas_max = median, min and max lenght of seasons

    '''

    features = []

    # compute length of season for each season present per pixel
    nx = seasons['nseasons'].shape[0]
    ny = seasons['nseasons'].shape[1]
    nz = seasons['sos'].shape[0]
    lSeas = np.zeros((nz, nx, ny))
    for i in range(nx):
        for j in range(ny):
            if seasons['nseasons'][i, j] == 0:
                continue
            nseas = seasons['nseasons'][i, j]
            length_seas = time[seasons['eos'][0:nseas, i, j]] - time[seasons['sos'][0:nseas, i, j]]
            for s in range(nseas):
                lSeas[s, i, j] = length_seas[s].days

    # convert all zeros to NaN's
    lSeas[lSeas == 0] = np.nan

    if 'lSeas_min' in listFeat:
        features.append(Features(np.nanmin(lSeas, axis=0), ['lSeas_min']))

    if 'lSeas_med' in listFeat:
        features.append(Features(np.nanmedian(lSeas, axis=0), ['lSeas_med']))

    if 'lSeas_max' in listFeat:
        features.append(Features(np.nanmax(lSeas, axis=0), ['lSeas_max']))

    return features


def pheno_mult_season_features(seasons, timeseries, listFeat=['nSeas', 'lSeas_min', 'lSeas_med', 'lSeas_max',
                                                              'aSeas_min', 'aSeas_med', 'aSeas_max']):
    '''
    Computes phenological features over multiple seasons. Useful to describe seasonality over larger timeframe.

    :param seasons: dicstionnary with season information (result of detect_seasons or select_seasons)
    :param timeseries: timeseries for which the features should be computed (should be a Timeseries object)
        Note that features will be computed for each band in the timeseries.
    :param listFeat: list of features to be computed
    nSeas = number of seasons
    lSeas_med, lSeas_min and lSeas_max = median, min and max lenght of seasons
    aSeas_med, aSeas_min and aSeas_max = median, min and max amplitude of seasons
    '''
    features = []

    if 'nSeas' in listFeat:
        features.append(Features(seasons['nseasons'], ['nSeas']))

    if ('lSeas_min' in listFeat) or ('lSeas_med' in listFeat) or ('lSeas_max' in listFeat):
        time = timeseries.timestamps
        feat_los = pheno_los_features(seasons, time, listFeat=listFeat)
        features.extend(feat_los)

    if ('aSeas_min' in listFeat) or ('aSeas_med' in listFeat) or ('aSeas_max' in listFeat):
        feat_ampl = pheno_amplitude_features(seasons, timeseries, listFeat=listFeat)
        features.extend(feat_ampl)

    features = Features.from_features(*features)
    return features


################################################################################################################################
# The following function is intended to be used for crop type mapping:
# once the season of interest is identified, compute some pheno-related features for that particular season
################################################################################################################################

def pheno_feat_season(seasons, timeseries):
    '''
    Computes a list of pheno features for one particular season and particular timeseries

    :param seasons: dicstionnary with season information (result of detect_seasons or select_seasons)
        Note that only the first season will be taken into account here!
    :param timeseries: timeseries for which the features should be computed (should be a Timeseries object)
        Note that features will be computed for each band in the timeseries.
    '''

    # prepare output
    features = []
    nx = seasons['nseasons'].shape[0]
    ny = seasons['nseasons'].shape[1]

    # first get some feature purely related to timing (day of year and length of season):
    time = timeseries.timestamps
    sos = np.ones((nx, ny)) * np.nan  # sos in day of year
    eos = np.ones((nx, ny)) * np.nan  # eos in day of year
    mos = np.ones((nx, ny)) * np.nan  # peak of season in day of year
    los = np.ones((nx, ny)) * np.nan  # length of season in days
    ttp = np.ones((nx, ny)) * np.nan  # time to peak in days
    tod = np.ones((nx, ny)) * np.nan  # time of decline in days
    for i in range(nx):
        for j in range(ny):
            nseas = seasons['nseasons'][i, j]
            if nseas == 0:
                continue
            sos[i, j] = time[seasons['sos'][0, i, j]].timetuple().tm_yday
            eos[i, j] = time[seasons['eos'][0, i, j]].timetuple().tm_yday
            mos[i, j] = time[seasons['mos'][0, i, j]].timetuple().tm_yday
            los[i, j] = (time[seasons['eos'][0, i, j]] - time[seasons['sos'][0, i, j]]).days
            ttp[i, j] = (time[seasons['mos'][0, i, j]] - time[seasons['sos'][0, i, j]]).days
            tod[i, j] = (time[seasons['eos'][0, i, j]] - time[seasons['mos'][0, i, j]]).days
     # add to features
    features.append(Features(sos, ['phenoSOS']))
    features.append(Features(eos, ['phenoEOS']))
    features.append(Features(mos, ['phenoMOS']))
    features.append(Features(los, ['phenoLOS']))
    features.append(Features(ttp, ['phenoTTP']))
    features.append(Features(tod, ['phenoTOD']))

    # now calculate some features related to amplitude of the timeseries:
    # loop over all bands in the input timeseries
    bands = timeseries.bands
    for b in bands:
        data = np.squeeze(timeseries.select_bands([b]).data)
        # compute basevalue, max value, amplitude and area under curve for each pixel
        phenoAmp = np.ones((nx, ny)) * np.nan
        phenoBase = np.ones((nx, ny)) * np.nan
        phenoMax = np.ones((nx, ny)) * np.nan
        phenoArea = np.ones((nx, ny)) * np.nan
        for i in range(nx):
            for j in range(ny):
                nseas = seasons['nseasons'][i, j]
                if nseas == 0:
                    continue
                d = np.squeeze(data[:, i, j])
                phenoMax[i, j] = d[seasons['mos'][0, i, j]]
                phenoBase[i, j] = np.average((d[seasons['sos'][0, i, j]], d[seasons['eos'][0, i, j]]))
                phenoAmp[i, j] = phenoMax[i, j] - phenoBase[i, j]
                phenoArea[i, j] = np.sum(d[seasons['sos'][0, i, j]: seasons['eos'][0, i, j]])
        # add to features
        features.append(Features(phenoBase, [b + '_phenoBase']))
        features.append(Features(phenoAmp, [b + '_phenoAmp']))
        features.append(Features(phenoMax, [b + '_phenoMax']))
        features.append(Features(phenoArea, [b + '_phenoArea']))

        # identify timing of phenological stages:
        phenoSGU = np.ones((nx, ny)) * np.nan
        phenoMGU = np.ones((nx, ny)) * np.nan
        phenoMAT = np.ones((nx, ny)) * np.nan
        phenoSGD = np.ones((nx, ny)) * np.nan
        phenoMGD = np.ones((nx, ny)) * np.nan
        phenoDOR = np.ones((nx, ny)) * np.nan
        for i in range(nx):
            for j in range(ny):
                nseas = seasons['nseasons'][i, j]
                if nseas == 0:
                    continue
                # cut out first half of the season
                d = np.squeeze(data[seasons['sos'][0, i, j]: seasons['mos'][0, i, j], i, j])
                t = time[seasons['sos'][0, i, j]: seasons['mos'][0, i, j]]
                # timing of green-up phases
                idx = np.where(d > 0.15 * phenoAmp[i, j])
                if idx[0].shape[0] > 0:
                    phenoSGU[i, j] = t[idx[0][0]].timetuple().tm_yday
                idx = np.where(d > 0.50 * phenoAmp[i, j])
                if idx[0].shape[0] > 0:
                    phenoMGU[i, j] = t[idx[0][0]].timetuple().tm_yday
                idx = np.where(d > 0.90 * phenoAmp[i, j])
                if idx[0].shape[0] > 0:
                    phenoMAT[i, j] = t[idx[0][0]].timetuple().tm_yday
                # cut out second half of the season
                d = np.squeeze(data[seasons['mos'][0, i, j]: seasons['eos'][0, i, j], i, j])
                t = time[seasons['mos'][0, i, j]: seasons['eos'][0, i, j]]
                # timing of green-down phases
                idx = np.where(d < 0.90 * phenoAmp[i, j])
                if idx[0].shape[0] > 0:
                    phenoSGD[i, j] = t[idx[0][0]].timetuple().tm_yday
                idx = np.where(d < 0.50 * phenoAmp[i, j])
                if idx[0].shape[0] > 0:
                    phenoMGD[i, j] = t[idx[0][0]].timetuple().tm_yday
                idx = np.where(d < 0.15 * phenoAmp[i, j])
                if idx[0].shape[0] > 0:
                    phenoDOR[i, j] = t[idx[0][0]].timetuple().tm_yday
        # add to features
        features.append(Features(phenoSGU, [b + '_phenoSGU']))
        features.append(Features(phenoMGU, [b + '_phenoMGU']))
        features.append(Features(phenoMAT, [b + '_phenoMAT']))
        features.append(Features(phenoSGD, [b + '_phenoSGD']))
        features.append(Features(phenoMGD, [b + '_phenoMGD']))
        features.append(Features(phenoDOR, [b + '_phenoDOR']))

    features = Features.from_features(*features)
    return features
