import numpy as np

from satio.features import (percentile_iqr, std,
                            kurto, skewness,
                            summation, tsteps)
from worldcereal.utils.masking import SCL_MASK_VALUES
from worldcereal.features.feat_sen2agri import sen2agri_temp_feat
from worldcereal.features.feat_irr import (calc_gvmi, calc_ndwi_veg,
                                           calculate_smad,
                                           cumfeat, cumtsteps)


def get_comp_mode_dict():
    return {
        'dewpoint_temperature': 'median',
        'precipitation_flux': 'sum',
        'solar_radiation_flux': 'median',
        'temperature_max': 'max',
        'temperature_mean': 'median',
        'temperature_min': 'min',
        'vapour_pressure': 'median',
        'wind_speed': 'median',
        'et0': 'median'
    }


# CROPLAND

def get_default_settings():
    return {
        'l2a':
            {
                'bands': ["B02", "B03", "B04", "B05", "B06",
                          "B07", "B08", "B11", "B12"],
                'rsis': ["ndvi", "evi", "anir",
                         "ndmi", "ndwi", "brightness",
                         "ndre1", "ndre2", "ndre3",
                         "ndre4", "ndre5",
                         "savi", "ndgi"
                         ],
                'composite': {
                    'freq': 10,
                    'window': 20,
                    'start': None,
                    'end': None},
                'mask': {
                    'erode_r': 3,
                    'dilate_r': 21,
                    'mask_values': SCL_MASK_VALUES,
                    'multitemporal': True},
                'seasons': {'rsi': 'evi',
                            'amp_thr1': 0.15,
                            'amp_thr2': 0.35,
                            'max_window': 250}
            },
        'sigma0':
            {
                'bands': ['VV', 'VH'],
                'rsis': ['vh_vv'],
                'composite': {
                    'freq': 12,
                    'window': 18,
                    'mode': 'mean',
                    'start': None,
                    'end': None},
                'mask': {
                    'AgERA5_col': '',
                    'precipitation_threshold': 10}
            },
        'agera5':
            {
                'bands': ['precipitation_flux',
                          'temperature_mean',
                          'temperature_min'],
                'composite': {
                    'freq': 1,  # Beware: changing this can have unexpected results for precipitation!  # NOQA
                    'window': 1,
                    'mode': get_comp_mode_dict(),
                    'start': None,
                    'end': None}
            }
    }


def calc_rgbBR(B02, B03, B04):
    return np.sqrt(np.power(B02, 2) + np.power(B03, 2)
                   + np.power(B04, 2))


def get_default_rsi_meta():
    return {
        'l2a':
        {
            'rgbBR': {'bands': ['B02', 'B03', 'B04'],
                      'range': [0, 1],
                      'clamp_range': False,
                      'scale': 1,
                      'native_res': 10,
                      'func': calc_rgbBR
                      }}}


def get_cropland_features_meta():
    return {
        'l2a':
            {
                "std": {
                    "function": std,
                    "names": ['std']},
                "kurt": {
                    "function": kurto,
                    "names": ['kurt']},
                "skew": {
                    "function": skewness,
                    "names": ['skew']},
                "sen2agri_temp_feat": {
                    "function": sen2agri_temp_feat,
                    "parameters": {
                        'time_start': None,
                        'time_freq': 10,
                        'w': 2,
                        'delta': 0.05,
                        'tsoil': 0.2},
                    "bands": ['ndvi'],
                    "names": ['maxdif', 'mindif', 'difminmax',
                              'peak', 'lengthpeak', 'areapeak',
                              'ascarea', 'asclength', 'ascratio',
                              'descarea', 'desclength',
                              'descratio', 'soil1', 'soil2']},
                "pheno_mult_season": {}
            },

        'sigma0':
            {
                "std": {
                    "function": std,
                    "names": ['std']},
            },
        'agera5':
            {
                "sum": {
                    "function": summation,
                    "names": ['sum'],
                    "bands": ['precipitation_flux']},
                "percentile_iqr": {
                    "function": percentile_iqr,
                    "parameters": {'q': [10, 50, 90],
                                   'iqr': [25, 75]},
                    "names": ['p10', 'p50', 'p90', 'iqr'],
                    "bands": ['temperature_mean',
                              'temperature_min',
                              'temperature_max']}
            }
    }


def get_default_ignore_def_feat():
    return {
        'l2a': False,
        'sigma0': False,
        'agera5': True
    }


def get_cropland_tsteps_settings():
    return {
        'l2a':
        {
            'bands': ["B02", "B03", "B04", "B05",
                      "B06", "B08", "B11", "B12"],
            'rsis': [],
            'composite': {
                'freq': 10,
                'window': 20,
                'start': None,
                'end': None},
            'mask': {
                'erode_r': 3,
                'dilate_r': 21,
                'mask_values': SCL_MASK_VALUES,
                'multitemporal': True},
        },
        'sigma0':
        {
            'bands': ['VV', 'VH'],
            'rsis': [],
            'composite': {
                'freq': 12,
                'window': 18,
                'mode': 'mean',
                'start': None,
                'end': None},
            'mask': {
                'AgERA5_col': '',
                'precipitation_threshold': 10}
        },
        'agera5':
        {
            'bands': [
                'temperature_mean',
            ],
            'composite': {
                'freq': 30,
                'window': 30,
                'mode': get_comp_mode_dict(),
                'start': None,
                'end': None}
        }
    }


def get_cropland_tsteps_features_meta():
    return {
        'l2a':
        {
            "tsteps": {
                "function": tsteps,
                "parameters": {
                    'n_steps': 36,
                },
                "names": [f'ts{i}' for i in range(36)]
            }
        },
        'sigma0':
        {
            "tsteps": {
                "function": tsteps,
                "parameters": {
                    'n_steps': 30,
                },
                "names": [f'ts{i}' for i in range(30)]
            }
        },
        'agera5':
        {
            "tsteps": {
                "function": tsteps,
                "parameters": {
                    'n_steps': 12,
                },
                "names": [f'ts{i}' for i in range(12)],
                "bands": ['temperature_mean']
            }
        }
    }


def get_cropland_tsteps_parameters():
    return {
        'settings': get_cropland_tsteps_settings(),
        'features_meta': get_cropland_tsteps_features_meta(),
        'rsi_meta': get_default_rsi_meta(),
        'ignore_def_feat': {
            'l2a': True,
            'sigma0': True,
            'agera5': True
        },
        'gddnormalization': False
    }


# CROP TYPE


def get_croptype_features_meta():
    return {
        'l2a':
            {
                "std": {
                    "function": std,
                    "names": ['std']},
                "kurt": {
                    "function": kurto,
                    "names": ['kurt']},
                "skew": {
                    "function": skewness,
                    "names": ['skew']},
                "sen2agri_temp_feat": {
                    "function": sen2agri_temp_feat,
                    "parameters": {
                        'time_start': None,
                        'time_freq': 10,
                        'w': 2,
                        'delta': 0.05,
                        'tsoil': 0.2},
                    "bands": ['ndvi'],
                    "names": ['maxdif', 'mindif', 'difminmax',
                              'peak', 'lengthpeak', 'areapeak',
                              'ascarea', 'asclength', 'ascratio',
                              'descarea', 'desclength',
                              'descratio', 'soil1', 'soil2']},
                "pheno_mult_season": {},
                "pheno_single_season": {"select_season": {"mode": "length",
                                                          "param": 1}}
            },

        'sigma0':
            {
                "std": {
                    "function": std,
                    "names": ['std']}
            },
        'agera5':
            {
                "sum": {
                    "function": summation,
                    "names": ['sum'],
                    "bands": ['precipitation_flux']},
                "percentile_iqr": {
                    "function": percentile_iqr,
                    "parameters": {'q': [10, 50, 90],
                                   'iqr': [25, 75]},
                    "names": ['p10', 'p50', 'p90', 'iqr'],
                    "bands": ['temperature_mean',
                              'temperature_min',
                              'temperature_max']}
            }
    }


def get_croptype_tsteps_settings():
    return {
        'l2a':
        {
            'bands': ["B02", "B03", "B04", "B05",
                      "B06", "B08", "B11", "B12"],
            'rsis': ["evi"],
            'composite': {
                'freq': 10,
                'window': 20,
                'start': None,
                'end': None},
            'mask': {
                'erode_r': 3,
                'dilate_r': 21,
                'mask_values': SCL_MASK_VALUES,
                'multitemporal': True},
            'seasons': {'rsi': 'evi',
                        'amp_thr1': 0.15,
                        'amp_thr2': 0.35,
                        'max_window': 250}
        },
        'sigma0':
        {
            'bands': ['VV', 'VH'],
            'rsis': [],
            'composite': {
                'freq': 12,
                'window': 18,
                'mode': 'mean',
                'start': None,
                'end': None},
            'mask': {
                'AgERA5_col': '',
                'precipitation_threshold': 10}
        },
        'agera5':
        {
            'bands': [
                'temperature_mean'
            ],
            'composite': {
                'freq': 30,
                'window': 30,
                'mode': get_comp_mode_dict(),
                'start': None,
                'end': None}
        }
    }


def get_croptype_tsteps_features_meta():
    return {
        'l2a':
        {
            "tsteps": {
                "function": tsteps,
                "parameters": {
                    'n_steps': 18,
                },
                "names": [f'ts{i}' for i in range(18)]
            },
            "pheno_mult_season": {},
        },
        'sigma0':
        {
            "tsteps": {
                "function": tsteps,
                "parameters": {
                    'n_steps': 15,
                },
                "names": [f'ts{i}' for i in range(15)]
            }
        },
        'agera5':
        {
            "tsteps": {
                "function": tsteps,
                "parameters": {
                    'n_steps': 6,
                },
                "names": [f'ts{i}' for i in range(6)],
                "bands": ['temperature_mean']
            }
        }
    }


def get_croptype_tsteps_parameters():
    return {
        'settings': get_croptype_tsteps_settings(),
        'features_meta': get_croptype_tsteps_features_meta(),
        'rsi_meta': get_default_rsi_meta(),
        'ignore_def_feat': {
            'l2a': True,
            'sigma0': True,
            'agera5': True
        },
        'gddnormalization': True
    }

# IRRIGATION


def get_irr_parameters():
    return {
        'settings': get_irr_settings(),
        'features_meta': get_irr_features_meta(),
        'rsi_meta': get_irr_rsi_meta(),
        'ignore_def_feat': get_irr_ignore_def_feat(),
        'gddnormalization': False
    }


def get_irr_settings():
    return {
        'l2a': {
            'bands': [],
            'rsis': ["ndvi", "evi", "mndwi", "ndwiveg", "gvmi"],
            'composite': {
                'freq': 10,
                'window': 20,
                'start': None,
                'end': None},
            'mask': {
                'erode_r': 3,
                'dilate_r': 21,
                'mask_values': SCL_MASK_VALUES,
                'multitemporal': True},
            'smad': {'bands_10': ['B02', 'B03', 'B04', 'B08'],
                     'bands_20': ['B11', 'B12'],
                     'function': calculate_smad}
        },
        'agera5': {'bands': ['precipitation_flux'],
                   'rsis': ['et0', 'et', 'prdef',
                            'sm', 'smstress'],
                   'composite': {
            'freq': 10,
            'window': None,
            'mode': get_comp_mode_dict(),
            'start': None,
            'end': None},
            'et': {'method': 'ndvi_linear'}
        },
        'l8': {
            'bands': [],
            'rsis': ['lst_ta'],
            'composite': {
                'freq': 16,
                'window': 32,
                'mode': 'median',
                'start': None,
                'end': None
            },
            'mask': {
                'erode_r': 3,
                'dilate_r': 21,
                'max_invalid_ratio': 1}
        }
    }


def get_irr_rsi_meta():
    return {
        'l2a': {
            'ndwiveg': {'bands': ['B08', 'B12'],
                        'range': [-1, 1],
                        'clamp_range': False,
                        'scale': 10000,
                        'native_res': 20,
                        'func': calc_ndwi_veg},
            'gvmi': {'bands': ['B08', 'B12'],
                     'range': [-1, 1],
                     'clamp_range': False,
                     'scale': 10000,
                     'native_res': 20,
                     'func': calc_gvmi}
        },
        'agera5': {
            'sm': {
                'bands': ['dewpoint_temperature',
                          'temperature_mean',
                          'wind_speed',
                          'vapour_pressure',
                          'solar_radiation_flux'],
                'range': [0, 1],
                'clamp_range': False,
                'scale': 1,
                'native_res': 20},
            'smstress': {
                'bands': ['dewpoint_temperature',
                          'temperature_mean',
                          'wind_speed',
                          'vapour_pressure',
                          'solar_radiation_flux'],
                'range': [0, 1],
                'clamp_range': False,
                'scale': 1,
                'native_res': 20}
        },
        'l8': {
            'lst_ta': {'bands': ['ST-B10'],
                       'range': [200, 400],
                       'clamp_range': False,
                       'scale': 1,
                       'native_res': 10}
        }
    }


def get_irr_features_meta():
    return {
        'agera5': {
            "std": {
                "function": std,
                "names": ['std']},
            "cumfeat": {
                "function": cumfeat,
                "parameters": {
                    'composite_freq': 10},
                "names": ['cum_max', 'cum_min',
                          'cum_npeaks', 'cum_maxdur',
                          'cum_maxslope'],
                "bands": ['prdef']},
            "cumtsteps": {
                "function": cumtsteps,
                "parameters": {
                    'n_steps': 6},
                "names": ['cumts0', 'cumts1', 'cumts2',
                          'cumts3', 'cumts4', 'cumts5'],
                "bands": ['prdef']},
            "sum": {
                "function": summation,
                "names": ['sum'],
                "bands": ['precipitation_flux',
                          'et0',
                          'et',
                          'prdef']}
        },
        'l2a': {
            "std": {
                "function": std,
                "names": ['std']}
        },
        'l8': {
            "std": {
                "function": std,
                "names": ['std']},
            "cumtsteps": {
                "function": cumtsteps,
                "parameters": {
                    'n_steps': 6,
                },
                "names": ['cumts0', 'cumts1', 'cumts2',
                          'cumts3', 'cumts4', 'cumts5']
            }
        }
    }


def get_irr_ignore_def_feat():
    return {
        'l2a': False,
        'agera5': False,
        'l8': False
    }
