import xarray as xr
from loguru import logger
import pandas as pd
import numpy as np
from satio.timeseries import Timeseries

from worldcereal.seasons import (doy_from_tiff,
                                 doy_to_date_after)


def computeGDD(tcube: xr.DataArray,
               tbase: int = 0,
               upper_limit: int = None,
               units: str = "C"):
    """
    Compute GDD value for each timestep in the tcube

    INPUTS
    - tcube - A xarray.dataarray.dataArray or numpy.ndarray object
              with X, Y and Timestamp as coordinates
    - tbase - Integer. Base temperature to the GDD computation
    - upper_limit - Max temperature to not accumulate temperature
    - units - Temperature units
    """

    if units == "C":
        tcube = tcube - 273.15
    elif units == "K":
        pass
    else:
        raise RuntimeError((f"`units` should be one of "
                            "['C', 'K'] but got {units}."))

    if "ndarray" in str(type(tcube)):
        gdd = tcube.copy()
        if upper_limit:
            gdd[gdd >= upper_limit] = upper_limit
        gdd = gdd - tbase
        gdd[gdd < 0] = 0

    else:
        gdd = tcube.copy()
        gdd.values[gdd.values <= tbase] = 0
        if upper_limit:
            gdd.values[gdd.values >= upper_limit] = upper_limit
        gdd.values = gdd.values - tbase
        gdd.values[gdd.values < 0] = 0

    return gdd


def winterwheatGDD(band):
    out = computeGDD(band, tbase=0)
    return out


def maizeGDD(band):
    out = computeGDD(band, tbase=10)
    return out


class GDDcomputer:

    def __init__(self,
                 agera5coll,
                 tbase: int,
                 bounds: tuple = None,
                 epsg: int = None,
                 upper_limit: int = None,
                 units: str = "C",
                 start_date: str = None,
                 end_date: str = None):

        self.coll = agera5coll
        self.tbase = tbase
        self.bounds = bounds
        self.epsg = epsg
        self.upper_limit = upper_limit
        self.units = units
        self.start_date = start_date
        self.end_date = end_date

    def load_temperature(self):
        logger.info('Loading temperature_mean time series.')
        ts = self.coll.load_timeseries('temperature_mean',
                                       resolution=10000)

        if self.start_date is not None:
            idx = np.where(ts.timestamps >= pd.to_datetime(
                self.start_date).to_pydatetime())[0]
            ts.data = ts.data[:, idx, ...]
            ts.timestamps = ts.timestamps[idx]

        if self.end_date is not None:
            idx = np.where(ts.timestamps <= pd.to_datetime(
                self.end_date).to_pydatetime())[0]
            ts.data = ts.data[:, idx, ...]
            ts.timestamps = ts.timestamps[idx]

        return ts

    def compute_accumulated_gdd(self, season: str = None):
        '''
        Main method to compute accumulated GDD.
        If the optional `season` argument is provided
        then accumulation will be done from the respective SOS onward.
        Otherwise accumulation is done over the entire time series.
        '''
        logger.info('Start computation accumulated GDD.')

        # Load temperature time series
        temperature_ts = self.load_temperature()

        # if all Nan's -> block is situated completely over the ocean
        if np.all(np.isnan(temperature_ts.data)):
            raise ValueError(('Accumulated GDD resulted in all NaNs, '
                              'block probably not located over land!'))

        # If a season is specified we need to determine
        # the start date for accumulation
        if season is not None:

            if self.bounds is None:
                raise ValueError(('`bounds` cannot be None when '
                                  'requesting a specific season'))
            if self.epsg is None:
                raise ValueError(('`epsg` cannot be None when '
                                  'requesting a specific season'))

            sos_doy = doy_from_tiff(season,
                                    'SOS',
                                    self.bounds,
                                    self.epsg,
                                    10000)
            if sos_doy.size != 1:
                raise ValueError(('Extracted SOS DOY data has size '
                                  f'`{sos_doy.size}` while only size '
                                  '1 is supported'))
            elif int(sos_doy) == 0:
                raise ValueError(('Extracted SOS DOY is 0, meaning '
                                  'no seasonality for this crop is '
                                  'available in this region'))
            else:
                sos_doy = int(sos_doy)

                # Now convert DOY to an actual date
                sos_date = doy_to_date_after(sos_doy,
                                             temperature_ts.timestamps[0])

                logger.info(f'Accumulating GDD from: {sos_date}')

                # Finally, put all temperature before the sos_date
                # low enough so they won't contribute to accumulation
                idx = np.where(temperature_ts.timestamps < pd.to_datetime(
                    sos_date).to_pydatetime())[0]

                temperature_ts.data[:, idx, ...] = -100

        # Compute accumulated GDD
        gdd = computeGDD(temperature_ts.data, tbase=self.tbase,
                         upper_limit=self.upper_limit, units=self.units)
        accumulated_gdd = gdd.cumsum(axis=1)
        accumulated_gdd_ts = Timeseries(
            data=accumulated_gdd,
            timestamps=temperature_ts.timestamps,
            bands=['accumulated_gdd'],
            attrs=dict(
                tbase=self.tbase
            )
        )
        logger.info('Computation accumulated GDD finished.')
        return accumulated_gdd_ts
