from loguru import logger
from pathlib import Path
import numpy as np
import matplotlib.cm

from satio.utils.cog import cog_from_folder


def get_confidence_cmap():
    colormap = matplotlib.cm.get_cmap('RdYlGn')
    cmap = {}
    for i in range(101):
        cmap[i] = tuple((np.array(colormap(int(2.55 * i))) * 255).astype(int))
    return cmap


COLORMAP = {
    'active-cropland': {
        0: (232, 55, 39, 255),  # inactive
        1: (77, 216, 39, 255),  # active
    },
    'annualcropland': {
        0: (186, 186, 186, 0),  # no cropland
        100: (224, 24, 28, 200),  # cropland
    },
    'cropland': {
        0: (186, 186, 186, 0),  # no cropland
        100: (224, 24, 28, 200),  # cropland
    },
    'winterwheat': {
        0: (186, 186, 186, 255),  # other
        1: (186, 113, 53, 255),  # wheat
    },
    'springwheat': {
        0: (186, 186, 186, 255),  # other
        1: (186, 113, 53, 255),  # wheat
    },
    'wheat': {
        0: (186, 186, 186, 255),  # other
        1: (186, 113, 53, 255),  # wheat
    },
    'maize1': {
        0: (186, 186, 186, 255),  # other
        1: (252, 207, 5, 255),  # maize
    },
    'maize2': {
        0: (186, 186, 186, 255),  # other
        1: (252, 207, 5, 255),  # maize
    },
    'irr-maize1': {
        0: (255, 132, 132, 255),  # rainfed
        1: (104, 149, 197, 255)  # irrigated
    },
    'irr-maize2': {
        0: (255, 132, 132, 255),  # rainfed
        1: (104, 149, 197, 255)  # irrigated
    },
    'irr-winterwheat': {
        0: (255, 132, 132, 255),  # rainfed
        1: (104, 149, 197, 255)  # irrigated
    },
    'confidence': get_confidence_cmap(),
    'metafeatures': None
}
NODATA = {
    'active-cropland': 255,
    'annualcropland': None,
    'cropland': None,
    'winterwheat': 255,
    'springwheat': 255,
    'wheat': 255,
    'maize1': 255,
    'maize2': 255,
    'irr-maize1': 255,
    'irr-maize2': 255,
    'irr-winterwheat': 255,
    'confidence': 255,
    'metafeatures': None
}


def postprocess_tile(indir, outdir, product, tile, year, modelname,
                     skip_processed=False, in_memory=True):

    blocks_folder = Path(indir) / 'blocks'

    for name in ['classification',
                 'confidence',
                 'metafeatures',
                 'active-cropland']:
        if name == 'active-cropland':
            tif_folder = (blocks_folder / tile /
                          (str(year) + f'_{product}') / name)
            out_fn = outdir / tile / '_'.join([str(year), tile, product,
                                               name + '.tif'])
        else:
            tif_folder = (blocks_folder / tile /
                          (str(year) + f'_{product}') / modelname /
                          name)
            out_fn = outdir / tile / '_'.join([str(year), tile, product,
                                               name, modelname + '.tif'])

        out_fn.parent.mkdir(parents=True, exist_ok=True)

        # If we're looking at the classification itself we'll get
        # the options from the product.
        identifier = name if name != 'classification' else product

        if out_fn.is_file() and skip_processed:
            logger.warning(f'{out_fn} already exists, skipping')
        elif not Path(tif_folder).is_dir():
            logger.warning(f'{tif_folder} does not exist, skipping')
        else:
            if out_fn.is_file() and not skip_processed:
                logger.warning(f'{out_fn} exists, removing')
                out_fn.unlink()

            if identifier in COLORMAP.keys():
                colormap = COLORMAP[identifier]
            else:
                logger.warning(('Unknown identifier `{identifier}`: '
                                'cannot assign colormap'))
                colormap = None

            if identifier in NODATA.keys():
                nodata = NODATA[identifier]
            else:
                logger.warning((f'Unknown identifier `{identifier}`: '
                                'cannot assign nodata value'))
                nodata = None

            try:
                import glob
                import rasterio
                ex_file = glob.glob(str(tif_folder / '*.tif'))[0]
                with rasterio.open(ex_file) as src:
                    band_names = eval(src.get_tag_item('bands'))

                if nodata is None:
                    cog_from_folder(tif_folder,
                                    out_fn,
                                    in_memory=in_memory,
                                    colormap=colormap,
                                    band_names=band_names)
                else:
                    cog_from_folder(tif_folder,
                                    out_fn,
                                    in_memory=in_memory,
                                    colormap=colormap,
                                    nodata=nodata,
                                    band_names=band_names)
            except Exception as e:
                logger.error(f'{e}')
