from satio.utils.errors import EmptyCollection
from typing import Dict, List

from loguru import logger
import geopandas as gpd
from shapely.geometry import Polygon
import numpy as np
import pandas as pd
import gc
import os
import json
import glob
from pathlib import Path
import copy
import rasterio
import time
from skimage.segmentation import felzenszwalb
import tensorflow as tf

from satio.features import Features
from satio.utils import run_parallel, TaskTimer
from satio.utils import rasterize
from satio.utils.geotiff import write_geotiff
from satio.utils.geotiff import get_rasterio_profile

from worldcereal.classification.models import WorldCerealModel
from worldcereal.utils import io
from worldcereal.utils.scalers import minmaxscaler
from worldcereal.utils.io import check_collection
from worldcereal.utils import is_real_feature
from worldcereal.features.settings import (get_default_settings,
                                           get_cropland_features_meta,
                                           get_default_rsi_meta,
                                           get_default_ignore_def_feat)
from worldcereal.fp import (L2AFeaturesProcessor,
                            L8ThermalFeaturesProcessor,
                            SIGMA0FeaturesProcessor,
                            WorldCerealSIGMA0FeaturesProcessor,
                            WorldCerealOpticalFeaturesProcessor,
                            WorldCerealThermalFeaturesProcessor,
                            WorldCerealAgERA5FeaturesProcessor,
                            AgERA5FeaturesProcessor)
from worldcereal.clf import WorldCerealClassifier, mask, majority_filter
from worldcereal.gdd import GDDcomputer
from worldcereal.seasons import infer_season_dates
from worldcereal import SEASON_BUFFER, SUPPORTED_SEASONS, TBASE
from worldcereal.postprocess import COLORMAP, NODATA


class BlockProcessor:

    supported_sensors = ['L2A',
                         'SIGMA0',
                         'AgERA5',
                         'DEM',
                         'WorldCover',
                         'L8']

    def __init__(self,
                 collections_dict,
                 settings,
                 rsi_meta,
                 features_meta,
                 ignore_def_feat,
                 bounds,
                 epsg,
                 custom_fps: Dict = None,
                 featresolution=10):

        self.collections = collections_dict
        self.bounds = bounds
        self.epsg = epsg

        self._feat_handles = custom_fps or {
            'L2A': L2AFeaturesProcessor,
            'SIGMA0': SIGMA0FeaturesProcessor,
            'L8': L8ThermalFeaturesProcessor,
            'AgERA5': AgERA5FeaturesProcessor,
            'DEM': Features.from_dem,
            'WorldCover': Features.from_worldcover
        }

        self.settings = settings
        self.rsi_meta = rsi_meta
        self.features_meta = features_meta
        self.ignore_def_feat = ignore_def_feat

        self.featresolution = featresolution

        self._check_sensors()

    def _check_sensors(self):

        for s in self.collections.keys():
            if s not in self.supported_sensors:
                raise ValueError(f"{s} is not a supported sensor. "
                                 "Supported sensors: "
                                 f"'{self.supported_sensors}'.")

    def get_features(self) -> Features:

        collections = self.collections
        feat_handles = self._feat_handles
        settings = self.settings
        rsi_meta = self.rsi_meta
        features_meta = self.features_meta
        ignore_def_feat = self.ignore_def_feat

        sensors = [s for s in collections.keys()
                   if s not in ['LABELS', 'DEM',
                                'WorldCover']]

        # now run all other feature computations
        features = []

        # add agera5col to sigma0 settings if available and needed
        if settings.get('sigma0', None):
            if settings['sigma0'].get('mask', None):
                if settings['sigma0']['mask'].get('AgERA5_col', None) == '':
                    if 'AgERA5' in sensors:
                        settings['sigma0']['mask'][
                            'AgERA5_col'] = collections['AgERA5']
                    else:
                        raise ValueError('When masking of sigma0 is required,'
                                         ' you need to specify an AgERA5 '
                                         'collection')

        # add dem, bounds and epsg to agera5 settings if available and needed
        if settings.get('agera5', None):
            if settings['agera5'].get('rsis', None):
                etrsis = ['et0', 'et', 'prdef']
                etcheck = [i for i in etrsis
                           if i in settings['agera5']['rsis']]
                if len(etcheck) > 0:
                    if 'DEM' in collections.keys():
                        settings['agera5']['demcol'] = collections['DEM']
                        settings['agera5']['bounds'] = self.bounds
                        settings['agera5']['epsg'] = self.epsg
                    else:
                        raise ValueError(
                            'When ET0, ET or prdef needs to be computed,'
                            'a DEM collection is required!')

        # add L2A and L8 collections to agera5 settings if ET or prdef required
        if settings.get('agera5', None):
            if settings['agera5'].get('rsis', None):
                etrsis = ['et', 'prdef', 'sm', 'smstress']
                etcheck = [i for i in etrsis
                           if i in settings['agera5']['rsis']]
                if len(etcheck) > 0:
                    if settings['agera5'].get('et', None):
                        if 'L2A' in collections.keys():
                            settings[
                                'agera5']['et']['col'] = collections['L2A']
                        else:
                            raise ValueError(
                                'When ET or prdef needs to be computed,'
                                'an L2A collection is required!')
                    else:
                        raise ValueError(
                            'Need ET settings for ET calculation!')
                    if 'L8' not in collections.keys():
                        raise ValueError('Need L8 collection for '
                                         ' computation of ET, SM or prdef!')
                    else:
                        settings['agera5']['sm'] = {'l8col': collections['L8']}

        # add agera5 collection to l8 settings if lst_ta required
        if settings.get('l8', None):
            if settings['l8'].get('rsis', None):
                if 'lst_ta' in settings['l8']['rsis']:
                    if 'AgERA5' in collections.keys():
                        settings['l8']['agera5col'] = collections['AgERA5']
                    else:
                        raise ValueError('lst_ta feature required -> '
                                         'provide agera5 collection!')

        # Compute accumulated GDD and add to settings if needed
        accumulated_gdd = None
        for s in sensors:
            if 'normalize_gdd' in settings.get(s.lower(), {}).keys():
                if accumulated_gdd is None:
                    if 'AgERA5' not in collections:
                        raise RuntimeError(('An `AgERA5` collection '
                                            'is required when GDD '
                                            'normalization is requested.'))
                    # Compute accumulated GDD
                    tbase = settings[s.lower()]['normalize_gdd']['tbase']
                    season = settings[s.lower()]['normalize_gdd']['season']
                    start_date = settings[s.lower()]['composite']['start']
                    end_date = settings[s.lower()]['composite']['end']
                    gddcomp = GDDcomputer(collections['AgERA5'],
                                          tbase=tbase,
                                          bounds=self.bounds,
                                          epsg=self.epsg,
                                          start_date=start_date,
                                          end_date=end_date)
                    accumulated_gdd = gddcomp.compute_accumulated_gdd(
                        season=season
                    )
                else:
                    if settings[s.lower()]['normalize_gdd']['tbase'] != tbase:
                        raise RuntimeError(
                            (f'The `Tbase` value of {s} '
                             f'is different from the Tbase value {tbase}'
                             'that was used for another sensor. These '
                             'should be equal.'))

                # Add accumulated gdd to settings
                settings[s.lower()]['normalize_gdd'][
                    'accumulated_gdd'] = accumulated_gdd

        for s in sensors:
            sensor_settings = settings.get(s.lower(), {})
            if sensor_settings:
                feat = feat_handles[s](
                    collections[s],
                    sensor_settings,
                    rsi_meta=rsi_meta.get(s.lower(), {}),
                    features_meta=features_meta.get(s.lower(),
                                                    {}),
                    ignore_def_features=ignore_def_feat.get(
                        s.lower(), False)).compute_features()

                # append sensor to feature names
                featnames = feat.names
                featnames = [s + '-' + f for f in featnames]
                feat.names = featnames
                features.append(feat)

        if 'DEM' in collections.keys():
            feat = feat_handles['DEM'](
                collections['DEM'],
                settings.get('dem', {}),
                resolution=self.featresolution)
            features.append(feat)

        if 'WorldCover' in collections.keys():
            feat = feat_handles['WorldCover'](
                collections['WorldCover'],
                resolution=self.featresolution)
            features.append(feat)

        features = Features.from_features(*features)
        features = features.add_latlon(self.bounds,
                                       self.epsg,
                                       resolution=self.featresolution)

        features = features.add_pixelids()

        return features


class TrainingBlockProcessor(BlockProcessor):

    supported_sensors = ['L2A',
                         'SIGMA0',
                         'DEM',
                         'WorldCover',
                         'AgERA5',
                         'LABELS',
                         'L8']

    def __init__(self,
                 collections_dict,
                 settings,
                 rsi_meta,
                 features_meta,
                 ignore_def_feat,
                 bounds,
                 epsg,
                 location_id,
                 ref_id,
                 start_date,
                 end_date,
                 **kwargs):

        super().__init__(collections_dict,
                         settings,
                         rsi_meta,
                         features_meta,
                         ignore_def_feat,
                         bounds,
                         epsg,
                         **kwargs)

        self.location_id = location_id
        self.ref_id = ref_id
        self.start_date = start_date
        self.end_date = end_date

    def _get_labels(self):

        location_id = self.location_id

        labels_collection = self.collections['LABELS']
        labels = labels_collection.filter_location(location_id)

        # Load the labels
        labelsdata = labels.load()

        labelfeatures = Features(data=labelsdata['CT'], names=[
            'CT'], dtype='uint16')
        labelfeatures = labelfeatures.merge(Features(data=labelsdata['LC'],
                                                     names=['LC'],
                                                     dtype='uint16'),
                                            Features(data=labelsdata['IRR'],
                                                     names=['IRR'],
                                                     dtype='uint16'))

        if self.featresolution == 20:
            labelfeatures = labelfeatures.downsample_categorical()

        return labelfeatures

    def get_features(self):

        start_date = self.start_date
        end_date = self.end_date
        settings = self.settings
        features_meta = self.features_meta

        # Now make sure start_date and end_date are set in the settings
        # of all collections for this sample
        for coll in settings.keys():
            settings[coll]['composite']['start'] = start_date
            settings[coll]['composite']['end'] = end_date

        if 'sen2agri_temp_feat' in features_meta.get('l2a', {}):
            features_meta['l2a'][
                'sen2agri_temp_feat'][
                    'parameters']['time_start'] = start_date

        features = super().get_features()

        # add labels
        features = features.merge(self._get_labels())

        # add location_id
        features = features.add_attribute(self.location_id,
                                          'location_id')
        features = features.add_attribute(self.ref_id, 'ref_id')

        # add the split category
        # check which sensor is present
        if 'L2A' in self.collections.keys():
            sens = 'L2A'
        elif 'SIGMA0' in self.collections.keys():
            sens = 'SIGMA0'
        elif 'L8' in self.collections.keys():
            sens = 'L8'
        else:
            sens = 'AgERA5'
        features = features.add_attribute(
            self.collections[sens].df['split'].values[0],
            'split')

        return features


class TrainingChain:
    """
    Trains a model based on the training collections and settings provided.

    Attributes
    ----------
    collections_dict : dict
        Supported keys: ['L2A', 'GAMMA0', 'SIGMA0', 'AgERA5', 'DEM', 'LABELS'].
        'L2A' and 'LABELS' are required and should be instances of
        L2ATrainingCollection and PatchLabelsTrainingCollection respectively.

    settings : Dict
        Specifies parameters for the features computation.
        Dictionary with the same keys as `collections_dict` except for
        'LABELS'. It specifies different FeaturesProcessor parameters, as
        well as providing features functions and features names.

    rsi_meta : Dict
        Optionally provides information on how to calculate particular RSIs.
        Dictionary with same keys as "collections_dict". If a certain key
        is not provided, then the default in satio will be
        selected for that sensor.

    features_meta : Dict
        Optionally provides information on how to calculate certain features.
        Dictionary with same keys as "collections_dict". If a certain key
        is not provided, then the default in satio will be selected for that
        sensor.

    location_ids : List
        List of location_ids from which to build the training dataframe

    geometry : Polygon
        shapely polygon in lat lon to retrieve location_ids.

    trainingdb : geopandas.GeoDataFrame
        gdf of the training locations with required attributes

    aez : List
        should be a list of tuples (geodataframe, name, value_column).
        If value_column is None, the default will be the 'Index'. These
        layers will be added as a rasterized feature and one hot encoded.

    season: str
        specification of the season for which this chain is valid.
        Currently one of [WW, M1, M2, annual, custom].

    start_date: str (yyyy-mm-dd)
        an optional custom start date for which to run the chain.
        Is only taken into account for season `custom`

    end_date: str (yyyy-mm-dd)
        an optional custom end date for which to run the chain.
        Is only taken into account for season `custom`

    prior_buffer: int
        an optional buffer (days) to add prior to the season start.

    avg_segm: boolean
        whether or not to compute the average of features per segment

    segm_feat: list
        list of features to be used for segmentation

    featresolution: int (optional)
        resolution of computed features (default = 10)

    """

    def __init__(self,
                 collections_dict: Dict,
                 settings: Dict,
                 rsi_meta: Dict = None,
                 features_meta: Dict = None,
                 ignore_def_feat: Dict = None,
                 location_ids: List = None,
                 geometry: Polygon = None,
                 trainingdb: gpd.GeoDataFrame = None,
                 aez: gpd.GeoDataFrame = None,
                 basedir: str = None,
                 model: WorldCerealModel = None,
                 season: str = 'custom',
                 start_date: str = None,
                 end_date: str = None,
                 prior_buffer: int = None,
                 avg_segm: bool = False,
                 segm_feat: List = None,
                 featresolution: int = 10):

        # -------------------------------------------------------------
        # FIRST DO THE NECESSARY CHECKS IF WE HAVE EVERYTHING WE NEED

        if season not in SUPPORTED_SEASONS:
            raise ValueError(f'Unrecognized `season` parameter: {season}')

        if season != 'custom' and (
                start_date is not None or end_date is not None):
            logger.warning((f'Using season `{season}`. '
                            'start_date and end_date arguments are ignored'))

        if season == 'custom' and (start_date is None or end_date is None):
            raise ValueError(('When using `custom` season, start_date and '
                              'end_date arguments are required'))

        if (location_ids is None) & (geometry is None):
            raise ValueError("One argument between `location_ids` "
                             "and `geometry` should be provided. If both are "
                             "provided `geometry` is ignored.")

        if trainingdb is None:
            raise ValueError('`trainingdb` cannot be None')

        # -------------------------------------------------------------

        self.collections = collections_dict
        self.settings = settings
        self.rsi_meta = rsi_meta or get_default_rsi_meta()
        self.features_meta = features_meta or get_cropland_features_meta()
        self.ignore_def_feat = ignore_def_feat or get_default_ignore_def_feat()
        self.season = season
        self.start_date = start_date
        self.end_date = end_date
        self.prior_buffer = prior_buffer or SEASON_BUFFER[season]
        self.trainingdb = trainingdb
        self.location_ids = (location_ids if location_ids is not None
                             else self._get_geometry_locations(geometry))
        self.aez = aez
        self.avg_segm = avg_segm
        self.segm_feat = segm_feat

        self._basedir = basedir
        if basedir is not None:
            os.makedirs(basedir, exist_ok=True)
            logger.add(Path(basedir) / 'main.log', mode='w')

        if model is not None:
            if not isinstance(model, WorldCerealModel):
                raise ValueError(('`model` should be instance '
                                  'of WorldCerealModel but got: '
                                  f'{type(model)}.'))
            self.model = model

        self.featresolution = featresolution
        self.timer = _ChainTimer()

    def _get_geometry_locations(self, geometry):
        trainingdb = self.trainingdb
        location_ids = trainingdb[trainingdb.intersects(
            geometry)].location_id.values
        return location_ids

    def _filter_collections(self, location_id, bounds=None, epsg=None):

        if (bounds is None) | (epsg is None):
            bounds, epsg = self._get_location_bounds(location_id)

        new_collections = self.collections.copy()
        for s in self.collections.keys():
            if s in ['L2A', 'SIGMA0', 'AgERA5']:
                try:
                    new_collections[s] = (new_collections[s]
                                          .filter_location(location_id))
                except EmptyCollection:
                    _ = new_collections.pop(s)

            elif s in ['DEM', 'WorldCover']:
                new_collections[s] = (new_collections[s]
                                      .filter_bounds(bounds, epsg))
            else:
                pass

        return new_collections

    def _get_location_features(self, location_id, custom_fps=None):

        bounds, epsg = self._get_location_bounds(location_id)
        collections = self._filter_collections(location_id, bounds, epsg)
        logger.debug(f"Processing location_id {location_id}")

        # ----------------------------------------------
        # INFER START DATE AND END DATE FOR THIS SAMPLE
        # ----------------------------------------------

        sample = self.trainingdb.set_index('location_id').loc[location_id]
        ref_id = sample['ref_id']

        if self.season == 'custom':
            start_date = self.start_date
            end_date = self.end_date
        else:
            # Sample validity time
            valid_date = sample.validityTi
            start_date, end_date = infer_season_dates(self.season,
                                                      bounds, epsg,
                                                      valid_date)

        # Subtract optional buffer prior to season start
        if self.prior_buffer > 0:
            start_date = (pd.to_datetime(start_date) -
                          pd.Timedelta(
                days=self.prior_buffer)).strftime('%Y-%m-%d')

        # We have our start_date and end_date; now check
        # if this range is covered by the sample
        adjusted = False
        if pd.to_datetime(sample['start_date']) > pd.to_datetime(start_date):
            if self.season == 'annual':
                adjusted = True
                # Temporary (?) fix for annual cropland
                logger.warning(('Default annual cropland range not covered'
                                ' by sample. Adjusting to sample range!'))
                start_date = pd.to_datetime(sample['start_date'])
                end_date = start_date + pd.Timedelta(days=364)
                start_date = start_date.strftime('%Y-%m-%d')
                end_date = end_date.strftime('%Y-%m-%d')
            else:
                raise ValueError((f'Needed start_date ({start_date}) '
                                  ' is before sample start_date '
                                  f'({sample["start_date"]})!'))
        if pd.to_datetime(sample['end_date']) < pd.to_datetime(end_date):
            if self.season == 'annual' and not adjusted:
                # Temporary (?) fix for annual cropland
                logger.warning(('Default annual cropland range not covered'
                                ' by sample. Adjusting to sample range!'))
                end_date = pd.to_datetime(sample['end_date'])
                start_date = end_date - pd.Timedelta(days=364)
                start_date = start_date.strftime('%Y-%m-%d')
                end_date = end_date.strftime('%Y-%m-%d')
            else:
                raise ValueError((f'Needed end_date ({end_date}) '
                                  ' is after sample end_date '
                                  f'({sample["end_date"]})!'))

        logger.debug(('Start and end of to be used for this sample: '
                      f'{start_date} - {end_date}'))

        # -----------------------------------------------

        settings = copy.deepcopy(self.settings)
        rsi_meta = self.rsi_meta
        features_meta = self.features_meta
        ignore_def_feat = self.ignore_def_feat
        featresolution = self.featresolution

        train_processor = TrainingBlockProcessor(collections,
                                                 settings,
                                                 rsi_meta,
                                                 features_meta,
                                                 ignore_def_feat,
                                                 bounds,
                                                 epsg,
                                                 location_id,
                                                 ref_id,
                                                 start_date,
                                                 end_date,
                                                 custom_fps=custom_fps,
                                                 featresolution=featresolution)

        loc_feat = train_processor.get_features()

        avg_segm = self.avg_segm
        segm_feat = self.segm_feat
        if avg_segm:
            logger.info('Running segmentation...')
            # select the features for segmentation
            if segm_feat is None:
                raise ValueError('Need list of features to '
                                 'perform the segmentation on!')
            elif len(segm_feat) < 3:
                raise ValueError('You need at least 3 features'
                                 ' to perform segmentation.')
            sfeat = loc_feat.select(segm_feat)
            # scale the features
            scaledFeat = []
            for feat in sfeat.names:
                f = sfeat.select([feat]).data
                scaledData = minmaxscaler(f, feat)
                scaledFeat.append(Features(scaledData, [feat]))
            scaledFeat = Features.from_features(*scaledFeat)
            # replace nan's by zeros
            scaledFeat.data[np.isnan(scaledFeat.data)] = 0
            # run PCA if more than 3 features selected
            if len(sfeat.names) > 3:
                scaledFeat = scaledFeat.pca(num_components=3)
            # transform data into right shape
            inputs = np.moveaxis(scaledFeat.data, 0, -1)
            # apply segmentation algorithms
            segments_fz = felzenszwalb(inputs, scale=100, sigma=0.5,
                                       min_size=20)
            logger.info('Segmentation done!')
            logger.info('Computing average features per segment')
            newdata = loc_feat.data.copy()
            for segm in np.unique(segments_fz):
                msk = segments_fz == segm
                for i in range(newdata.shape[0]):
                    newdata[i, msk == 1] = np.mean(loc_feat[i, msk == 1])
            loc_feat = loc_feat._clone(data=newdata)
            logger.info('Done!')

        if self.aez is not None:
            groupid = int(
                rasterize(self.aez, bounds, epsg,
                          resolution=640,  # One pixel in 640m patch
                          value_column='groupID',
                          fill_value=-9999,
                          dtype=np.int32).ravel()[0])
            zoneid = int(
                rasterize(self.aez, bounds, epsg,
                          resolution=640,  # One pixel in 640m patch
                          value_column='zoneID',
                          fill_value=-9999,
                          dtype=np.int32).ravel()[0])

            loc_feat = loc_feat.add_attribute(groupid,
                                              'aez_groupid')
            loc_feat = loc_feat.add_attribute(zoneid,
                                              'aez_zoneid')
        else:
            loc_feat = loc_feat.add_attribute(None, 'aez_groupid')
            loc_feat = loc_feat.add_attribute(None, 'aez_zoneid')

        return loc_feat

    def _get_location_bounds(self, location_id):
        trainingdb = self.trainingdb
        trainingdb_row = trainingdb[
            trainingdb.location_id == location_id].iloc[0]
        epsg = int(trainingdb_row['epsg'])
        bounds = trainingdb_row['bounds']

        if isinstance(bounds, str):
            bounds = eval(bounds)
        bounds = np.array(bounds)

        return bounds, epsg

    def _get_feat(self, location_id, ft_selection=None, custom_fps=None):
        try:
            features = self._get_location_features(location_id,
                                                   custom_fps=custom_fps)

            if ft_selection is not None:
                if type(ft_selection) is not list:
                    raise ValueError('`ft_selection` should be a list!')
                features = features.select(ft_selection)

        except Exception as e:
            # raise e
            features = None
            logger.error(f"Error processing location_id = {location_id}: "
                         f"{e}.")
        return features

    def select_fts_with_meta(self, features: Features, ft_selection):
        if type(ft_selection) is not list:
            raise ValueError('`ft_selection` should be a list!')
        all_fts = features.names
        real_fts = [ft for ft in all_fts if self.is_real_feature(ft)]
        real_selected_fts = [ft for ft in real_fts if ft in ft_selection]
        meta_fts = list(set(all_fts) - set(real_fts))
        selected_fts = real_selected_fts + meta_fts

        return features.select(selected_fts)

    @ staticmethod
    def is_real_feature(ft_name):
        return is_real_feature(ft_name)

    def save_settings(self, outdir=None):

        def _make_serial(d):
            if isinstance(d, dict):
                return {_make_serial(k): _make_serial(v)
                        for k, v in d.items()}
            elif isinstance(d, list):
                return [_make_serial(i) for i in d]
            elif callable(d):
                return d.__name__
            else:
                return d

        outdir = outdir or self._basedir

        logger.info(f'Saving settings to: {outdir}')

        with open(Path(outdir) / 'settings.json', 'w') as f:
            json.dump(_make_serial(self.settings), f)
        with open(Path(outdir) / 'rsi_meta.json', 'w') as f:
            json.dump(_make_serial(self.rsi_meta), f)
        with open(Path(outdir) / 'features_meta.json', 'w') as f:
            json.dump(_make_serial(self.features_meta), f)
        with open(Path(outdir) / 'ignore_def_feat.json', 'w') as f:
            json.dump(_make_serial(self.ignore_def_feat), f)

    def train(self):
        raise NotImplementedError

    def predict(self, location_id):
        if self.model is None:
            raise ValueError(('No model assigned to '
                              'this chain.'))

        self.clf = WorldCerealClassifier(self.model)

        # Compute the features
        starttime = time.time()
        features = self._get_feat(location_id)
        time_feat = time.time() - starttime

        if features is None:
            return None, None, None
        else:
            # Make the prediction and return together with
            # the features
            starttime = time.time()
            prediction, confidence = self.clf.predict(features)
            time_pred = time.time() - starttime
            return prediction, confidence, features, (time_feat, time_pred)


class PatchTrainingChain(TrainingChain):

    def __init__(self, windowsize=64, **kwargs):

        self.chaintype = 'patch'
        self.windowsize = windowsize
        super().__init__(**kwargs)

    def get_training_patches(self,
                             sparkcontext=None,
                             ft_selection=None,
                             **kwargs
                             ):

        logger.info((f'Processing {len(self.location_ids)} samples ...'))

        # Save the current settings
        self.save_settings()

        if sparkcontext is None:
            df = self._get_training_patches_local(
                ft_selection=ft_selection,
                **kwargs)
            return df
        else:
            self._get_training_patches_spark(sparkcontext,
                                             ft_selection=ft_selection,
                                             **kwargs)

    def _get_training_patches_spark(self,
                                    sparkcontext,
                                    ft_selection=None,
                                    outputbasename='inputpatches',
                                    outputformat='tfrecord',
                                    nrparts=None):
        '''
        Method to get and save patch features on spark
        '''

        if self._basedir is None:
            raise ValueError(('Cannot run export job without '
                              f'specifying `basedir`.'))

        if outputformat not in ['tfrecord', 'parquet']:
            raise ValueError(f'Unrecognized outputformat: `{outputformat}`')

        outdir = Path(self._basedir)

        logger.info('Running jobs on spark ...')
        rdd = sparkcontext.parallelize(self.location_ids,
                                       len(self.location_ids)).map(
            lambda location_id: self._get_feat(
                location_id,
            )).filter(lambda ft: ft is not None)

        # Check if we actually have something left
        if rdd.isEmpty():
            logger.warning('No samples left in RDD -> aborting job')
            if (Path(outdir) / str(outputbasename)).is_dir():
                (Path(outdir) / str(outputbasename)).unlink()
            return
        # Convert features to spark SQL rows
        rdd = rdd.map(lambda ft: self._convert_ft_to_row(
            ft,
            windowsize=self.windowsize,
            ft_selection=ft_selection,
            on_spark=True))

        # Convert to dataframe
        output_df = rdd.toDF()

        # Get feature names in the rdd
        present_fts = output_df.schema.names
        present_fts_file = (Path(self._basedir) /
                            'present_features.txt')
        if present_fts_file.is_file():
            present_fts_file.unlink()
        with open(present_fts_file, 'w') as f:
            for ft in present_fts:
                f.write(str(ft) + '\n')

        # Amount of output files
        nrparts = nrparts or int(output_df.count() / 500 + 1)

        if outputformat == 'tfrecord':
            self._rdd_to_tfrecords(output_df, nrparts,
                                   outdir, str(outputbasename))
        elif outputformat == 'parquet':
            self._rdd_to_parquet(output_df, nrparts,
                                 outdir, str(outputbasename))

    def _get_training_patches_local(self,
                                    ft_selection=None,
                                    max_workers=2,
                                    save=True,
                                    debug=False,
                                    **kwargs):

        def get_feat_df(location_id):
            try:
                features = self._get_location_features(location_id)

                self._convert_ft_to_row(features,
                                        windowsize=self.windowsize,
                                        ft_selection=ft_selection,
                                        on_spark=False,
                                        scalefactor=1)
            except Exception as e:
                if debug:
                    raise e
                logger.error(f"Error processing location_id = {location_id}: "
                             f"{e}.")
                return None

        results = run_parallel(get_feat_df,
                               self.location_ids,
                               max_workers=max_workers)

        train_df = pd.DataFrame(results)

        if save:

            # Get feature names in the rdd
            present_fts = train_df.columns.tolist()
            present_fts_file = (Path(self._basedir) /
                                'present_features.txt')
            if present_fts_file.is_file():
                present_fts_file.unlink()
            with open(present_fts_file, 'w') as f:
                for ft in present_fts:
                    f.write(str(ft) + '\n')

            train_df_file = (Path(self._basedir)
                             / f'training_df.csv')
            logger.info(f'Saving training df to: {train_df_file}')
            train_df.to_csv(train_df_file)

        return train_df

    def _convert_ft_to_row(self, ft, windowsize=64,
                           ft_selection=None, scalefactor=10000,
                           on_spark=True):
        '''
        helper function to transform the resulting features into a
        proper spark dataframe row for saving
        :param ft: features to save
        :param windowsize: the patch size that is being used
        :param scalefactor: integer factor used to scale the values
        :return: spark SQL row
        '''
        if on_spark:
            from pyspark.sql.types import Row

        location_id = ft.attrs['location_id']
        ref_id = ft.attrs['ref_id']
        split = ft.attrs['split']
        aez_groupid = ft.attrs['aez_groupid']
        aez_zoneid = ft.attrs['aez_zoneid']

        if ft_selection is not None:
            ft = self.select_fts_with_meta(ft,
                                           ft_selection)

        converted_ft = {}

        for ft_name in ft.names:
            # Get np.ndarray data of this feature
            ftdata = ft[ft_name]

            if on_spark:
                dtype = int
            else:
                dtype = ftdata.dtype

            converted_ft[ft_name] = (ftdata.reshape(
                (windowsize*windowsize)) * scalefactor).astype(
                    dtype).tolist()

        if on_spark:
            return Row(**converted_ft, location_id=location_id, split=split,
                       aez_groupid=aez_groupid, aez_zoneid=aez_zoneid,
                       ref_id=ref_id)
        else:
            return dict(**converted_ft, location_id=location_id, split=split,
                        aez_groupid=aez_groupid, aez_zoneid=aez_zoneid,
                        ref_id=ref_id)

    @ staticmethod
    def _rdd_to_tfrecords(sparkdf, nrparts, outdir, outname):
        logger.info('Saving results to tfrecords ...')
        sparkdf.repartition(nrparts).write.format('tfrecords')\
            .mode('overwrite').option('recordType', 'SequenceExample')\
            .option("codec", "org.apache.hadoop.io.compress.GzipCodec")\
            .save('file://' + os.path.join(outdir, outname))

    @ staticmethod
    def _rdd_to_parquet(sparkdf, nrparts, outdir, outname):
        logger.info('Saving results to parquet ...')
        sparkdf.repartition(nrparts).write.format('parquet')\
            .mode('overwrite')\
            .save('file://' + os.path.join(outdir, outname))


class PixelTrainingChain(TrainingChain):

    def __init__(self, **kwargs):

        self.chaintype = 'pixel'
        super().__init__(**kwargs)

    @ staticmethod
    def _get_df(features):
        if features is None:
            return None
        else:
            return features.df

    def get_training_df(self, sparkcontext=None, **kwargs):

        # timer = self.timer
        # timer.trainingdf.start()

        # Save the current settings
        self.save_settings()

        if sparkcontext is None:
            train_df = self._get_training_df_local(**kwargs)
        else:
            train_df = self._get_training_df_spark(sparkcontext,
                                                   **kwargs)
        # timer.trainingdf.stop()
        # if self.location_ids is not None:
        #     timer.trainingdf._total = (timer.trainingdf._total /
        #                                len(self.location_ids))
        # timer.trainingdf.log()

        return train_df

    def _get_training_df_local(self,
                               ft_selection=None,
                               label='LC',
                               max_pixels=1,
                               max_workers=2,
                               save=True,
                               debug=False,
                               seed=1234,
                               format='csv',
                               custom_fps=None):

        if format not in ['csv', 'parquet']:
            raise ValueError(f'Unsupported output format: `{format}`')

        def get_feat_df(location_id):
            try:
                features = self._get_location_features(location_id,
                                                       custom_fps=custom_fps)

                if ft_selection is not None:
                    features = self.select_fts_with_meta(features,
                                                         ft_selection)

            except Exception as e:
                if debug:
                    raise e
                features = None
                logger.error(f"Error processing location_id = {location_id}: "
                             f"{e}.")
            return self.sample_pixels(
                self._get_df(features),
                nrpixels=max_pixels,
                label=label,
                seed=seed)

        results = run_parallel(get_feat_df,
                               self.location_ids,
                               max_workers=max_workers)
        results = [result for result in results if result is not None]

        train_df = pd.concat(results, ignore_index=True)
        train_df = train_df.rename(columns={label: 'OUTPUT'})

        if save:
            if format == 'csv':
                train_df_file = (Path(self._basedir)
                                 / f'training_df_{label}.csv')
                logger.info(f'Saving training df to: {train_df_file}')
                train_df.to_csv(train_df_file)
            if format == 'parquet':
                train_df_file = (Path(self._basedir)
                                 / f'training_df_{label}.parquet')
                logger.info(f'Saving training df to: {train_df_file}')
                train_df.to_parquet(train_df_file)
        return train_df

    def _get_training_df_spark(self,
                               sparkcontext,
                               ft_selection=None,
                               label='LC',
                               max_pixels=1,
                               save=True,
                               seed=1234,
                               format='csv',
                               custom_fps=None,
                               **kwargs
                               ):

        if format not in ['csv', 'parquet']:
            raise ValueError(f'Unsupported output format: `{format}`')

        logger.info('Running feature extraction on spark ...')
        rdd = sparkcontext.parallelize(self.location_ids,
                                       len(self.location_ids)).map(
            lambda location_id: self._get_feat(
                location_id, ft_selection=ft_selection,
                custom_fps=custom_fps
            )).filter(lambda t: t is not None).map(
                self._get_df).map(
                    lambda df: self.sample_pixels(
                        df, nrpixels=max_pixels,
                        label=label, seed=seed)).filter(
                            lambda t: t is not None)

        # Check if we actually have something left
        if rdd.isEmpty():
            logger.warning('No samples left in RDD -> aborting job')
            return

        dfs = rdd.collect()
        train_df = pd.concat(dfs, ignore_index=True)
        train_df = train_df.rename(columns={label: 'OUTPUT'})

        if save:
            if format == 'csv':
                train_df_file = (Path(self._basedir)
                                 / f'training_df_{label}.csv')
                logger.info(f'Saving training df to: {train_df_file}')
                train_df.to_csv(train_df_file)
            if format == 'parquet':
                train_df_file = (Path(self._basedir)
                                 / f'training_df_{label}.parquet')
                logger.info(f'Saving training df to: {train_df_file}')
                train_df.to_parquet(train_df_file)

        return train_df

    def train_one_vs_all(self,
                         traindf,
                         targetlabels,
                         nodatalabel=0):
        """Method to train a binary pixel-based classifier
        on a training dataframe

        Args:
            traindf (pd.DataFrame or str): training dataframe or path to csv
            targetlabels (str or list): the target classes to identify
            nodatalabel (int, optional): Target label corresponding to unknown.
            Defaults to 0.

        Raises:
            ValueError: if no model has been assigned to the trainingchain
            ValueError: if training dataframe is not a dataframe or string
        """
        timer = self.timer

        if self.model is None:
            raise ValueError('No model assigned yet.')

        if type(traindf) is str:
            traindf = pd.read_csv(traindf, index_col=0)
        elif type(traindf) is pd.DataFrame:
            pass
        else:
            raise ValueError(('`traindf` should be either '
                              ' a path to a csv or a '
                              'pd.DataFrame instance. Got: '
                              f'{type(traindf)}'))

        inputfeatures = self.model.feature_names

        # Check if all required features are in the DF
        for ft in inputfeatures:
            if ft not in traindf.columns.tolist():
                logger.debug(f'traindf shape: {traindf.shape}')
                raise ValueError((f'Expected input feature `{ft}` '
                                  'not found in trainingDF.'))

        logger.info(('Start training on '
                     f'{len(inputfeatures)} features ...'))

        traindfcal = traindf[traindf['split'] == 'CAL']
        traindfval = traindf[traindf['split'] == 'VAL']

        inputscal = traindfcal[inputfeatures].values
        outputscal = traindfcal['OUTPUT'].values
        inputsval = traindfval[inputfeatures].values
        outputsval = traindfval['OUTPUT'].values

        inputsval, outputsval = io.drop_nan(
            *io.drop_unknown(inputsval, outputsval))
        inputscal, outputscal = io.drop_nan(
            *io.drop_unknown(inputscal, outputscal))

        outputscal = io.convert_to_binary(
            outputscal,
            target_labels=targetlabels)
        outputsval = io.convert_to_binary(
            outputsval,
            target_labels=targetlabels)

        timer.trainModel.start()
        self.model.train(inputscal, outputscal)
        timer.trainModel.stop()
        timer.evalModel.start()
        self.model.evaluate(inputsval, outputsval)
        timer.evalModel.stop()
        modelfile = (Path(self._basedir) /
                     self.model.modeltype)
        self.model.save(modelfile)

        timer.trainModel.log()
        timer.evalModel.log()

    @ staticmethod
    def sample_pixels(df: pd.DataFrame,
                      nrpixels=1,
                      label='LC',
                      nodata_value=0,
                      seed=1234):
        '''
        Method to sample one or more valid pixels
        from the training dataframe
        '''
        np.random.seed(seed)

        labeldata = df[label].values

        if label != 'LC':
            # also filter on annual cropland
            idxvalid = np.where((labeldata != nodata_value)
                                & (df['LC'].values == 11))[0]
        else:
            idxvalid = np.where(labeldata != nodata_value)[0]

        if len(idxvalid) == 0:
            return None
        elif len(idxvalid) > nrpixels:
            # first, select the center pixel
            idxcenter = int(df.shape[0]/2)
            # check whether center pixel is valid
            if idxcenter in idxvalid:
                idxchoice = [idxcenter]
                nrpixels = nrpixels - 1
            else:
                idxchoice = []
            if nrpixels > 1:
                if len(idxchoice) > 0:
                    # remove idxchoice from idxvalid
                    idxvalid = np.setdiff1d(idxvalid, np.array(idxchoice))
                # now add random pixels
                idxchoice.extend(list(np.random.choice(
                    idxvalid,
                    size=nrpixels,
                    replace=False)))

            df_sample = df.iloc[idxchoice, :]
        else:
            df_sample = df.iloc[idxvalid, :]

        return df_sample


class SegmentTrainingChain(TrainingChain):

    def __init__(self, **kwargs):

        self.chaintype = 'segment'
        super().__init__(**kwargs)

        raise NotImplementedError


class ClassificationProcessor:

    supported_sensors = ['L2A',
                         'SIGMA0',
                         'AgERA5',
                         'DEM',
                         'L8']

    def __init__(self,
                 output_folder,
                 product: str,
                 models: Dict,
                 collections: Dict,
                 season: str = 'annual',
                 settings: Dict = None,
                 rsi_meta: Dict = None,
                 features_meta: Dict = None,
                 ignore_def_feat: Dict = None,
                 gdd_normalization: bool = False,
                 fps: Dict = None,
                 aez: gpd.GeoDataFrame = None,
                 start_date='20190101',
                 end_date='20200101',
                 save_confidence=False,
                 save_meta=False,
                 save_features=False,
                 filtersettings: Dict = None,
                 featresolution: int = 10,
                 **kwargs):

        logger.debug("Initializing ClassificationProcessor")

        self.collections = {k: v for k, v in collections.items()
                            if v is not None}
        self.aez = aez
        self.output = Path(output_folder)
        self.product = end_date[0:4] + f'_{product}'
        self.models = models
        self.filtersettings = filtersettings
        self.season = season
        # self._check_models()

        # # Initialize classifiers
        # self.classifiers = {name: WorldCerealClassifier(model)
        #                     for name, model in self.models.items()}

        self.settings = settings or get_default_settings()
        self.rsi_meta = rsi_meta or get_default_rsi_meta()
        self.features_meta = features_meta or get_cropland_features_meta()
        self.ignore_def_feat = ignore_def_feat or get_default_ignore_def_feat()
        self.gdd_normalization = gdd_normalization
        self._start_date = start_date
        self._end_date = end_date
        self._save_features = save_features
        self._save_confidence = save_confidence
        self._save_meta = save_meta
        self.featresolution = featresolution

        self._check_sensors()

        # Configure the featuresprocessors
        fps = fps or {}
        self._fps = {
            'L2A': fps.get('L2A', WorldCerealOpticalFeaturesProcessor),
            'SIGMA0': fps.get('SIGMA0', WorldCerealSIGMA0FeaturesProcessor),
            'AgERA5': fps.get('AgERA5', WorldCerealAgERA5FeaturesProcessor),
            'DEM': fps.get('DEM', Features.from_dem),
            'L8': fps.get('L8', WorldCerealThermalFeaturesProcessor)
        }

        # Now make sure start_date and end_date are set in the settings
        # of all collections
        for coll in self.settings.keys():
            self.settings[coll]['composite']['start'] = start_date
            self.settings[coll]['composite']['end'] = end_date

        if 'sen2agri_temp_feat' in self.features_meta.get('l2a', {}):
            self.features_meta['l2a'][
                'sen2agri_temp_feat'][
                    'parameters']['time_start'] = start_date

        # Add optional GDD normalization to the settings as well
        if self.gdd_normalization:
            for coll in self.settings.keys():
                self.settings[coll]['normalize_gdd'] = dict(
                    tbase=TBASE[season],
                    season=season
                )

        logger.debug("ClassificationProcessor initialized")

    def _check_sensors(self):

        for s in self.collections.keys():
            if s not in self.supported_sensors:
                raise ValueError(f"{s} is not a supported sensor. "
                                 "Supported sensors: "
                                 f"'{self.supported_sensors}'.")

    def _filtered_collections(self, collections, tile, bounds, epsg):

        # Filter collections spatially
        data_collections = {k: v.filter_bounds(bounds, epsg)
                            for k, v in collections.items()}

        # Filter temporal collections
        # We filter a bit broader to allow for extra data
        # prior to the requested start can be used in compositing
        collstart = (pd.to_datetime(self._start_date) -
                     pd.Timedelta(days=10)).strftime('%Y-%m-%d')
        collend = (pd.to_datetime(self._end_date) +
                   pd.Timedelta(days=10)).strftime('%Y-%m-%d')
        for collection in data_collections:
            if collection != 'DEM':
                data_collections[collection] = (
                    data_collections[collection].filter_dates(
                        collstart,
                        collend
                    )
                )

        # If we have L2A product, additionally filter on the
        # tile
        if 'L2A' in data_collections:
            data_collections['L2A'] = (
                data_collections['L2A'].filter_tiles(
                    tile)
            )

        # If we have SIGMA0 product, additionally filter on the
        # tile
        if 'SIGMA0' in data_collections:
            data_collections['SIGMA0'] = (
                data_collections['SIGMA0'].filter_tiles(
                    tile)
            )

        # If we have L8 product, additionally filter on the
        # tile
        if 'L8' in data_collections:
            data_collections['L8'] = (
                data_collections['L8'].filter_tiles(
                    tile)
            )

        # Finally filter on SCL 60m which we do
        # AFTER checking collection completeness
        if 'L2A' in data_collections:
            data_collections['L2A'] = data_collections['L2A'].filter_nodata()

        logger.info('-' * 50)

        return data_collections

    def get_features(self, tile, bounds, epsg,
                     settings_override=None,
                     rsi_meta_override=None,
                     features_meta_override=None,
                     ignore_def_feat_override=None,
                     collections_override=None
                     ) -> Features:

        def num_to_feat(data, names, shapex, shapey):

            data = [np.full((shapex, shapey), d) for d in data]
            data = np.stack(data, axis=0)
            return Features(data, names, dtype=np.int16)

        settings = settings_override or self.settings
        rsi_meta = rsi_meta_override or self.rsi_meta
        features_meta = features_meta_override or self.features_meta
        ignore_def_feat = ignore_def_feat_override or self.ignore_def_feat
        collections = collections_override or self.collections

        logger.info("Filtering collections")
        data_collections = self._filtered_collections(collections,
                                                      tile, bounds, epsg)

        logger.info("Starting features computation")
        # Compute the features
        features = BlockProcessor(
            data_collections,
            settings,
            rsi_meta,
            features_meta,
            ignore_def_feat,
            bounds,
            epsg,
            custom_fps=self._fps,
            featresolution=self.featresolution).get_features()

        # Adding meta features related to gaps in collections
        for collection in data_collections:
            if 'DEM' not in collection:
                coll_gaps, names = check_collection(
                    data_collections[collection],
                    collection,
                    self._start_date, self._end_date)
                coll_gaps_feat = num_to_feat(coll_gaps, names,
                                             features.data.shape[1],
                                             features.data.shape[2])
                features = features.merge(coll_gaps_feat)

        if self.aez is not None:
            logger.debug("Adding aez feature.")
            raise NotImplementedError('Cannot use aez like this yet.')

        return features

    def process(self,
                tile,
                bounds,
                epsg,
                block_id):

        # Set output folder
        output_folder = self.output / 'blocks'

        # Compute the features
        features = self.get_features(tile, bounds, epsg)

        # Put feature data on disk
        import zarr
        logger.info('Saving to zarr on disk')
        z = zarr.open('features.zarr', mode='w',
                      shape=features.data.shape,
                      chunks=(512, 512), dtype=np.float32)
        z[:] = features.data
        logger.info('Switching features data to disk version')
        features.data = z

        # Now do the predictions based on all provided models
        # and write result to disk
        try:

            for modelname, model in self.models.items():

                wcmodel = WorldCerealModel.from_config(model)
                self.classifier = WorldCerealClassifier(
                    wcmodel,
                    filtersettings=self.filtersettings)

                prediction, confidence = self.classifier.predict(features)
                prediction = prediction * 100
                prediction = prediction.astype(np.uint8)
                confidence = confidence.astype(np.uint8)
                self.save(prediction, tile, bounds, epsg,
                          output_folder, block_id=block_id,
                          modelname=modelname, tag='classification')

                if self._save_confidence:
                    self.save(confidence, tile, bounds, epsg,
                              output_folder, block_id=block_id,
                              modelname=modelname, tag='confidence')

                if self._save_features:
                    self.save_features(features, output_folder, modelname,
                                       tile, block_id, bounds, epsg)

                if self._save_meta:
                    self.save_meta(features, output_folder, modelname,
                                   tile, block_id, bounds, epsg)

                # Clear session to avoid memory leak
                tf.keras.backend.clear_session()
                prediction = None
                confidence = None
                gc.collect()

        except Exception as e:
            logger.error(f"Error predicting for {tile} - {bounds} "
                         f"- {block_id}: {e}")
            raise

    def save(self, predictiondata, tile, bounds,
             epsg, save_folder, block_id=999, modelname='model',
             tag='classification', colormap=None,
             nodata=None):

        fname = self._product_fname(save_folder,
                                    self.product,
                                    modelname,
                                    tag,
                                    tile,
                                    block_id)
        fname.parent.mkdir(parents=True, exist_ok=True)
        self._to_geotiff(predictiondata, bounds, epsg, fname,
                         colormap=colormap, nodata=nodata)

    @ staticmethod
    def _to_geotiff(data, bounds, epsg, filename,
                    band_names=[], colormap=None,
                    nodata=None):

        if data.ndim == 2:
            data = np.expand_dims(data, axis=0)

        profile = get_rasterio_profile(data, bounds, epsg)
        logger.debug(f"Saving {filename}...")

        write_geotiff(data, profile, filename,
                      band_names=band_names, colormap=colormap,
                      nodata=nodata)

    @ staticmethod
    def _product_fname(output_folder, product, modelname, tag, tile, block_id):
        product_filename = (output_folder / str(tile) / product / modelname /
                            tag / f"{tile}_{block_id:03d}_{tag}.tif")
        return product_filename

    @ staticmethod
    def is_real_feature(ft_name):
        return is_real_feature(ft_name)

    def scale_features(self, features):

        scaled_features = []

        # Scale the data
        for ft in features.names:
            ftdata = features.select([ft]).data
            ftscaled = minmaxscaler(ftdata,
                                    ft_name=ft,
                                    clamp=(-1.5, 1.5))
            scaled_features.append(Features(data=ftscaled,
                                            names=[ft]))

        return Features.from_features(*scaled_features)

    def save_features(self, features, output_folder, modelname,
                      tile, block_id, bounds, epsg):
        logger.info('Saving features ...')
        used_fts = features.select(self.classifier.feature_names)
        scaled_fts = self.scale_features(used_fts)
        # Fts clamped from [-0.5, 1.5]
        scaled_fts.data = (scaled_fts.data + 0.5) * 100
        scaled_fts.data = scaled_fts.data.astype(np.uint8)
        fname = self._product_fname(
            output_folder, self.product, modelname,
            'inputfeatures', tile, block_id)
        fname.parent.mkdir(parents=True, exist_ok=True)
        scaled_fts.to_geotiff(bounds, epsg, fname)

    def save_meta(self, features, output_folder, modelname,
                  tile, block_id, bounds, epsg):
        logger.info('Saving metadata ...')
        meta_ft_names = [
            ft_name for ft_name in features.names
            if not self.is_real_feature(ft_name) and
            ft_name not in ['lat', 'lon', 'pixelids']]
        if len(meta_ft_names) == 0:
            logger.warning('No metafeatures found.')
            return
        meta_fts = features.select(meta_ft_names)
        meta_fts.data = meta_fts.data.astype(np.uint8)
        fname = self._product_fname(
            output_folder, self.product, modelname,
            'metafeatures', tile, block_id)
        fname.parent.mkdir(parents=True, exist_ok=True)
        meta_fts.to_geotiff(bounds, epsg, fname)


class CropTypeProcessor(ClassificationProcessor):
    def __init__(self, *args,
                 season='custom',
                 gdd_normalization=True,
                 cropland_mask=None,
                 active_marker=True,
                 irrigation=False,
                 irrcollections=None,
                 irrparameters=None,
                 irrmodels=None,
                 **kwargs):

        if season == 'custom':
            logger.warning(('Using `custom` season. Are '
                            'you sure this is what you want?'))

        super().__init__(*args,
                         season=season,
                         gdd_normalization=gdd_normalization,
                         **kwargs)

        self.active_marker = active_marker
        self.cropland_mask = cropland_mask
        self.irrigation = irrigation
        self.irrcollections = irrcollections
        self.irrparameters = irrparameters
        self.irrmodels = irrmodels

        self._check_season()

    def _check_season(self):
        if self.season not in SUPPORTED_SEASONS:
            raise ValueError(f'Unknown season `{self.season}`')

    def process(self,
                tile,
                bounds,
                epsg,
                block_id):

        # Set output folder
        output_folder = self.output / 'blocks'

        # Check if we need to load the cropland mask
        if self.cropland_mask is not None:

            cropland_sourcepath = (
                output_folder / tile / '**' /
                self.cropland_mask / 'classification'
                / f'{tile}_{block_id:03d}_classification.tif')
            cropland_files = glob.glob(str(cropland_sourcepath),
                                       recursive=True)

            if len(cropland_files) == 0:
                raise ValueError(('Required cropland mask not found in '
                                  f'`{cropland_sourcepath}`.'))
            with rasterio.open(cropland_files[0], 'r') as src:
                logger.info(f'Loading cropland mask: {cropland_files[0]}')
                window = rasterio.windows.from_bounds(
                    *bounds, transform=src.transform)
                cropland_mask_data = src.read(1, window=window)
        else:
            cropland_mask_data = None

        # Compute the features
        features = self.get_features(tile, bounds, epsg)

        # Put feature data on disk
        import zarr
        logger.info('Saving to zarr on disk')
        z = zarr.open('features.zarr', mode='w',
                      shape=features.data.shape,
                      chunks=(512, 512), dtype=np.float32)
        z[:] = features.data
        logger.info('Switching features data to disk version')
        features.data = z

        # Now do the predictions based on all provided models
        # and write result to disk
        try:

            for modelname, model in self.models.items():

                wcmodel = WorldCerealModel.from_config(model)
                self.classifier = WorldCerealClassifier(
                    wcmodel,
                    filtersettings=self.filtersettings,
                    maskdata=cropland_mask_data)

                # get correct colormap and nodatavalue
                colormap = COLORMAP[self.product.split('_')[-1]]
                nodatavalue = NODATA[self.product.split('_')[-1]]

                prediction, confidence = self.classifier.predict(
                    features,
                    nodatavalue=nodatavalue)

                prediction = prediction.astype(np.uint8)
                confidence = confidence.astype(np.uint8)

                self.save(prediction, tile, bounds, epsg,
                          output_folder, block_id=block_id,
                          modelname=modelname,
                          tag='classification', colormap=colormap,
                          nodata=nodatavalue)

                if self._save_confidence:
                    self.save(confidence, tile, bounds, epsg,
                              output_folder, block_id=block_id,
                              modelname=modelname, tag='confidence',
                              nodata=nodatavalue)

                if self._save_features:
                    self.save_features(features, output_folder, modelname,
                                       tile, block_id, bounds, epsg)

                if self._save_meta:
                    self.save_meta(features, output_folder, modelname,
                                   tile, block_id, bounds, epsg)

                # Clear session to avoid memory leak
                tf.keras.backend.clear_session()
                prediction = None
                confidence = None
                gc.collect()

            if self.active_marker:

                activemarker_ft = 'L2A-evi-nSeas-10m'
                if activemarker_ft not in features.names:
                    raise ValueError(
                        ('The active marker needs the '
                            f'`{activemarker_ft}` feature which is '
                            'not available in the computed features.'))
                nr_seasons = features.select(
                    ['L2A-evi-nSeas-10m']).data[0, ...]
                nr_seasons[nr_seasons > 0] = 1
                nr_seasons[nr_seasons != 1] = 0
                nr_seasons = nr_seasons.astype(np.uint8)

                colormap = COLORMAP['active-cropland']
                nodatavalue = NODATA['active-cropland']

                if self.cropland_mask is not None:
                    nr_seasons = mask(nr_seasons, cropland_mask_data,
                                      maskedvalue=nodatavalue)

                # apply post-classification majority filtering
                if self.filtersettings['kernelsize'] > 0:
                    nr_seasons = majority_filter(
                        nr_seasons,
                        self.filtersettings['kernelsize'],
                        no_data_value=nodatavalue)

                # To correct dtype
                nr_seasons = nr_seasons.astype(np.uint8)

                fname = (output_folder / str(tile) / self.product /
                         'active-cropland' /
                         f"{tile}_{block_id:03d}_active-cropland.tif")
                fname.parent.mkdir(parents=True, exist_ok=True)
                self._to_geotiff(nr_seasons, bounds, epsg, fname,
                                 colormap=colormap,
                                 nodata=nodatavalue)

        except Exception as e:
            logger.error(f"Error predicting crop type for "
                         f"{tile} - {bounds} "
                         f"- {block_id}: {e}")
            raise

        if self.irrigation:

            if self.irrmodels is None:
                raise ValueError('No dedicated irrigation models have been'
                                 ' supplied for the irrigation detection!')

            if self.irrparameters is None:
                raise ValueError(
                    'No dedicated irrigation parameters have '
                    'been provided for irrigation feature computation!')

            logger.info('Preparing settings for irrigation')
            irr_settings = self.irrparameters.get('settings', None)
            if irr_settings is not None:
                # Now make sure start_date and end_date are set in settings
                # of all collections
                start_date = self.settings['l2a']['composite'].get('start')
                end_date = self.settings['l2a']['composite'].get('end')
                for coll in irr_settings.keys():
                    irr_settings[coll]['composite']['start'] = start_date
                    irr_settings[coll]['composite']['end'] = end_date

            irr_features_meta = self.irrparameters.get('features_meta', None)
            irr_rsi_meta = self.irrparameters.get('rsi_meta', None)
            irr_ignore_def_feat = self.irrparameters.get(
                'ignore_def_feat', None)
            irr_collections = {k: v for k, v in self.irrcollections.items()
                               if v is not None}
            self.product = (self.product.split('_')[0] + '_irr-'
                            + self.season)

            logger.info('Computing irr features')
            irrfeatures = self.get_features(
                tile, bounds, epsg,
                settings_override=irr_settings,
                features_meta_override=irr_features_meta,
                rsi_meta_override=irr_rsi_meta,
                ignore_def_feat_override=irr_ignore_def_feat,
                collections_override=irr_collections)

            # Put feature data on disk
            logger.info('Saving irr features to zarr on disk')
            z = zarr.open('irrfeatures.zarr', mode='w',
                          shape=irrfeatures.data.shape,
                          chunks=(512, 512), dtype=np.float32)
            z[:] = irrfeatures.data
            logger.info('Switching features data to disk version')
            irrfeatures.data = z

            # Now do the predictions based on all provided models
            # and write results to disk
            try:

                for modelname, model in self.irrmodels.items():

                    wcmodel = WorldCerealModel.from_config(model)
                    self.classifier = WorldCerealClassifier(
                        wcmodel,
                        filtersettings=self.filtersettings,
                        maskdata=cropland_mask_data)

                    # get correct colormap and nodatavalue
                    colormap = COLORMAP[self.product.split('_')[-1]]
                    nodatavalue = NODATA[self.product.split('_')[-1]]

                    prediction, confidence = self.classifier.predict(
                        irrfeatures,
                        nodatavalue=nodatavalue)

                    prediction = prediction.astype(np.uint8)
                    confidence = confidence.astype(np.uint8)

                    self.save(prediction, tile, bounds, epsg,
                              output_folder, block_id=block_id,
                              modelname=modelname,
                              tag='classification', colormap=colormap,
                              nodata=nodatavalue)

                    if self._save_confidence:
                        self.save(confidence, tile, bounds, epsg,
                                  output_folder, block_id=block_id,
                                  modelname=modelname, tag='confidence',
                                  nodata=nodatavalue)

                    if self._save_features:
                        self.save_features(
                            irrfeatures, output_folder, modelname,
                            tile, block_id, bounds, epsg)

                    if self._save_meta:
                        self.save_meta(irrfeatures, output_folder, modelname,
                                       tile, block_id, bounds, epsg)

                    # Clear session to avoid memory leak
                    tf.keras.backend.clear_session()
                    prediction = None
                    confidence = None
                    gc.collect()

            except Exception as e:
                logger.error(f"Error predicting irrigation for "
                             f"{tile} - {bounds} "
                             f"- {block_id}: {e}")
                raise


class _ChainTimer():

    def __init__(self):

        self.trainingdf = TaskTimer('building training df per sample',
                                    unit='seconds')
        self.trainModel = TaskTimer('model training',
                                    unit='seconds')
        self.evalModel = TaskTimer('model evaluation',
                                   unit='seconds')
