# EWoC ARD specification

The EWoC pre-processing system generates a set of ARD products which are stored into a s3 bucket and used by the classification processor.

## Open questions

- s3 bucket structure organisation / local filesystem organisation
- merge S2 and L8 in the pre processing
- `MSK` or `MASK` for SR and TIR
- TIFF tags or/and metadata.json file
- 30m tif caracteritics
- TIR mask = QA_PIxel ?
- TIR pixel value = quantification?
- Eleaf validation!
- Add production datetime in product/file id

## EWoC ARD S3 bucket structure

The EWoC ARD products are stored in to a s3 bucket with the following structure:

```sh
EWOC_ARD/
├── OPTICAL
│   └── <s2_tile_id>
│       └── <YYYYMMDD>
│           └── <optical_product_id>
│               └── <optical_file_id>.tif
├── SAR
│   └── <s2_tile_id>
│       └── <YYYYMMDD>
│           └── <sar_product_id>
│               └── <sar_file_id>.tif
└── TIR
    └── <s2_tile_id>
        └── <YYYYMMDD>
            └── <tir_product_id>
                └── <tir_file_id>.tif
```

With this structure it is easy to retrieve all the high level products `OPTICAL`, `SAR`, `TIR`, but not to retrieve all the products which correspond to a tile.

Second proposal (possible due to the product_id uniqueness):

```sh
EWOC_ARD/
├── <s2_tile_id>
│   └── <product_id>
│       └── <file_id>.tif
```

with:

- `product_id` = `optical_product_id` or `sar_product_id` or `tir_product_id`
- `file_id` = `optical_file_id` or `sar_file_id` or `tir_file_id` according the `product_id`

## OPTICAL products

The EWoC optical ARD products are generated from Sentinel-2 L2A and Landsat 8 Level 2 Collection 2 Tier 1 products by the ewoc_optical processor. (TBD) This processor merge the Sentinel-2 and Landsat 8 SR data into one product by day. If not we need to replace `OPT` by `S2A` or `S2A` or `L08` and `HHMMSS`. Are we able to merge the L8 and S2 Level 2 by pixel in pre-processing?

The surface reflectances (SR) are provided as GeoTiff per spectral bands:

- B02, B03, B04 and B08 for 10 meters bands
- B11, B12 for 20 meters bands
- B05, B06, B07 for 20 meters bands only if input products contains Sentinel-2 product (TBD: special case if only L8?)

The SR files contains unsigned 16-bit integers with a nodata value (no acquisition) equal to 0 (set in the `GDAL_NODATA` tag). The pixel values need to be divided by 10000 to retrieve the original SR value.

An unified mask is provided to identify usable, unusable and no data pixel. It is provided as GeoTiff:

- at 20 meters
- which contains unsigned 8-bit integers with 0 for usable pixels, 1 for unusable pixels and 255 for no_data (no acquisition) pixels
- no data value is set to 255  in the `GDAL_NODATA` tag.

### Optical naming convention

- `optical_product_id` = `EWOC_ARD_OPT_<s2_tile_id>_<YYYYMMDD>`
- `optical_file_id` = `EWOC_ARD_OPT_<s2_tile_id>_<YYYYMMDD>_<file_type>_<band>.tif`

if second proposal:

- `optical_product_id` = `EWOC_ARD_<s2_tile_id>_OPT_<YYYYMMDD>`
- `optical_file_id` = `EWOC_ARD_<s2_tile_id>_OPT_<YYYYMMDD>_<file_type>_<band>.tif`

With the following keys:

- `file_type` = `SR` or `MSK` (TBD: `MASK`)
- `band` = `B02` or `B03` or `B04` or `B05` or `B06` or `B07` or `B08` or `B11` or `B12` or `None`

### Optical GeoTiff metadata

TODO

`TIFFTAG_IMAGEDESCRIPTION` =  `EWOC ARD SR_band/file_type`

### Annexe 1: Conversion table between spectral bands of Sentinel2, Landsat8 and EWoc ARD

| Sentinel-2<sup>1</sup> | Landsat 8       | WorldCereal | Resolution |
| ---------------------- | --------------- | ----------- | ---------- |
| B02                    | B02             | B02         | 10m        |
| B03                    | B03             | B03         | 10m        |
| B04                    | B04             | B04         | 10m        |
| B05                    | -               | B05         | 20m        |
| B06                    | -               | B06         | 20m        |
| B07                    | -               | B07         | 20m        |
| B08                    | B05<sup>2</sup> | B08         | 10m        |
| B11                    | B06             | B11         | 20m        |
| B12                    | B07             | B12         | 20m        |
| SCL                    | QA_PIXEL        | MASK        | 20m        |

<sup>1</sup>Note that the bands not listed in this table are currently not used by the WorldCereal system.

<sup>2</sup>Note that according to Sen2Like convention, B05 should be matched with B8A. However, we opt to match it with B08, understanding that the calibration will be not as good, but that the 10m resolution of Sentinel-2 B08 is more important.

### Annexe 2: Optical MASK generation

The WorldCereal classification module will make use of a unified binary mask that is generated from either the Sen2Cor scene classifcation (SCL, Sentinel-2) or the FMASK output (QA_PIXEL, Landsat 8). The rules that need to be applied to each mask for binarization are explained in the sections below.

#### Sen2Cor Scene Classification -> Binary MASK

The rules below are valid for **Sentinel-2 L2A products** processed with Sen2Cor V2.9.

```python
# Set the to-be-masked SCL values
SCL_MASK_VALUES = [0, 1, 3, 8, 9, 10, 11]

# Set the nodata value in SCL
SCL_NODATA_VALUE = 0

# Contruct the final binary 0-1-255 mask
mask = np.zeros_like(scl)
mask[scl == SCL_NODATA_VALUE] = 255
mask[~np.isin(scl, SCL_MASK_VALUES)] = 1

```

#### FMASK QA_PIXEL -> Binary MASK

The rules below are valid for **Landsat 8 Level 2 Collection 2**

```python
# Define the nodata 
nodata = qa_pixel == 1      

# Define the to-be-masked qa_pixel values based on the bitmask
cirrus = 1 << 2
cloud = 1 << 3
shadow = 1 << 4
snow = 1 << 5

# Construct the "clear" mask
clear = ((qa_pixel & shadow == 0) &
         (qa_pixel & cloud == 0) &
         (qa_pixel & cirrus == 0) &
         (qa_pixel & snow == 0))

# Contruct the final binary 0-1-255 mask
mask = np.zeros_like(qa_pixel)
mask[clear] = 1
mask[nodata] = 255

```

## SAR products

The EWoC optical ARD products are generated from Sentinel-1 GRD products by the ewoc_s1 processor.

The $\sigma_0$ backscatter (BS) values are provided as GeoTiff per polarization at 20 meters bands for the both polarization VV and VH merged by day. It contains unsigned 16-bit integers in dB with a nodata value (no acquisition) equal to 65535 (set in the `GDAL_NODATA` tag).

### SAR naming convention

- `sar_product_id` = `EWOC_ARD_SAR_<s2_tile_id>_<YYYYMMDD>_<orbit_direction>`
- `sar_file_id` = `EWOC_ARD_SAR_<s2_tile_id>_<YYYYMMDD>_<orbit_direction>_<polarization>.tif`

if second proposal:

- `sar_product_id` = `EWOC_ARD_<s2_tile_id>_SAR_<YYYYMMDD>`
- `sar_file_id` = `EWOC_ARD_<s2_tile_id>_SAR_<YYYYMMDD>_<file_type>_<band>.tif`

With the following keys:

- `orbit_direction` = `ASC` or `DES`
- `polarization` = `VV` or `VH`

### SAR GeoTiff metadata

TODO

`TIFFTAG_IMAGEDESCRIPTION` = `EWOC ARD SIGMA0 VH` or `EWOC ARD SIGMA0 VV`

## TIR products

The EWoC TIR ARD products are generated from Landsat 8 Level 2 Collection 2 Tier 1 product by the ewoc_l8 processor.

The surface temperature (ST) values are provided as GeoTiff at 30 meters and contains unsigned 16-bit integers with nodata value flagged to 0 in the `GDAL_NODATA` tag. The data need to be divided by 10000 to retrieve the original ST (TBC).

A mask is provided as GeoTiff at the same resolution than the ST file. It contains unsigned 16-bit integers values coded as the original Landsat 8 level 2 QA Band file.

### TIR GeoTiff metadata

TODO

`TIFFTAG_IMAGEDESCRIPTION` = `EWOC ARD TIR ST` or `EWOC ARD TIR MASK`

### TIR product naming

- `tir_product_id` = `EWOC_ARD_TIR_<s2_tile_id>_<YYYYMMDD>`
- `tir_file_id` = `EWOC_ARD_TIR_<s2_tile_id>_<YYYYMMDD>_<file_type>.tif`

if second proposal:

- `tir_product_id` = `EWOC_ARD_<s2_tile_id>_TIR_<YYYYMMDD>`
- `tir_file_id` = `EWOC_ARD_<s2_tile_id>_TIR_<YYYYMMDD>_<file_type>_<band>.tif`

With the followings keys:

- `file_type` = `ST`  or `MSK` (TBD: `MASK`)

## Common TIF options required

Common GeoTiff tags:

- `LINEAGE_PRODUCT_ID` = `list of input products id`
- `S2_TILE_ID` = `s2_tile_id`
- `TIFFTAG_DATETIME`= `production datetime in UTC`
- `TIFFTAG_SOFTWARE`= `<ewoc processor name>-<ewoc processor version>`

For **10m** output files:

```sh
blockxsize=1024
blockysize=1024
compress=deflate
tiled=True
```

For **20m** output files:

```sh
blockxsize=512
blockysize=512
compress=deflate
tiled=True
```

For **30m** output files (TIR products):

```sh
TBD
```
