from pathlib import Path
from loguru import logger
from worldcereal.utils.aez import get_matching_aez_id, group_from_id


LC_VALUES = {
    'Unknown': 0,
    'Cropland': 10,
    'Annual cropland': 11,
    'Perennial cropland': 12,
    'Grassland': 13,
    'Herbaceous vegetation': 20,
    'Shrubland': 30,
    'Deciduous forest': 40,
    'Evergreen forest': 41,
    'Mixed/unknown forest': 42,
    'Bare/Sparse vegetation': 50,
    'Builtup/urban': 60,
    'Water': 70,
    'Snow/ice': 80,
    'Non-cropland': 99
}

LC_COLORS = {
    'Unknown': [255/255., 255/255., 255/255.],
    'Cropland': [244/255., 233/255., 12/255.],
    'Annual cropland': [244/255., 178/255., 12/255.],
    'Perennial cropland': [137/255., 100/255., 5/255.],
    'Grassland': [134/255., 229/255., 25/255.],
    'Herbaceous vegetation': [56/255., 237/255., 153/255.],
    'Shrubland': [149/255., 249/255., 234/255.],
    'Deciduous forest': [21/255., 155/255., 17/255.],
    'Evergreen forest': [12/255., 119/255., 8/255.],
    'Mixed/unknown forest': [6/255., 73/255., 4/255.],
    'Bare/Sparse vegetation': [73/255., 56/255., 4/255.],
    'Builtup/urban': [201/255., 201/255., 201/255.],
    'Water': [14/255., 194/255., 234/255.],
    'Snow/ice': [187/255., 238/255., 249/255.],
    'Non-cropland': [163/255., 78/255., 109/255.]
}

LC_MAPPER = {
    0: 0,
    10: 1,
    11: 2,
    12: 3,
    13: 4,
    20: 5,
    30: 6,
    40: 7,
    41: 8,
    42: 9,
    50: 10,
    60: 11,
    70: 12,
    80: 13,
    99: 14
}


def is_real_feature(ft_name):
    '''
    Helper function to distinguish real input features
    from meta-features.
    '''
    if len(ft_name.split('-')) == 4 or 'DEM' in ft_name:
        return True
    else:
        return False


def get_sensor_config(feature_names, sensors=None, pattern=''):
    all_sensors = [x.split('-')[0] for x in feature_names]
    unique_sensors = list(set(all_sensors))

    if sensors is not None:
        if type(sensors) is str:
            sensors = [sensors]
        unique_sensors = [sensor for sensor in unique_sensors
                          if sensor in sensors]

    sensorconfig = {}
    for sensor in unique_sensors:
        sensorconfig[sensor] = dict(
            channels=len([x for x in feature_names if sensor in x
                          and pattern in x and '-ts' not in x]),
            positions=[i for i, ft in enumerate(feature_names)
                       if sensor in ft
                       and pattern in ft and '-ts' not in ft]
        )

    return sensorconfig


def get_sensor_config_timeseries(feature_names):
    all_sensors = [x.split('-')[0] for x in feature_names if '-ts' in x]
    unique_sensors = list(set(all_sensors))

    sensorconfig = {}
    for sensor in unique_sensors:
        allsensorfts = [x for x in feature_names if sensor in x if '-ts' in x]
        sensorchannels = list(set([x.split('-')[1] for x in allsensorfts]))
        sensorconfig[sensor] = {'channels': sensorchannels}

        for channel in sensorchannels:
            channelts = [ft for ft in allsensorfts if channel in ft]
            tsteps = sorted([int(ft.split('-')[2][2:]) for ft in channelts])
            positions = []
            for tstep in tsteps:
                positions += [i for i, ft in enumerate(feature_names)
                              if sensor in ft and channel
                              in ft and f'-ts{tstep}-' in ft]

            sensorconfig[sensor][channel] = dict(
                positions=positions
            )

    return sensorconfig


def check_url(url):
    """Method to check if a URL is valid

    Args:
        url (str): URL to check

    Returns:
        bool: True if URL is valid, False if not
    """

    import urllib

    try:
        status_code = urllib.request.urlopen(url).getcode()
        return status_code == 200
    except urllib.error.HTTPError as e:
        # logger.debug(f'URLError: {e.reason}')
        return False


def _get_best_model_fromurl(parentmodel, aez_id):
    '''Helper function to find locally trained models.
    This is the version that looks on Artifactory!
    '''

    model_baselink = '/'.join(parentmodel.split('/')[:-1])
    aez_group = group_from_id(aez_id)

    groupmodel = f'{model_baselink}/retrained/Group_{aez_group}/config.json'
    zonemodel = f'{model_baselink}/retrained/Zone_{aez_id}/config.json'

    if check_url(zonemodel):
        logger.info(f'Found a zone-specific model: {zonemodel}')
        return str(zonemodel)
    elif check_url(groupmodel):
        logger.info(f'Found a group-specific model: {groupmodel}')
        return str(groupmodel)
    else:
        logger.info(f'Only found a parent model: {parentmodel}')
        return str(parentmodel)
    # return str(parentmodel)


def _get_best_model_fromfile(parentmodel, aez_id):
    '''Helper function to find locally trained models
    '''

    model_basedir = Path(parentmodel).parent
    aez_group = group_from_id(aez_id)

    groupmodel = (model_basedir / 'retrained' /
                  f'Group_{aez_group}' / 'config.json')
    zonemodel = (model_basedir / 'retrained' /
                 f'Zone_{aez_id}' / 'config.json')

    if zonemodel.is_file():
        logger.info(f'Found a zone-specific model: {zonemodel}')
        return str(zonemodel)
    elif groupmodel.is_file():
        logger.info(f'Found a group-specific model: {groupmodel}')
        return str(groupmodel)
    else:
        logger.info(f'Only found a parent model: {parentmodel}')
        return str(parentmodel)
    # return str(parentmodel)


def get_best_model(parentmodel, aez_id):

    if parentmodel.startswith('http'):
        return _get_best_model_fromurl(parentmodel, aez_id)
    else:
        return _get_best_model_fromfile(parentmodel, aez_id)
