import numpy as np
import pandas as pd
from loguru import logger
from satio.collections import BaseCollection


class CollectionError(Exception):
    pass


def drop_unknown(inputs, outputs, nodatavalue=0):
    idx_known = np.where(outputs != nodatavalue)[0]
    return inputs[idx_known, :], outputs[idx_known]


def drop_nan(inputs, outputs=None):
    idx = np.where(np.sum(np.isfinite(inputs), axis=1) == inputs.shape[1])[0]

    if outputs is not None:
        return inputs[idx, :], outputs[idx]
    else:
        return inputs[idx, :]


def convert_to_binary(outputs, target_labels):
    if type(target_labels) is int:
        target_labels = [target_labels]
    outputs_bin = outputs.copy()
    outputs_bin[~np.isin(outputs, target_labels)] = 0
    outputs_bin[np.isin(outputs, target_labels)] = 1
    return outputs_bin


def check_collection(collection: BaseCollection, name,
                     start_date, end_date):
    """Helper function to check collection completeness
    and report gap lengths

    Args:
        collection (BaseCollection): a satio collection
        name (str): name of the collection used for reporting
        start_date (str): processing start date (Y-m-d)
        end_date (str): processing end date (Y-m-d)

    """

    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)

    collstart = collection.df.date.min()
    collend = collection.df.date.max()

    # Check collection start
    gapstart = collstart - start_date
    gapstart = gapstart.days
    if gapstart < 0:
        gapstart = 0

    # Check collection end
    gapend = end_date - collend
    gapend = gapend.days
    if gapend < 0:
        gapend = 0

    # Check max gap
    maxgap = collection.df.date.diff().max().days

    # Report on collection
    logger.info('-' * 50)
    logger.info(f'{name} first image: {collstart}')
    logger.info(f'{name} last image: {collend}')
    logger.info(f'{name} largest gap: {collection.df.date.diff().max()}')

    names = [f'{name}-gap_start', f'{name}-gap_end', f'{name}-gap_max']
    return [gapstart, gapend, maxgap], names
