import glob
from pathlib import Path
from loguru import logger
import tensorflow as tf
import numpy as np
import pandas as pd
import geopandas as gpd
import matplotlib.pyplot as plt
from shapely.geometry import Point

from worldcereal.utils.scalers import minmaxscaler, scale_df
from worldcereal.classification.weights import get_refid_weight
from worldcereal.utils import aez


def get_tfrecord_files(indirs: list,
                       aez_ids: list = None):
    if type(aez_ids) is str or type(aez_ids) is int:
        aez_ids = [aez_ids]

    if type(indirs) is str:
        indirs = [indirs]

    all_tfrecord_files = []

    for indir in indirs:

        if aez_ids is None:
            tfrecord_files = glob.glob(str(Path(indir) / '**' / '*part*.gz'),
                                       recursive=True)
        else:
            tfrecord_files = []
            for aez_id in aez_ids:
                tfrecord_files += glob.glob(
                    str(Path(indir) / str(aez_id) / '*part*.gz'),
                    recursive=True)

        all_tfrecord_files += tfrecord_files

    if len(all_tfrecord_files) == 0:
        raise RuntimeError(('No matching TFrecord files'
                            f' found in `{indirs}`'))

    logger.info(f'TFrecord files found: {len(all_tfrecord_files)}')

    return all_tfrecord_files


def parse_record(example_proto, features, windowsize):
    """Helper function to parse TFrecords based on a recipee

    Args:
        example_proto (TFrecord): one TFrecord entry
        features (list): list of features in the TFrecord
        windowsize (int): patch size

    Returns:
        parsed TFrecord: parsed TFrecord
    """
    parse_features = {}
    for feature in features:
        parse_features[feature] = tf.io.FixedLenFeature(
            shape=[windowsize * windowsize], dtype=tf.int64
        )
    parse_features["location_id"] = tf.io.FixedLenFeature(shape=[],
                                                          dtype=tf.string)
    parse_features["split"] = tf.io.FixedLenFeature(shape=[], dtype=tf.string)
    parse_features["aez_groupid"] = tf.io.FixedLenFeature(shape=[],
                                                          dtype=tf.int64)
    parse_features["aez_zoneid"] = tf.io.FixedLenFeature(shape=[],
                                                         dtype=tf.int64)
    parse_features["ref_id"] = tf.io.FixedLenFeature(shape=[],
                                                     dtype=tf.string)

    example = tf.io.parse_single_example(
        example_proto,
        features=parse_features
    )

    return example


def create_dataset(tfrecordfiles: list, features: list,
                   windowsize=64, batchsize=1):
    """Function to create a TFrecord dataset from TFrecord files

    Args:
        tfrecordfiles (list): list of paths to TFrecord files
        features (list): list of features present in TFrecords
        windowsize (int, optional): patch size. Defaults to 64.
        batchsize (int, optional): batch size. Defaults to 1.

    Returns:
        TFdataset: TFdataset for use in training
    """
    nrFiles = len(tfrecordfiles)

    compression_type = ""
    if tfrecordfiles[0].endswith(".gz"):
        compression_type = "GZIP"

    dataset = tf.data.Dataset.from_tensor_slices(
        tfrecordfiles).shuffle(nrFiles)

    # increase cycle length to read multiple files in parallel
    dataset = dataset.interleave(
        lambda path: (tf.data
                      .TFRecordDataset(path,
                                       compression_type=compression_type)),
        cycle_length=4,
        block_length=1,
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )

    return (
        dataset.map(lambda x: parse_record(x, features, windowsize),
                    num_parallel_calls=nrFiles)
        .repeat()
        .batch(batchsize)
        .prefetch(5)
    )


def convert_inputs(inputbatch, inputfeatures: list, outputname: str,
                   targetlabels: list, scalefactor: int = 10000,
                   ignorelabels: list = None, focuslabels: list = None,
                   targetmultiplier: int = 1):
    """
    Function to convert raw TFrecords to training-ready numpy arrays
    Here we implement pixel-based loss weighting as well.

    Args:
        inputbatch (tfrecords): batch of TFrecords
        inputfeatures (list): list of features to extract from raw TFrecord
        outputname (str): name of the output label to extract
        targetlabels (list): list of target label values that are positives
        scalefactor (int, optional): scale factor to apply. Defaults to 10000.
        ignorelabels (list, optional): list of target label values to ignore
        focuslabels (list, optional): list of target label values
                                      to double weights
        targetmultiplier (int, optional): multiplier for weights of
                          target class (defaults to 1)

    Returns:
        tuple: numpy arrays of (inputs, outputs, weights)
    """

    inputs = []
    for feature in inputfeatures:
        inputs.append(
            minmaxscaler(
                np.expand_dims(
                    inputbatch[
                        feature].numpy().astype(
                            float), axis=2) / scalefactor,
                ft_name=feature)
        )

    inputs = np.concatenate(tuple(inputs), axis=2)
    outputs = np.expand_dims(
        inputbatch[outputname].numpy().astype(int) / scalefactor,
        axis=2)
    weights = np.ones_like(outputs)

    # First set weights to zero if we don't know the label
    weights[outputs == 0] = 0

    # Put weights of ignorelabels also to zero
    if ignorelabels is not None:
        for ignore in ignorelabels:
            weights[outputs == ignore] = 0

    # Multiply the weights by the ref_id weighting factor
    ref_ids = inputbatch['ref_id'].numpy().astype(str)
    ref_id_weights = [get_refid_weight(ref_id) for ref_id in ref_ids]
    ref_id_weights = np.expand_dims(ref_id_weights, axis=(1, 2))
    ref_id_weights = np.repeat(ref_id_weights, outputs.shape[1], axis=1)
    weights = np.multiply(weights, ref_id_weights)

    # Double the weights of the focus labels
    if focuslabels is not None:
        for focuslabel in focuslabels:
            weights[outputs == focuslabel] *= 2

    # Now make the classification binary
    for t in targetlabels:
        outputs[outputs == t] = 1  # cropland
    outputs[outputs != 1] = 0  # no cropland

    # Finally apply multiplier to the weights of the target
    weights[outputs == 1] *= targetmultiplier

    # Get rid of any NaN values in the inputs
    # to avoid breaking the neural net
    inputs[np.isnan(inputs)] = 0

    return inputs, outputs, weights.squeeze()


def data_generator(dataset_iterator, inputfeatures, outputname,
                   targetlabels, ignorelabels=None, focuslabels=None,
                   targetmultiplier=1, windowsize=64,
                   requestedwindowsize=None, batchsize=16,
                   clamp=(None, None), forcebinaryweights=False,
                   forcepositives=False,
                   samplepatience=1000,
                   ):
    """Function that acts as data generator for keras model training

    Args:
        dataset_iterator (iterator): a TFdataset iterator object
        inputfeatures (list): list of input features to parse
        outputname (str): name of the output label in the TFdataset
        targetlabels (list): list of target label values that are positives
        ignorelabels (list, optional): list of target label values to ignore
        focuslabels (list, optional): list of target label values
                                      to double weights
        targetmultiplier (int, optional): multiplier for weights of
                          target class (defaults to 1)
        windowsize (int, optional): size of the patch. Defaults to 64.
        requestedwindowsize (int, optional): optional random subset to cut out
        batchsize (int, optional): batch size. Defaults to 16.
        clamp (tuple, optional): min/max clamp values to apply on inputs
        forcebinaryweights (bool, optional): if True, force weights to be [0,1]
        forcepositives (bool, optional): if True, a sample must include
                  at least one positive pixel
        samplepatience (int, optional): how many samples we try if
                  forcepositives is True

    Yields:
        tuple: numpy arrays (inputs, outputs, weights))
    """

    requestedwindowsize = requestedwindowsize or windowsize

    while True:

        inputs = np.empty((batchsize, windowsize * windowsize,
                           len(inputfeatures)))
        outputs = np.empty((batchsize, windowsize * windowsize, 1))
        weights = np.empty((batchsize, windowsize * windowsize))

        for i in range(batchsize):

            if not forcepositives:
                # Get a sample
                sample = next(dataset_iterator)
            else:
                attempts = 0
                while True:

                    attempts += 1

                    if attempts > samplepatience:
                        raise RuntimeError(('Could not find an appropriate '
                                            f'sample after {samplepatience} '
                                            'attempt(s).'))

                    # Get a sample
                    sample = next(dataset_iterator)

                    # Check if we have a positive pixel
                    uniquelabels = np.unique(
                        (sample['CT'].numpy().ravel()/10000).astype(int))

                    if np.in1d(uniquelabels,
                               targetlabels,
                               assume_unique=True).any():

                        logger.debug((f'Good sample after {attempts} '
                                      'attempt(s) ...'))
                        break

                    else:
                        continue

            # Convert the data
            inputs[i, ...], outputs[i, ...], weights[i, ...] = convert_inputs(
                sample,
                inputfeatures,
                outputname,
                targetlabels,
                ignorelabels=ignorelabels,
                focuslabels=focuslabels,
                targetmultiplier=targetmultiplier)

            if forcebinaryweights:
                weights[weights > 1] = 1

            # Check if we need to clamp
            minclamp, maxclamp = clamp
            if minclamp is not None:
                inputs[inputs < minclamp] = minclamp
            if maxclamp is not None:
                inputs[inputs > maxclamp] = maxclamp

        if requestedwindowsize == windowsize:
            # Return for training
            yield inputs, outputs, weights

        elif requestedwindowsize > windowsize:
            raise ValueError(('`requestedwindowsize` cannot be larger '
                              'than `windowsize.`'))

        else:
            # Need to cut out a random part
            origindex = np.arange(windowsize ** 2).reshape(
                (windowsize, windowsize)
            )
            # Sample a random window
            randomstartX = np.random.randint(
                low=0,
                high=windowsize-requestedwindowsize)
            randomstartY = np.random.randint(
                low=0,
                high=windowsize-requestedwindowsize)
            sampledindex = origindex[
                randomstartX: randomstartX+requestedwindowsize,
                randomstartY: randomstartY+requestedwindowsize]
            sampledindex = sampledindex.ravel()

            # Return subsample for training
            yield (inputs[:, sampledindex, :],
                   outputs[:, sampledindex, :],
                   weights[:, sampledindex])


def get_sample_weight(sample, options):

    # Default weight
    if sample['OUTPUT'] == 0:
        weight = 0
    elif sample['OUTPUT'] in options.get('ignorelabels', []):
        weight = 0
    else:
        weight = 1

    # Multiply the weights by the ref_id weighting factor
    ref_id = sample['ref_id']
    ref_id_weight = get_refid_weight(ref_id, pixel=True)
    weight *= ref_id_weight

    # Double the weights of the focus labels
    if sample['OUTPUT'] in options.get('focuslabels', []):
        weight *= 2

    # Finally apply multiplier to the weights of the target
    if sample['OUTPUT'] in options.get('targetlabels', []):
        weight *= options.get('targetmultiplier', 1)

    return weight


def select_within_aez(df, aez_id, aez_group, buffer):
    '''
    Subset the dataframes
    We use the provided buffer for this and select all samples within
    the buffered polygon(s)
    '''

    # Create points from lat/lon in DF
    points = [Point(x) for x in zip(df['lon'], df['lat'])]

    # Make gdf from df
    df = gpd.GeoDataFrame(df, geometry=points, crs='epsg:4326')

    # Load AEZ
    aez_df = aez.load()

    # Subset on group or zone
    if aez_id is not None:
        aez_df = aez_df[aez_df['zoneID'] == aez_id]
    if aez_group is not None:
        aez_df = aez_df[aez_df['groupID'] == aez_group]

    # Check how many samples are truly in AEZ for reporting
    logger.info('Intersecting samples with original AEZ ...')
    nr_within = df.within(aez_df.unary_union).sum()

    # Buffer the zone/group and reproject back to WGS84
    logger.info(f'Buffering selected AEZs with buffer: {buffer}')
    aez_df['geometry'] = aez_df.to_crs(epsg=3857).buffer(
        buffer).to_crs(epsg=4326)

    # Finally intersect df with the buffered AEZ
    logger.info('Intersecting samples with buffered AEZ ...')
    df = df[df.within(aez_df.unary_union)]

    # Report on selected samples
    nr_surrounding = df.shape[0] - nr_within
    logger.info(f'{nr_within + nr_surrounding} samples selected '
                f'({nr_within} inside AEZ and {nr_surrounding} '
                'within the buffer)')

    return df


def do_smote(season, inputs, outputs, minsamples, figdir, options):
    '''Method to run SMOTE on inputs/outputs
    '''
    from imblearn.over_sampling import SMOTE

    # Classes that we don't want to oversample
    # by default
    if season == 'annual' or season == 'cropland':
        nonoversample_classes = [
            10,   # Cropland, not specified
            12,   # Perennial cropland
            40,   # Deciduous forest
            41,   # Evergreen forest
            42,   # Mixed/unknown forest
            60,   # Builtup/urban
            70,   # Water
            80,   # Snow/ice
            99,   # other non-crop (lpis)
        ]
    else:
        nonoversample_classes = []

    logger.info('Initializing SMOTE on samples...')
    counts = outputs.value_counts()

    if counts.sum() < minsamples:
        errormessage = (f'Got less than {minsamples} '
                        f'samples for this dataset. '
                        'Cannot continue!')
        logger.error(errormessage)
        raise NotEnoughSamplesError(errormessage)

    # SMOTE needs at least 6 samples for each class
    class_too_few = counts[counts < 6].index.tolist()

    # We add the classes that we don't want to oversample
    exclude_classes = list(set(class_too_few + nonoversample_classes))

    # Those samples will not be
    # oversampled, but we do keep them
    # for training
    new_inputs = inputs[outputs.isin(exclude_classes)]
    new_outputs = outputs[outputs.isin(exclude_classes)]

    # Get the samples that can participate in SMOTE
    idx_smote = np.where((~outputs.isin(exclude_classes)) & (
        np.sum(np.isnan(inputs), axis=1) == 0
    ))[0]

    # Get the majority class label and its sample size
    majority_class = outputs.value_counts().index[0]
    nr_samples_maj = outputs.value_counts().values[0]

    # Setup over-sampling strategy
    smote_dict = dict()
    oversampling_fraction = 1
    for current_class in counts[~counts.index.isin(
            exclude_classes)].index.tolist():
        if current_class == majority_class:
            # We keep the majority class as is
            smote_dict[current_class] = nr_samples_maj
        else:
            if counts[current_class] > int(
                    oversampling_fraction * nr_samples_maj):
                # If we have more then we keep them all
                smote_dict[current_class] = counts[current_class]
            else:
                # If we have less then we oversample
                smote_dict[current_class] = int(
                    oversampling_fraction * nr_samples_maj)

    # Initialize SMOTE with sampling strategy
    oversample = SMOTE(sampling_strategy=smote_dict)

    # Perform SMOTE
    logger.info('Running SMOTE ...')
    smote_inputs, smote_outputs = oversample.fit_resample(
        inputs[idx_smote, :],
        outputs.values[idx_smote])

    # Concatenate with the other samples
    logger.info('Concatenating enriched samples ...')
    inputs = np.concatenate([new_inputs, smote_inputs], axis=0)
    outputs = np.concatenate([new_outputs, smote_outputs], axis=0)

    if figdir is not None:
        # Plot histogram of SMOTE outputs
        labels, counts = np.unique(outputs, return_counts=True)
        plt.bar(range(len(labels)), counts)
        plt.xticks(range(len(labels)), labels.astype(int))
        plt.xlabel('Class')
        plt.ylabel('Amounts')
        plt.title('Output label distribution after SMOTE')
        outfile = Path(figdir) / 'output_distribution_SMOTE.png'
        plt.savefig(outfile)
        plt.close()

    origoutputs = np.copy(outputs)  # For use in evaluation

    # Make classification binary
    outputs[np.in1d(outputs, options['targetlabels'])] = 1
    outputs[outputs != 1] = 0

    # Set all weights to 1
    logger.warning(('When using SMOTE, all resulting '
                    'samples get equal weights!'))
    weights = np.ones_like(outputs)

    return inputs, outputs, weights, origoutputs


def get_trainingdata(df, inputfeatures, season, options, aez_id=None,
                     aez_group=None, minsamples=500, smote=False,
                     logdir=None, worldcover=False, outliers=False,
                     outlierinputs=None, buffer=500000):
    """Function that returns inputs/outputs from training DataFrame

    Args:
        df (pd.DataFrame): dataframe containing input/output features
        inputfeatures (list[str]): list of inputfeatures to be used
        season (str): season identifier
        options (dict): dictionary containing options
        aez_id (int, optional): ID of the AEZ to subset on. Defaults to None.
        aez_group (int, optional): AEZ group ID to subset on. Defaults to None.
        minsamples (int, optional): minimum number of samples. Defaults to 500.
        smote (bool, optional): whether or not to perform SMOTE. Defaults to
                                False.
        buffer (int, optional): the buffer (in m) to take around AEZs before
            selecting matching location_ids. Defaults to 500000 (500km)
        logdir (str, optional): output path for logs/figures. Defaults to None.
        worldcover (bool, optional): whether or not to remove outliers based on
            worldcover data. Defaults to False.
        outliers (bool, optional): whether or not to remove
            outliers from the class
            of interest based on KNN (pyod implementation). Defaults to False.
        outlierinputs (list, optional): in case outliers need to be removed,
            this list specifies which input variables need to be used here.
            Defaults to None.

    Raises:
        NotEnoughSamplesError: obviously when not enough samples were found

    Returns:
        inputs, outputs, weights: arrays to use in training
    """

    # Check if we have enough samples at all to start with
    # that's at least 2X minsamples (binary classification)
    if df.shape[0] < 2 * minsamples:
        errormessage = (f'Got less than {2 * minsamples} '
                        f'in total for this dataset. '
                        'Cannot continue!')
        logger.error(errormessage)
        raise NotEnoughSamplesError(errormessage)

    # Input clamping
    minclamp = -0.5  # Do not allow inputs below this value (None to disable)
    maxclamp = 1.5  # Do not allow inputs above this value (None to disable)

    if aez_id is not None and aez_group is not None:
        raise ValueError('Cannot set both `aez_id` and `aez_group`')

    # Select samples within (buffered) AEZ
    if aez_id is not None or aez_group is not None:
        df = select_within_aez(df, aez_id, aez_group, buffer)

    # If not looking at cropland
    # need to get rid of non-annual-cropland samples
    if season != 'annual' and season != 'cropland':
        logger.info('Removing samples that are not annual cropland.')
        df = df[df['LC'] == 11].copy()

    # Remove the unknown and ignore labels
    remove_idx = ((df['OUTPUT'] == 0) |
                  (df['OUTPUT'].isin(options['ignorelabels'])))
    df = df[~remove_idx].copy()
    logger.info(f'Removed {remove_idx.sum()} unknown/ignore samples.')

    # Get rid of urban/bare samples according to worldcover
    if worldcover:
        agri_labels = [10, 11, 12]
        ewoc_ignore = [50, 60]
        if (season == 'annual') or (season == 'cropland'):
            remove_idx = ((df['OUTPUT'].isin(agri_labels)) &
                          (df['WORLDCOVER-LABEL-10m'].isin(ewoc_ignore)))
        else:  # crop type case
            remove_idx = df['WORLDCOVER-LABEL-10m'].isin(ewoc_ignore)

        df = df[~remove_idx].copy()
        logger.info(f'Removed {remove_idx.sum()} bare/urban samples'
                    f' according to worldcover information.')

    # get rid of outliers (only for class of interest)
    if outliers:
        if outlierinputs is None:
            raise ValueError('No outlier inputs provided!')
        from pyod.models.knn import KNN

        # split the dataset and run outlier removal on only one part
        target_idx = df['OUTPUT'].isin(options['targetlabels'])
        nremoved = 0
        df1 = df[target_idx].copy()
        dfs = []
        if df1.shape[0] > 0:
            for label in options['targetlabels']:
                dftoclean = df1[df1['OUTPUT'] == label]
                if dftoclean.shape[0] < 6:
                    # Not enough samples for this label: skip
                    continue
                # get the variables used for outlier removal
                outlier_x = dftoclean[outlierinputs].values
                # Get rid of any existing NaN values
                outlier_x[np.isnan(outlier_x)] = 0
                # fit the model
                clf = KNN()
                clf.fit(outlier_x)
                # get the prediction labels
                y_train_pred = clf.labels_  # binary labels (0: inliers, 1: outliers)  # NOQA
                nremoved += y_train_pred.sum()
                retain_idx = np.where(y_train_pred == 0)
                dfclean = dftoclean.iloc[retain_idx]
                dfs.append(dfclean)
            # merge all cleaned dataframes
            if len(dfs) > 0:
                dfclean = pd.concat(dfs)
                # merge with other classes
                df0 = df[~target_idx].copy()
                df = pd.concat([df0, dfclean])
        logger.info(f'Removed {nremoved} outliers')

    # Scale the input data
    df[inputfeatures] = scale_df(df[inputfeatures],
                                 clamp=(minclamp, maxclamp),
                                 nodata=0,
                                 )

    # Get the inputs
    inputs = df[inputfeatures].values

    # Get the weights
    weights = df.apply(lambda row: get_sample_weight(
        row, options), axis=1).values

    # Log the number of samples still present
    # per ref_id
    if logdir is not None:
        counts = (df.groupby('ref_id')[
            'OUTPUT'].value_counts().unstack().fillna(0).astype(int))
        counts.to_csv(Path(logdir) / 'sample_counts.csv')

    # Get the outputs AFTER the weights
    outputs = df['OUTPUT'].copy()

    if logdir is not None:
        # Plot histogram of original outputs
        counts = outputs.value_counts()
        labels = counts.index.astype(int)
        plt.bar(range(len(labels)), counts.values)
        plt.xticks(range(len(labels)), labels)
        plt.xlabel('Class')
        plt.ylabel('Amounts')
        plt.title('Output label distribution')
        outfile = Path(logdir) / 'output_distribution.png'
        outfile.parent.mkdir(exist_ok=True)
        plt.savefig(outfile)
        plt.close()

    # Optionally perform SMOTE
    if smote:
        # if season != 'annual' and season != 'cropland':
        #     raise RuntimeError(('SMOTE oversampling not '
        #                         f'supported for season `{season}`'))

        # Run SMOTE
        inputs, outputs, weights, origoutputs = do_smote(
            season, inputs, outputs, minsamples, logdir, options
        )

    else:

        origoutputs = np.copy(outputs.values)  # For use in evaluation

        # Make classification binary
        outputs[outputs.isin(options['targetlabels'])] = 1
        outputs[outputs != 1] = 0
        outputs = outputs.values

    idx = np.where(np.sum(inputs, axis=1) == 0)
    logger.info(f'#Rows with all-zero inputs: {len(idx[0])}')

    # Make sure all NaNs are gone!
    if np.sum(np.isnan(inputs)) > 0:
        logger.warning(f'Removing {np.sum(np.isnan(inputs))} NaN values!')
    inputs[np.isnan(inputs)] = 0

    # Now check if we have enough samples to proceed
    for label in [0, 1]:
        if np.sum(outputs == label) < minsamples:
            errormessage = (f'Got less than {minsamples} '
                            f'`{label}` samples for this dataset. '
                            'Cannot continue!')
            logger.error(errormessage)
            raise NotEnoughSamplesError(errormessage)

    return (inputs, outputs, weights, origoutputs)


def get_pixel_data(season, options, inputfeatures,
                   aez_id=None, aez_group=None, buffer=500000,
                   logdir=None, outlierinputs=None, minsamples=500):
    cal_df = (options['trainingfiles_cal'] if
              type(options['trainingfiles_cal'])
              is list else [options['trainingfiles_cal']])
    val_df = (options['trainingfiles_val'] if
              type(options['trainingfiles_val'])
              is list else [options['trainingfiles_val']])
    test_df = (options['trainingfiles_test'] if
               type(options['trainingfiles_test'])
               is list else [options['trainingfiles_test']])

    cal_df_data = pd.DataFrame()
    val_df_data = pd.DataFrame()
    test_df_data = pd.DataFrame()

    for current_cal_df in cal_df:
        cal_df_data = pd.concat(
            [cal_df_data,
             pd.read_parquet(
                 (Path(current_cal_df) /
                  f'training_df_{options["outputlabel"]}.parquet'))])
    for current_val_df in val_df:
        val_df_data = pd.concat(
            [val_df_data,
             pd.read_parquet(
                 (Path(current_val_df) /
                  f'training_df_{options["outputlabel"]}.parquet'))])
    for current_test_df in test_df:
        test_df_data = pd.concat(
            [test_df_data,
             pd.read_parquet(
                 (Path(current_test_df) /
                  f'training_df_{options["outputlabel"]}.parquet'))])

    cal_data = get_trainingdata(cal_df_data, inputfeatures, season,
                                options,
                                aez_id=aez_id, aez_group=aez_group,
                                smote=options.get('smote', False),
                                logdir=logdir,
                                worldcover=options.get('worldcover', False),
                                outliers=options.get('outliers', False),
                                outlierinputs=outlierinputs,
                                buffer=buffer,
                                minsamples=minsamples)
    val_data = get_trainingdata(val_df_data, inputfeatures, season,
                                options,
                                aez_id=aez_id, aez_group=aez_group,
                                worldcover=options.get('worldcover', False),
                                outliers=options.get('outliers', False),
                                outlierinputs=outlierinputs,
                                buffer=buffer,
                                minsamples=minsamples)
    test_data = get_trainingdata(test_df_data, inputfeatures, season,
                                 options,
                                 aez_id=aez_id, aez_group=aez_group,
                                 worldcover=options.get('worldcover', False),
                                 outliers=options.get('outliers', False),
                                 outlierinputs=outlierinputs,
                                 buffer=buffer,
                                 minsamples=minsamples)

    logger.info(f'Training on {cal_data[0].shape[0]} samples.')
    logger.info(f'Validating on {val_data[0].shape[0]} samples.')
    logger.info(f'Testing on {test_data[0].shape[0]} samples.')

    return cal_data, val_data, test_data


class NotEnoughSamplesError(Exception):
    pass
