from typing import List

from loguru import logger
from pathlib import Path
import json

from satio.grid import get_blocks_gdf, buffer_bounds
from satio import layers
from satio.collections import (AgERA5Collection,
                               DEMCollection)
from satio.utils.logs import (exitlogs, proclogs,
                              memlogs, ExitLogsMonitor)

from worldcereal.seasons import get_processing_dates, NoSeasonError
from worldcereal.utils.aez import get_matching_aez_id, group_from_id
from worldcereal.utils import get_best_model
from worldcereal.fp import (WorldCerealAgERA5FeaturesProcessor,
                            WorldCerealOpticalFeaturesProcessor,
                            WorldCerealSIGMA0FeaturesProcessor,
                            WorldCerealThermalFeaturesProcessor)
from worldcereal.processors import (ClassificationProcessor,
                                    CropTypeProcessor)
from worldcereal.collections import (WorldCerealOpticalTiledCollection,
                                     WorldCerealSigma0TiledCollection,
                                     WorldCerealThermalTiledCollection)
from worldcereal.features.settings import (get_cropland_tsteps_parameters,
                                           get_croptype_tsteps_parameters,
                                           get_irr_parameters)
from worldcereal.postprocess import postprocess_tile

S2_GRID = layers.load('s2grid')

FEATURESETTINGS_LUT = {
    'cropland': get_cropland_tsteps_parameters(),
    'croptype': get_croptype_tsteps_parameters()
}


def clip_bounds(bounds, clipsize):
    '''Cut out a part of the block based on the clipsize.
    clipsize = resulting size of the block (in meters)
    '''
    bounds = [bounds[0], bounds[1], bounds[0] + clipsize,
              bounds[1] + clipsize]
    return bounds


def get_processing_blocks(tiles, parameters, debug=False,
                          blocks: List = None):
    '''Construct elementary processing blocks dataframe

    '''

    # If not blocks subset is provided, take by default all
    if type(blocks) == int:
        blocks = [blocks]
    blocks = blocks or list(range(121))

    # Get the processing blocks
    processingblocks = get_blocks_gdf(tiles, s2grid=S2_GRID)

    # Subset on the desired blocks
    processingblocks = processingblocks.loc[blocks]

    # Add the parameters to each block
    # this is some onconventional Pandas stuff
    processingblocks['parameters'] = processingblocks.apply(
        lambda x: parameters, axis=1)

    if debug:
        # reduce to one block per tile
        processingblocks = processingblocks.iloc[0:1, :]

        # # reduce blocks extent for testing
        # processingblocks["bounds"] = processingblocks["bounds"].apply(
        #     lambda x: clip_bounds(x, 3000))

    return processingblocks


def parse_featuresettings(featuresettings):
    if featuresettings not in FEATURESETTINGS_LUT.keys():
        raise ValueError('Unknown featuresettings describer: '
                         f'`{featuresettings}`')

    return FEATURESETTINGS_LUT[featuresettings]


def get_worldcereal_collections(inputpaths):
    """Function to create WorldCereal satio-based collections
    from folders (TODO: add S3 bucket option)

    Args:
        inputpaths (dict): dictionary with keys the different
        input sources and values the root path to the input data

    Returns:
        (dict[collection], dict[fp]): the resulting collections
        and features processors to be used.
    """

    # ----------------------------
    # Create the collections
    # ----------------------------
    logger.info('Initializing collections ...')
    S2coll = WorldCerealOpticalTiledCollection.from_folder(
        inputpaths['OPTICAL'])
    S1coll = WorldCerealSigma0TiledCollection.from_folder(
        inputpaths['SAR'])
    L8coll = WorldCerealThermalTiledCollection.from_folder(
        inputpaths['THERMAL'])
    demcoll = DEMCollection(
        folder=inputpaths['DEM'])
    AgERA5coll = AgERA5Collection.from_path(
        inputpaths['METEO'])

    collections = {
        'S2collection': S2coll,
        'S1collection': S1coll,
        'L8collection': L8coll,
        'AgERA5collection': AgERA5coll,
        'DEMcollection': demcoll
    }

    fps = {
        'L2A': WorldCerealOpticalFeaturesProcessor,
        'SIGMA0': WorldCerealSIGMA0FeaturesProcessor,
        'AgERA5': WorldCerealAgERA5FeaturesProcessor,
        'L8': WorldCerealThermalFeaturesProcessor
    }

    return collections, fps


def get_best_irr_model(parentmodel, aez_id):
    '''Helper function to find locally trained irr models
    '''
    modelname = Path(parentmodel).parent.stem
    model_basedir = Path(parentmodel).parents[2]
    aez_group = group_from_id(aez_id)

    groupmodel = (model_basedir / f'Group_{aez_group}' /
                  str(modelname) / 'config.json')
    zonemodel = (model_basedir / f'Zone_{aez_id}' /
                 str(modelname) / 'config.json')

    if zonemodel.is_file():
        logger.info(f'Found a zone-specific model: {zonemodel}')
        return str(zonemodel)
    elif groupmodel.is_file():
        logger.info(f'Found a group-specific model: {groupmodel}')
        return str(groupmodel)
    else:
        logger.info(f'Only found a parent model: {parentmodel}')
        return str(parentmodel)


def _run_block(output_folder, processing_tuple):
    '''Main WorldCereal products pipeline for 1 processing block

    '''
    block_id, block = processing_tuple

    # Get the processing parameters
    parameters = block.parameters
    tile = block.tile
    geometry = block.geometry
    product = parameters['product']
    year = parameters['year']
    season = parameters['season']
    models = parameters['models']
    use_local_models = parameters.get('localmodels', True)
    featuresettings = parameters['featuresettings']
    filtersettings = parameters['filtersettings']
    save_confidence = parameters.get('save_confidence', True)
    save_meta = parameters.get('save_meta', True)
    save_features = parameters.get('save_features', False)
    active_marker = parameters.get('active_marker', None)
    cropland_mask = parameters.get('cropland_mask', None)
    irrigation = parameters.get('irrigation', False)
    irrparameters = parameters.get('irrparameters', None)
    irrmodels = parameters.get('irrmodels', None)

    logger.info('-'*50)
    logger.info(f'Starting processing block: {block_id}')
    logger.info('-'*50)
    logger.info('PARAMETERS:')
    for parameter, value in parameters.items():
        logger.info(f'{parameter}: {value}')

    # Get AEZ ID
    # aez_id = get_matching_aez_id(geometry)
    aez_id = get_matching_aez_id(
        S2_GRID[S2_GRID.tile == tile].geometry.values[0])

    # Get processing dates
    try:
        start_date, end_date = get_processing_dates(season, aez_id, year)
    except NoSeasonError:
        logger.warning(f'No valid `{season}` season found for this block.')
        return

    # Get collections
    collections, fps = get_worldcereal_collections(parameters['inputs'])

    # Get best model
    for model in models.keys():
        if use_local_models:
            models[model] = get_best_model(models[model], aez_id)

    if irrmodels is not None:
        for model in irrmodels.keys():
            if use_local_models:
                irrmodels[model] = get_best_irr_model(irrmodels[model], aez_id)

    # Initialize processor
    if season == 'annual':
        processor = ClassificationProcessor
    else:
        processor = CropTypeProcessor

    # Initialize processing chain
    chain = processor(
        output_folder,
        product=product,
        models=models,
        season=season,
        collections=dict(
            L2A=collections['S2collection'],
            SIGMA0=collections['S1collection'],
            AgERA5=collections['AgERA5collection'],
            DEM=collections['DEMcollection'],
        ),
        settings=featuresettings['settings'],
        features_meta=featuresettings['features_meta'],
        rsi_meta=featuresettings['rsi_meta'],
        ignore_def_feat=featuresettings['ignore_def_feat'],
        gdd_normalization=featuresettings['gddnormalization'],
        fps=fps,
        start_date=start_date.strip('-'),
        end_date=end_date.strip('-'),
        active_marker=active_marker,
        cropland_mask=cropland_mask,
        save_features=save_features,
        save_meta=save_meta,
        save_confidence=save_confidence,
        filtersettings=filtersettings,
        irrcollections=dict(
            L2A=collections['S2collection'],
            AgERA5=collections['AgERA5collection'],
            DEM=collections['DEMcollection'],
            L8=collections['L8collection']
        ),
        irrigation=irrigation,
        irrparameters=irrparameters,
        irrmodels=irrmodels)

    # Run pipeline
    chain.process(block.tile, block.bounds,
                  block.epsg, block.block_id)


def run_tile(tile: str,
             configfile: str,
             outputfolder: str,
             blocks: List = None,
             skip_processed: bool = True,
             debug: bool = False, sparkcontext=None):
    """Generates WorldCereal products.

    Args:
        tile (str): MGRS tile ID to process. Example: '31UFS'
        configfile (str): path to config.json containing processing settings
        outputfolder (str): path to use for saving products and logs
        blocks (List, optional): Block ids of the blocksto process
                from the given tile. Should be a sequence of integers
                between 0 and 120 included. If not provided, all blocks
                will be processed.
        skip_processed (bool, optional): Skip already processed blocks
                by checking the existlogs folder. Defaults to True.
        debug (bool, optional): Run in debug mode, processing only
                one part of one block. Defaults to False.
        sparkcontext (optional): Optional sparkcontext to parallellize
                block processing using spark.
    """

    # Load the config file
    if not Path(configfile).is_file():
        raise FileNotFoundError(
            'Required config file '
            f'`{configfile}` not found. Cannot continue.')
    config = json.load(open(configfile, 'r'))

    # Get processing parameters
    parameters = config['parameters']

    # Get the right feature settings
    parameters['featuresettings'] = parse_featuresettings(
        parameters['featuresettings'])

    # Add models and inputs as parameters
    parameters['models'] = config['models']
    parameters['inputs'] = config['inputs']

    # ----------------------------
    # Get processing blocks
    # ----------------------------
    blocks = get_processing_blocks([tile], parameters, debug, blocks=blocks)

    # ----------------------------
    # Create processing tuples
    # ----------------------------
    logger.info('Getting processing tuples ...')
    processing_tuples = [(f'{row.tile}_{row.block_id:03d}', row)
                         for row in blocks.itertuples()]

    # Setup logging folders
    outputfolder = Path(outputfolder)
    year = parameters['year']
    product = parameters['product']
    exitlogs_folder = outputfolder / 'exitlogs' / f'{year}_{product}'
    proclogs_folder = outputfolder / 'proclogs' / f'{year}_{product}'
    memlogs_folder = outputfolder / 'memlogs' / f'{year}_{product}'

    @proclogs(proclogs_folder, level='DEBUG')
    @exitlogs(exitlogs_folder, skip_processed=skip_processed)
    @memlogs(memlogs_folder)
    def _log_run_block(processing_tuple):
        _run_block(outputfolder, processing_tuple)

    exitlogs_monitor = ExitLogsMonitor(exitlogs_folder,
                                       interval=300)
    exitlogs_monitor.start()

    if sparkcontext is not None:
        logger.info('Starting spark parallellization ...')
        sparkcontext.parallelize(
            processing_tuples,
            len(processing_tuples)).foreach(_log_run_block)

    else:
        logger.info('Running in serial ...')
        # TODO: allow processing in a few threads
        for tup in processing_tuples:
            _log_run_block(tup)

    exitlogs_monitor.task()
    exitlogs_monitor.stop()

    # ----------------------------
    # Do post-processing
    # ----------------------------

    logger.info('Start post-processing to COGs...')
    cog_folder = Path(outputfolder) / 'cogs'

    for modelname in config['models'].keys():
        logger.info(f'Working on model: {modelname}')

        postprocess_tile(outputfolder, cog_folder, product,
                         tile, year, modelname,
                         skip_processed=skip_processed,
                         in_memory=False)

    # post-process irrigation products
    if parameters.get('irrigation', False):
        for modelname in parameters['irrmodels'].keys():
            logger.info(f'Working on model: {modelname}')

            postprocess_tile(outputfolder, cog_folder,
                             'irr-' + product,
                             tile, year, modelname,
                             skip_processed=skip_processed,
                             in_memory=False)

    logger.success('Finished!')


if __name__ == '__main__':

    # -----------------
    tile = '30STF'
    blocks = [1]
    configfile = '/data/worldcereal/run_config.json'
    outputfolder = '/data/worldcereal/runs/test'

    debug = True
    spark = False
    skip_processed = True
    # ------------------

    # Initialize SC
    if spark:
        from worldcereal.utils.spark import get_spark_context
        sc = get_spark_context(pythonpath='/data/users/Public/jeroendegerickx/'
                               'conda_env/worldcereal/bin/python')
    else:
        sc = None

    run_tile(tile, configfile, outputfolder,
             blocks, skip_processed, debug, sc)
