from typing import List

from loguru import logger
from pathlib import Path

from satio.grid import get_blocks_gdf, buffer_bounds
from satio import layers
from satio.collections import (TerrascopeV200Collection,
                               AgERA5Collection,
                               DEMCollection)
from satio.utils.logs import (exitlogs, proclogs,
                              memlogs, ExitLogsMonitor)

from worldcereal.seasons import get_processing_dates, NoSeasonError
from worldcereal.utils.aez import get_matching_aez_id, group_from_id
from worldcereal.fp import (WorldCerealOpticalFeaturesProcessor,
                            WorldCerealSIGMA0FeaturesProcessor,
                            TSL2AFeaturesProcessor,
                            TSSIGMA0TiledFeaturesProcessor,
                            L8ThermalFeaturesProcessor)
from worldcereal.processors import (ClassificationProcessor,
                                    CropTypeProcessor)
from worldcereal.collections import (TerrascopeSigma0TiledCollection,
                                     WorldCerealOpticalTiledCollection,
                                     WorldCerealSigma0TiledCollection,
                                     WorldCerealThermalTiledCollection,
                                     L8ThermalTiledCollection)
from worldcereal.collections import (get_S2_products,
                                     create_products_df,
                                     _build_L8ThermalTiled_products_df)
from worldcereal.features.settings import (get_cropland_tsteps_parameters,
                                           get_croptype_tsteps_parameters,
                                           get_irr_parameters)
from worldcereal.postprocess import postprocess_tile

S2_GRID = layers.load('s2grid')


def get_processing_blocks(tiles, parameters, debug=False):
    '''Construct elementary processing blocks dataframe

    '''

    # Get the processing blocks
    blocks = get_blocks_gdf(tiles, s2grid=S2_GRID)

    # Add the parameters to each block
    # this is some onconventional Pandas stuff
    blocks['parameters'] = blocks.apply(lambda x: parameters, axis=1)

    if debug:
        # reduce to first block per tile
        # blocks = blocks[blocks.block_id == 0]
        blocks = blocks[blocks.block_id == 120]

        # # reduce blocks extent for testing
        # blocks["bounds"] = blocks["bounds"].apply(
        #     lambda x: buffer_bounds(x, - int((1024 - 128) / 2 * 10)))

    return blocks


def get_terrascope_collections(tile, start_date, end_date):

    def _get_S1_products(tile):
        basedir = Path('/data/worldcereal/data/s1_sigma0_tiled')
        currentdir = basedir / tile
        s1_products = list(currentdir.iterdir())
        s1_products = [str(x) for x in s1_products]

        return s1_products

    # ----------------------------
    # Look for S1 and S2 products
    # ----------------------------
    logger.info('Creating collections ...')
    epsg = S2_GRID[S2_GRID['tile'] == tile].epsg.values[0]
    s2products = get_S2_products(tile, start_date, end_date)
    s2_products_df = create_products_df(s2products, tile, epsg)
    s1products = _get_S1_products(tile)
    s1_products_df = create_products_df(s1products, tile, epsg,
                                        sensor='s1')
    l8folder = '/data/worldcereal/data/L8_LST_tiled'
    l8_products_df = _build_L8ThermalTiled_products_df(l8folder,
                                                       tiles=[tile])

    # ----------------------------
    # Create the collections
    # ----------------------------
    TSS2coll = TerrascopeV200Collection(s2_products_df)
    TSS1coll = TerrascopeSigma0TiledCollection(s1_products_df)
    L8Thermcoll = L8ThermalTiledCollection(l8_products_df)
    demcoll = DEMCollection(folder='/data/MEP/DEM/'
                            'COP-DEM_GLO-30_DTED/S2grid_20m')
    AgERA5coll = AgERA5Collection.from_path('/data/MTDA/AgERA5')

    collections = {
        'S2collection': TSS2coll,
        'S1collection': TSS1coll,
        'L8collection': L8Thermcoll,
        'AgERA5collection': AgERA5coll,
        'DEMcollection': demcoll
    }

    fps = {
        'L2A': TSL2AFeaturesProcessor,
        'SIGMA0': TSSIGMA0TiledFeaturesProcessor,
        'L8': L8ThermalFeaturesProcessor
    }

    return collections, fps


def get_worldcereal_collections():

    # ----------------------------
    # Create the collections
    # ----------------------------
    S2coll = WorldCerealOpticalTiledCollection.from_folder((
        '/data/worldcereal/runs/largescaletest/'
        'WORLDCEREAL_PREPROC/7tiles/OPTICAL'))
    S1coll = WorldCerealSigma0TiledCollection.from_folder(
        '/data/worldcereal/runs/largescaletest/'
        'WORLDCEREAL_PREPROC/7tiles/SAR')
    L8coll = WorldCerealThermalTiledCollection.from_folder(
        '/data/worldcereal/runs/largescaletest/'
        'WORLDCEREAL_PREPROC/7tiles/THERMAL')
    demcoll = DEMCollection(
        folder='/data/MEP/DEM/COP-DEM_GLO-30_DTED/S2grid_20m')
    AgERA5coll = AgERA5Collection.from_path('/data/MTDA/AgERA5')

    collections = {
        'S2collection': S2coll,
        'S1collection': S1coll,
        'L8collection': L8coll,
        'AgERA5collection': AgERA5coll,
        'DEMcollection': demcoll
    }

    fps = {
        'L2A': WorldCerealOpticalFeaturesProcessor,
        'SIGMA0': WorldCerealSIGMA0FeaturesProcessor,
        'L8': L8ThermalFeaturesProcessor
    }

    return collections, fps


def get_best_model(parentmodel, aez_id):
    '''Helper function to find locally trained models
    '''

    model_basedir = Path(parentmodel).parent
    aez_group = group_from_id(aez_id)

    groupmodel = (model_basedir / 'retrained' /
                  f'Group_{aez_group}' / 'config.json')
    zonemodel = (model_basedir / 'retrained' /
                 f'Zone_{aez_id}' / 'config.json')

    if zonemodel.is_file():
        logger.info(f'Found a zone-specific model: {zonemodel}')
        return str(zonemodel)
    elif groupmodel.is_file():
        logger.info(f'Found a group-specific model: {groupmodel}')
        return str(groupmodel)
    else:
        logger.info(f'Only found a parent model: {parentmodel}')
        return str(parentmodel)
    # return str(parentmodel)


def get_best_irr_model(parentmodel, aez_id):
    '''Helper function to find locally trained irr models
    '''
    modelname = Path(parentmodel).parent.stem
    model_basedir = Path(parentmodel).parents[2]
    aez_group = group_from_id(aez_id)

    groupmodel = (model_basedir / f'Group_{aez_group}' /
                  str(modelname) / 'config.json')
    zonemodel = (model_basedir / f'Zone_{aez_id}' /
                 str(modelname) / 'config.json')

    if zonemodel.is_file():
        logger.info(f'Found a zone-specific model: {zonemodel}')
        return str(zonemodel)
    elif groupmodel.is_file():
        logger.info(f'Found a group-specific model: {groupmodel}')
        return str(groupmodel)
    else:
        logger.info(f'Only found a parent model: {parentmodel}')
        return str(parentmodel)


def run_pipeline(output_folder, processing_tuple):
    '''Main WorldCereal products pipeline for 1 processing block

    '''
    block_id, block = processing_tuple

    # Get the processing parameters
    parameters = block.parameters
    geometry = block.geometry
    tile = block.tile
    name = parameters['name']
    year = parameters['year']
    season = parameters['season']
    models = parameters['models']
    featuresettings = parameters['featuresettings']
    filtersettings = parameters['filtersettings']
    save_confidence = parameters.get('save_confidence', True)
    save_meta = parameters.get('save_meta', True)
    save_features = parameters.get('save_features', False)
    active_marker = parameters.get('active_marker', None)
    cropland_mask = parameters.get('cropland_mask', None)
    irrigation = parameters.get('irrigation', False)
    irrparameters = parameters.get('irrparameters', None)
    irrmodels = parameters.get('irrmodels', None)

    logger.info('-'*50)
    logger.info(f'Starting processing block: {block_id}')
    logger.info('-'*50)
    logger.info('PARAMETERS:')
    for parameter, value in parameters.items():
        logger.info(f'{parameter}: {value}')

    # Get AEZ ID
    aez_id = get_matching_aez_id(geometry)

    # Get processing dates
    try:
        start_date, end_date = get_processing_dates(season, aez_id, year)
    except NoSeasonError:
        logger.warning(f'No valid `{season}` season found for this block.')
        return

    # Get collections
    collections, fps = get_terrascope_collections(tile, start_date, end_date)
    # collections, fps = get_worldcereal_collections()

    # Get best model
    for model in models.keys():
        models[model] = get_best_model(models[model], aez_id)

    if irrmodels is not None:
        for model in irrmodels.keys():
            irrmodels[model] = get_best_irr_model(irrmodels[model], aez_id)

    # Initialize processor
    if season == 'annual':
        processor = ClassificationProcessor
    else:
        processor = CropTypeProcessor

    chain = processor(
        output_folder,
        product=name,
        models=models,
        season=season,
        collections=dict(
            L2A=collections['S2collection'],
            SIGMA0=collections['S1collection'],
            AgERA5=collections['AgERA5collection'],
            DEM=collections['DEMcollection'],
        ),
        settings=featuresettings['settings'],
        features_meta=featuresettings['features_meta'],
        rsi_meta=featuresettings['rsi_meta'],
        ignore_def_feat=featuresettings['ignore_def_feat'],
        gdd_normalization=featuresettings['gddnormalization'],
        fps=fps,
        start_date=start_date.strip('-'),
        end_date=end_date.strip('-'),
        active_marker=active_marker,
        cropland_mask=cropland_mask,
        save_features=save_features,
        save_meta=save_meta,
        save_confidence=save_confidence,
        filtersettings=filtersettings,
        irrcollections=dict(
            L2A=collections['S2collection'],
            AgERA5=collections['AgERA5collection'],
            DEM=collections['DEMcollection'],
            L8=collections['L8collection']
        ),
        irrigation=irrigation,
        irrparameters=irrparameters,
        irrmodels=irrmodels)

    # Run pipeline
    chain.process(block.tile, block.bounds,
                  block.epsg, block.block_id)


def main(output_folder, processing_tuples, skip_processed,
         year, product, sc=None):

    # Setup logging folders
    output_folder = Path(output_folder)
    exitlogs_folder = output_folder / 'exitlogs' / f'{year}_{product}'
    proclogs_folder = output_folder / 'proclogs' / f'{year}_{product}'
    memlogs_folder = output_folder / 'memlogs' / f'{year}_{product}'

    @proclogs(proclogs_folder, level='DEBUG')
    @exitlogs(exitlogs_folder, skip_processed=skip_processed)
    @memlogs(memlogs_folder)
    def _log_run_pipeline(processing_tuple):
        run_pipeline(output_folder, processing_tuple)

    exitlogs_monitor = ExitLogsMonitor(exitlogs_folder,
                                       interval=300)
    exitlogs_monitor.start()

    if sc is not None:
        logger.info('Starting spark parallellization ...')
        sc.parallelize(processing_tuples,
                       len(processing_tuples)).foreach(_log_run_pipeline)

    else:
        logger.info('Testing locally ...')
        tup = processing_tuples[0]
        _log_run_pipeline(tup)

    exitlogs_monitor.task()
    exitlogs_monitor.stop()


def process_tiles(tiles: List[str], output_folder: str,
                  skip_processed: bool = True,
                  debug: bool = False, spark: bool = False):
    """
    Generates WorldCereal products.

    :param tiles: the list of MGRS tile id's to process,
                  for example: '["31UFS","50SMF"]'
    :param output_folder: an output folder
    :param skip_processed: True or False, setting True skips blocks
                           that were already successfully processed
    :param debug: True or False, setting to True runs the processing
                  on a smaller piece of data
    :param spark: True or False
    :returns:
    """

    # ANNUAL CROPLAND PARAMETERS
    parametersLC = dict(
        year=2020,
        season='annual',
        name='annualcropland',
        models={
            'WorldCerealPixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/cropland_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
        },
        featuresettings=get_cropland_tsteps_parameters(),
        filtersettings={
            'kernelsize': 5,
            'conf_threshold': 0.85
        },
        save_features=True
    )

    # WINTERWHEAT PARAMETERS
    parametersWW = dict(
        year=2020,
        season='winterwheat',
        name='winterwheat',
        models={
            'winterwheat_detector_PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/winterwheat_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'wintercereals_detector_PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/wintercereals_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
        },
        featuresettings=get_croptype_tsteps_parameters(),
        filtersettings={
            'kernelsize': 7,
            'conf_threshold': 1
        },
        active_marker=True,
        cropland_mask='WorldCerealPixelLSTM',
        irrigation=True,
        irrparameters=get_irr_parameters(),
        irrmodels={
            'irr_detector_RF': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr/Full_CIB/small_noTS_nosum/config.json'
        },
        save_features=True
    )

    # MAIZE1 PARAMETERS
    parametersM1 = dict(
        year=2020,
        season='maize1',
        name='maize1',
        models={
            'maize-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/maize_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'springwheat-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/springwheat_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'springcereals-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/springcereals_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
        },
        featuresettings=get_croptype_tsteps_parameters(),
        filtersettings={
            'kernelsize': 7,
            'conf_threshold': 1
        },
        active_marker=True,
        cropland_mask='WorldCerealPixelLSTM',
        irrigation=True,
        irrparameters=get_irr_parameters(),
        irrmodels={
            'irr_detector_RF': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr/Full_CIB/small_noTS_nosum/config.json'
        },
        save_features=True
    )

    # MAIZE2 PARAMETERS
    parametersM2 = dict(
        year=2020,
        season='maize2',
        name='maize2',
        models={
            'maize-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/maize_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'springwheat-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/springwheat_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'springcereals-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/springcereals_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
        },
        featuresettings=get_croptype_tsteps_parameters(),
        filtersettings={
            'kernelsize': 7,
            'conf_threshold': 1
        },
        active_marker=True,
        cropland_mask='WorldCerealPixelLSTM',
        irrigation=True,
        irrparameters=get_irr_parameters(),
        irrmodels={
            'irr_detector_RF': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr/Full_CIB/small_noTS_nosum/config.json'
        },
        save_features=True
    )

    # Initialize SC
    if spark:
        from worldcereal.utils.spark import get_spark_context
        sc = get_spark_context()
    else:
        sc = None

    params = [parametersLC,
              parametersM1, parametersWW,
              #   parametersM2
              ]
    for param in params:
        # Get processing blocks
        blocks = get_processing_blocks(tiles, param, debug)

        # # DEBUG SUBSET
        # blocks = blocks[(blocks['tile'] == '37MBV') & (blocks['block_id'] == 16)]

        # ----------------------------
        # Create processing tuples
        # ----------------------------
        logger.info('Getting processing tuples ...')
        processing_tuples = [(f'{row.tile}_{row.block_id:03d}', row)
                             for row in blocks.itertuples()]

        # ----------------------------
        # Launch processing
        # ----------------------------

        main(output_folder, processing_tuples, skip_processed,
             year=param['year'], product=param['name'], sc=sc)

        # ----------------------------
        # Do post-processing
        # ----------------------------

        logger.info('Start post-processing to COGs...')
        cog_folder = Path(output_folder) / 'cogs'

        for modelname in param['models'].keys():
            logger.info(f'Working on model: {modelname}')

            def _postprocess(tile):
                postprocess_tile(output_folder, cog_folder, param['name'],
                                 tile, param['year'], modelname,
                                 skip_processed=False, in_memory=False)

            if spark:
                sc.parallelize(tiles, len(tiles)).foreach(_postprocess)
            else:
                for tile in tiles:
                    _postprocess(tile)

    logger.success('All done!')


if __name__ == '__main__':

    debug = False
    spark = True
    skip_processed = True
    output_folder = ('/data/worldcereal/runs/'
                     'largescaletest/WORLDCEREAL_PRODUCTS_TS')

    #  ----------------
    #  TILES TO PROCESS
    #  ----------------
    tiles = ['30SWJ', '21HTC', '22LHQ', '37MBV', '35UQR', '50SMF', '31UFS']

    process_tiles(tiles, output_folder, skip_processed, debug, spark)
