from typing import List
import pandas as pd
from loguru import logger
from pathlib import Path

from satio.grid import get_blocks_gdf, buffer_bounds
from satio import layers
from satio.collections import (TerrascopeV200Collection,
                               AgERA5Collection,
                               DEMCollection)
from satio.utils.logs import (exitlogs, proclogs,
                              memlogs, ExitLogsMonitor)

from worldcereal.seasons import get_processing_dates, NoSeasonError
from worldcereal.utils.aez import get_matching_aez_id, group_from_id
from worldcereal.fp import (TSAgERA5FeaturesProcessor,
                            WorldCerealAgERA5FeaturesProcessor,
                            WorldCerealOpticalFeaturesProcessor,
                            WorldCerealSIGMA0FeaturesProcessor,
                            TSL2AFeaturesProcessor,
                            TSSIGMA0TiledFeaturesProcessor,
                            L8ThermalFeaturesProcessor,
                            WorldCerealThermalFeaturesProcessor)
from worldcereal.processors import (ClassificationProcessor,
                                    CropTypeProcessor)
from worldcereal.collections import (TerrascopeSigma0TiledCollection,
                                     WorldCerealOpticalTiledCollection,
                                     WorldCerealSigma0TiledCollection,
                                     WorldCerealThermalTiledCollection,
                                     L8ThermalTiledCollection)
from worldcereal.collections import (get_S2_products,
                                     create_products_df,
                                     _build_L8ThermalTiled_products_df)
from worldcereal.features.settings import (get_cropland_tsteps_parameters,
                                           get_croptype_tsteps_parameters,
                                           get_irr_parameters)
from worldcereal.postprocess import postprocess_tile

S2_GRID = layers.load('s2grid')


def get_processing_blocks(countries, parameters, debug=False,
                          spark=False):
    '''Construct elementary processing blocks dataframe

    '''
    list_blocks = []

    for country, tiledict in countries.items():

        # Get the processing blocks
        tiles = tiledict.keys()
        blocks = get_blocks_gdf(tiles, s2grid=S2_GRID)

        # Add the parameters to each block
        # this is some onconventional Pandas stuff
        blocks['parameters'] = blocks.apply(lambda x: parameters, axis=1)
        blocks['country'] = [country] * len(blocks)

        if debug:
            # reduce to requested blocks per tile
            for tile, req_blocks in tiledict.items():
                list_blocks.append(
                    blocks[(blocks.tile == tile) &
                           (blocks.block_id.isin(req_blocks))])
        else:
            list_blocks.append(blocks)

    blocks = pd.concat(list_blocks, axis=0)

    if debug and (not spark):
        # reduce blocks extent for testing
        blocks["bounds"] = blocks["bounds"].apply(
            lambda x: buffer_bounds(x, - int((1024 - 256) / 2 * 10)))

    return blocks


def get_terrascope_collections(tile, start_date, end_date):

    def _get_S1_products(tile):
        basedir = Path('/data/worldcereal/data/s1_sigma0_tiled')
        currentdir = basedir / tile
        s1_products = list(currentdir.iterdir())
        s1_products = [str(x) for x in s1_products]

        return s1_products

    # ----------------------------
    # Look for S1 and S2 products
    # ----------------------------
    logger.info('Creating collections ...')
    epsg = S2_GRID[S2_GRID['tile'] == tile].epsg.values[0]
    s2products = get_S2_products(tile, start_date, end_date)
    s2_products_df = create_products_df(s2products, tile, epsg)
    s1products = _get_S1_products(tile)
    s1_products_df = create_products_df(s1products, tile, epsg,
                                        sensor='s1')
    l8folder = '/data/worldcereal/data/L8_LST_tiled'
    l8_products_df = _build_L8ThermalTiled_products_df(l8folder,
                                                       tiles=[tile])

    # ----------------------------
    # Create the collections
    # ----------------------------
    TSS2coll = TerrascopeV200Collection(s2_products_df)
    TSS1coll = TerrascopeSigma0TiledCollection(s1_products_df)
    L8Thermcoll = L8ThermalTiledCollection(l8_products_df)
    demcoll = DEMCollection(folder='/data/MEP/DEM/'
                            'COP-DEM_GLO-30_DTED/S2grid_20m')
    AgERA5coll = AgERA5Collection.from_path('/data/MTDA/AgERA5')

    collections = {
        'S2collection': TSS2coll,
        'S1collection': TSS1coll,
        'L8collection': L8Thermcoll,
        'AgERA5collection': AgERA5coll,
        'DEMcollection': demcoll
    }

    fps = {
        'L2A': TSL2AFeaturesProcessor,
        'SIGMA0': TSSIGMA0TiledFeaturesProcessor,
        'AgERA5': TSAgERA5FeaturesProcessor,
        'L8': L8ThermalFeaturesProcessor
    }

    return collections, fps


def get_worldcereal_collections(country):

    # ----------------------------
    # Create the collections
    # ----------------------------
    logger.info('Initializing collections ...')
    S2coll = WorldCerealOpticalTiledCollection.from_folder((
        f'/data/worldcereal/runs/largescaletest/'
        f'WORLDCEREAL_PREPROC/{country}/OPTICAL'))
    S1coll = WorldCerealSigma0TiledCollection.from_folder(
        f'/data/worldcereal/runs/largescaletest/'
        f'WORLDCEREAL_PREPROC/{country}/SAR')
    L8coll = WorldCerealThermalTiledCollection.from_folder(
        f'/data/worldcereal/runs/largescaletest/'
        f'WORLDCEREAL_PREPROC/{country}/TIR')
    demcoll = DEMCollection(
        folder='/data/MEP/DEM/COP-DEM_GLO-30_DTED/S2grid_20m')
    AgERA5coll = AgERA5Collection.from_path('/data/MTDA/AgERA5')

    collections = {
        'S2collection': S2coll,
        'S1collection': S1coll,
        'L8collection': L8coll,
        'AgERA5collection': AgERA5coll,
        'DEMcollection': demcoll
    }

    fps = {
        'L2A': WorldCerealOpticalFeaturesProcessor,
        'SIGMA0': WorldCerealSIGMA0FeaturesProcessor,
        'AgERA5': WorldCerealAgERA5FeaturesProcessor,
        'L8': WorldCerealThermalFeaturesProcessor
    }

    return collections, fps


def get_best_model(parentmodel, aez_id):
    '''Helper function to find locally trained models
    '''

    model_basedir = Path(parentmodel).parent
    aez_group = group_from_id(aez_id)

    groupmodel = (model_basedir / 'retrained' /
                  f'Group_{aez_group}' / 'config.json')
    zonemodel = (model_basedir / 'retrained' /
                 f'Zone_{aez_id}' / 'config.json')

    if zonemodel.is_file():
        logger.info(f'Found a zone-specific model: {zonemodel}')
        return str(zonemodel)
    elif groupmodel.is_file():
        logger.info(f'Found a group-specific model: {groupmodel}')
        return str(groupmodel)
    else:
        logger.info(f'Only found a parent model: {parentmodel}')
        return str(parentmodel)
    # return str(parentmodel)


def get_best_irr_model(parentmodel, aez_id):
    '''Helper function to find locally trained irr models
    '''
    modelname = Path(parentmodel).parent.stem
    model_basedir = Path(parentmodel).parents[2]
    aez_group = group_from_id(aez_id)

    groupmodel = (model_basedir / f'Group_{aez_group}' /
                  str(modelname) / 'config.json')
    zonemodel = (model_basedir / f'Zone_{aez_id}' /
                 str(modelname) / 'config.json')

    if zonemodel.is_file():
        logger.info(f'Found a zone-specific model: {zonemodel}')
        return str(zonemodel)
    elif groupmodel.is_file():
        logger.info(f'Found a group-specific model: {groupmodel}')
        return str(groupmodel)
    else:
        logger.info(f'Only found a parent model: {parentmodel}')
        return str(parentmodel)


def run_pipeline(output_folder, processing_tuple):
    '''Main WorldCereal products pipeline for 1 processing block

    '''
    block_id, block = processing_tuple

    # Get the processing parameters
    country = block.country
    parameters = block.parameters
    geometry = block.geometry
    tile = block.tile
    name = parameters['name']
    year = parameters['year']
    season = parameters['season']
    models = parameters['models']
    featuresettings = parameters['featuresettings']
    filtersettings = parameters['filtersettings']
    save_confidence = parameters.get('save_confidence', True)
    save_meta = parameters.get('save_meta', True)
    save_features = parameters.get('save_features', False)
    active_marker = parameters.get('active_marker', None)
    cropland_mask = parameters.get('cropland_mask', None)
    irrigation = parameters.get('irrigation', False)
    irrparameters = parameters.get('irrparameters', None)
    irrmodels = parameters.get('irrmodels', None)

    logger.info('-'*50)
    logger.info(f'Starting processing block: {block_id}')
    logger.info('-'*50)
    logger.info('PARAMETERS:')
    for parameter, value in parameters.items():
        logger.info(f'{parameter}: {value}')

    # Get AEZ ID
    aez_id = get_matching_aez_id(geometry)

    # Get processing dates
    try:
        start_date, end_date = get_processing_dates(season, aez_id, year)
    except NoSeasonError:
        logger.warning(f'No valid `{season}` season found for this block.')
        return

    # Get collections
    # collections, fps = get_terrascope_collections(tile, start_date, end_date)
    collections, fps = get_worldcereal_collections(country)

    # Get best model
    for model in models.keys():
        models[model] = get_best_model(models[model], aez_id)

    if irrmodels is not None:
        for model in irrmodels.keys():
            irrmodels[model] = get_best_irr_model(irrmodels[model], aez_id)

    # Initialize processor
    if season == 'annual':
        processor = ClassificationProcessor
    else:
        processor = CropTypeProcessor

    output_folder = Path(output_folder) / country

    chain = processor(
        output_folder,
        product=name,
        models=models,
        season=season,
        collections=dict(
            L2A=collections['S2collection'],
            SIGMA0=collections['S1collection'],
            AgERA5=collections['AgERA5collection'],
            DEM=collections['DEMcollection'],
        ),
        settings=featuresettings['settings'],
        features_meta=featuresettings['features_meta'],
        rsi_meta=featuresettings['rsi_meta'],
        ignore_def_feat=featuresettings['ignore_def_feat'],
        gdd_normalization=featuresettings['gddnormalization'],
        fps=fps,
        start_date=start_date.strip('-'),
        end_date=end_date.strip('-'),
        active_marker=active_marker,
        cropland_mask=cropland_mask,
        save_features=save_features,
        save_meta=save_meta,
        save_confidence=save_confidence,
        filtersettings=filtersettings,
        irrcollections=dict(
            L2A=collections['S2collection'],
            AgERA5=collections['AgERA5collection'],
            DEM=collections['DEMcollection'],
            L8=collections['L8collection']
        ),
        irrigation=irrigation,
        irrparameters=irrparameters,
        irrmodels=irrmodels)

    # Run pipeline
    chain.process(block.tile, block.bounds,
                  block.epsg, block.block_id)


def main(output_folder, processing_tuples, skip_processed,
         year, product, sc=None):

    # Setup logging folders
    output_folder = Path(output_folder)
    exitlogs_folder = output_folder / 'exitlogs' / f'{year}_{product}'
    proclogs_folder = output_folder / 'proclogs' / f'{year}_{product}'
    memlogs_folder = output_folder / 'memlogs' / f'{year}_{product}'

    @proclogs(proclogs_folder, level='DEBUG')
    @exitlogs(exitlogs_folder, skip_processed=skip_processed)
    @memlogs(memlogs_folder)
    def _log_run_pipeline(processing_tuple):
        run_pipeline(output_folder, processing_tuple)

    exitlogs_monitor = ExitLogsMonitor(exitlogs_folder,
                                       interval=300)
    exitlogs_monitor.start()

    if sc is not None:
        logger.info('Starting spark parallellization ...')
        sc.parallelize(processing_tuples,
                       len(processing_tuples)).foreach(_log_run_pipeline)

    else:
        logger.info('Testing locally ...')
        tup = processing_tuples[0]
        _log_run_pipeline(tup)

    exitlogs_monitor.task()
    exitlogs_monitor.stop()


def process_tiles(countries: dict, output_folder: str,
                  skip_processed: bool = True,
                  debug: bool = False, spark: bool = False):
    """
    Generates WorldCereal products.

    :param tiles: the list of MGRS tile id's to process,
                  for example: '["31UFS","50SMF"]'
    :param output_folder: an output folder
    :param skip_processed: True or False, setting True skips blocks
                           that were already successfully processed
    :param debug: True or False, setting to True runs the processing
                  on a smaller piece of data
    :param spark: True or False
    :returns:
    """

    # ANNUAL CROPLAND PARAMETERS
    parametersLC = dict(
        year=2019,
        season='annual',
        name='annualcropland',
        models={
            'WorldCerealPixelLSTM_8': '/data/worldcereal/models/WorldCerealPixelLSTM/cropland_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'WorldCerealPixelLSTM_10': '/data/worldcereal/models/WorldCerealPixelLSTM/cropland_detector_WorldCerealPixelLSTM_10Bands/parentmodel/config.json'  # NOQA
        },
        featuresettings=get_cropland_tsteps_parameters(),
        filtersettings={
            'kernelsize': 5,
            'conf_threshold': 0.85
        },
        save_features=True
    )

    # WINTERWHEAT PARAMETERS
    parametersWW = dict(
        year=2019,
        season='winterwheat',
        name='winterwheat',
        models={
            'winterwheat_detector_PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/winterwheat_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'wintercereals_detector_PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/wintercereals_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
        },
        featuresettings=get_croptype_tsteps_parameters(),
        filtersettings={
            'kernelsize': 7,
            'conf_threshold': 1
        },
        active_marker=True,
        cropland_mask='WorldCerealPixelLSTM',
        irrigation=True,
        irrparameters=get_irr_parameters(),
        irrmodels={
            'irr_RF_smallnoTS': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr/Full_CIB/small_noTS/config.json',  # NOQA
            'irr_RF_smallnoTS_noWW': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr_noWW/Full_CIB/small_noTS/config.json'  # NOQA
        },
        save_features=True
    )

    # MAIZE1 PARAMETERS
    parametersM1 = dict(
        year=2019,
        season='maize1',
        name='maize1',
        models={
            'maize-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/maize_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'springwheat-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/springwheat_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'springcereals-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/springcereals_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
        },
        featuresettings=get_croptype_tsteps_parameters(),
        filtersettings={
            'kernelsize': 7,
            'conf_threshold': 1
        },
        active_marker=True,
        cropland_mask='WorldCerealPixelLSTM',
        irrigation=True,
        irrparameters=get_irr_parameters(),
        irrmodels={
            'irr_RF_smallnoTS': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr/Full_CIB/small_noTS/config.json',  # NOQA
            'irr_RF_smallnoTS_noWW': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr_noWW/Full_CIB/small_noTS/config.json'  # NOQA
        },
        save_features=True
    )

    # MAIZE2 PARAMETERS
    parametersM2 = dict(
        year=2020,
        season='maize2',
        name='maize2',
        models={
            'maize-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/maize_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'springwheat-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/springwheat_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
            'springcereals-PixelLSTM': '/data/worldcereal/models/WorldCerealPixelLSTM/springcereals_detector_WorldCerealPixelLSTM_8Bands/parentmodel/config.json',  # NOQA
        },
        featuresettings=get_croptype_tsteps_parameters(),
        filtersettings={
            'kernelsize': 7,
            'conf_threshold': 1
        },
        active_marker=True,
        cropland_mask='WorldCerealPixelLSTM',
        irrigation=True,
        irrparameters=get_irr_parameters(),
        irrmodels={
            'irr_RF_smallnoTS': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr/Full_CIB/small_noTS/config.json',  # NOQA
            'irr_RF_smallnoTS_noWW': '/data/worldcereal/benchmark_irr/CIB_V1/RF/irr_noWW/Full_CIB/small_noTS/config.json'  # NOQA
        },
        save_features=True
    )

    # Initialize SC
    if spark:
        from worldcereal.utils.spark import get_spark_context
        sc = get_spark_context(pythonpath='/data/users/Public/jeroendegerickx/'
                               'conda_env/worldcereal/bin/python')
    else:
        sc = None

    params = [
        parametersLC,
        # parametersM1, parametersWW,
        # parametersM2
    ]
    for param in params:
        # Get processing blocks
        blocks = get_processing_blocks(countries, param, debug,
                                       spark)

        # # DEBUG SUBSET
        # blocks = blocks[(blocks['tile'] == '37MBV') & (blocks['block_id'] == 16)]

        # ----------------------------
        # Create processing tuples
        # ----------------------------
        logger.info('Getting processing tuples ...')
        processing_tuples = [(f'{row.tile}_{row.block_id:03d}', row)
                             for row in blocks.itertuples()]

        # ----------------------------
        # Launch processing
        # ----------------------------

        main(output_folder, processing_tuples, skip_processed,
             year=param['year'], product=param['name'], sc=sc)

        # ----------------------------
        # Do post-processing
        # ----------------------------
        logger.info('Start post-processing to COGs...')
        for country, tiledict in countries.items():
            tiles = tiledict.keys()

            outfolder = Path(output_folder) / country
            cog_folder = Path(outfolder) / 'cogs'

            for modelname in param['models'].keys():
                logger.info(f'Working on model: {modelname}')

                def _postprocess(tile):
                    postprocess_tile(outfolder, cog_folder, param['name'],
                                     tile, param['year'], modelname,
                                     skip_processed=False, in_memory=False)
                if spark:
                    sc.parallelize(tiles, len(tiles)).foreach(_postprocess)
                else:
                    for tile in tiles:
                        _postprocess(tile)

            # post-process irrigation products
            if param.get('irrigation', False):
                for modelname in param['irrmodels'].keys():
                    logger.info(f'Working on model: {modelname}')

                    def _postprocess(tile):
                        postprocess_tile(outfolder, cog_folder,
                                         'irr-' + param['name'],
                                         tile, param['year'], modelname,
                                         skip_processed=False,
                                         in_memory=False)
                    if spark:
                        sc.parallelize(tiles, len(tiles)).foreach(_postprocess)
                    else:
                        for tile in tiles:
                            _postprocess(tile)

    logger.success('All done!')


if __name__ == '__main__':

    # if debug == True, processor will only process a selection
    # of blocks per tile
    debug = True
    # if spark == False, processor will reduce the extent of the blocks
    # in order to run the processing locally
    spark = False
    skip_processed = True
    output_folder = ('/data/worldcereal/runs/'
                     'largescaletest/WORLDCEREAL_PRODUCTS_10bands')

    #  ----------------
    #  TILES TO PROCESS
    #  ----------------

    # use year = 2020 for the following tiles
    # (adjust in parameters)!
    # countries = {
    #     '7tiles': {'tiles': ['30SWJ',
    #                          '21HTC', '22LHQ', '37MBV', '35UQR', '50SMF', '31UFS'
    #                          ],
    #                'blocks': [105]}}

    # use year = 2019 for the following tiles
    # (adjust in parameters)!
    countries = {
        # 'FRANCE': {'tiles': [''],
        #            'blocks': ['']},
        'SPAIN': {'30SVG': [86, 102],
                  '30STG': [60],
                  '30SVF': [38],
                  '30SUF': [38],
                  '30SWH': [22],
                  '30SUH': [34]
                  },
        # 'UKRAINE': {'tiles': [''],
        #             'blocks': []},
        'ARGENTINA': {'20GMA': [29],
                      '20HNB': [57, 66],
                      '21HTU': [6],
                      '20HMF': [93, 94],
                      '21HTC': [64, 75]},

        # 'TANZANIA': {'tiles': [''],
        #              'blocks': []}
    }

    # spain
    # tiles = ['29SPC', '29SQC', '30STH', '30SUH',
    #          '30SVH', '30SWH', '30SXH', '29SPB',
    #          '30SUG', '30SVG', '30SWG', '30SXG',
    #          '29SQB', '30STG', '29SQA', '30STF',
    #          '30SUF', '30SVF', '30SWF']

    # # france
    # tiles = ['31TBF', '31TCF', '30TWS',
    #          '30TWT', '30TXN', '30TXP',
    #          '30TXQ', '30TXR', '30TXS',
    #          '30TXT', '30TYN', '30TYP',
    #          '30TYQ', '30TYR', '30TYS',
    #          '30TYT', '30UXU', '30UYU',
    #          '31TBG', '31TCG', '31TCH',
    #          '31TCJ', '31TCK', '31TCL',
    #          '31TCM', '31TCN', '31TDF',
    #          '31TDG', '31TDH', '31TDJ',
    #          '31TDK', '31TDL', '31TDM',
    #          '31TDN', '31UCP', '31UDP']

    # # # Ukraine
    # tiles = ['36UUB', '36UUC', '35UPS',
    #          '35UPT', '35UQS', '35UQT',
    #          '36UUA', '36UVA', '35UQR',
    #          '35UPR', '36TUS', '36TUT',
    #          '36UUU', '36TVS', '36UUV',
    #          '36TVT', '36UVU', '36TWS',
    #          '36UVV', '36TWT', '35UPQ',
    #          '36UWU', '36TXS', '36UWV',
    #          '36TXT', '35UQP', '35UQQ']

    # # Tanzania
    # tiles = ['36MTB', '36MTC', '36MTD',
    #          '36MUA', '36MUB', '36MUC',
    #          '36MUD', '36MVA', '36MVB',
    #          '36MVC', '36MVD', '36MWA',
    #  '36MWB', '36MWC', '36MWD',
    #  '36MXB', '36MXC', '36MXD',
    #  '36MYC', '36MYD', '36MXA',
    #  '36MYA', '36MYB', '36MZA',
    #  '36MZB', '36MZC', '37MBR',
    #  '37MBS', '37MBT', '37MCR',
    #  '37MCS'
    #  ]

    # ARGENTINA
    # tiles = [
    #     # '20HMB', '20HMC', '20HMD',
    #     #          '21HWV', '20HME', '20HMF',
    #     #          '20HMG', '20HNB', '20HNC',
    #     #          '20HND', '20HNE', '20HNF',
    #     #          '20HNG', '20HNH', '20HPB',
    #     #          '20HPC', '20HPD', '20HPE',
    #     #          '20HPF', '20HPG', '20GMA',
    #     '20HPH', '20HQC', '20HQD',
    #     '20HQE', '20HQF', '20HQG',
    #     '20GNA', '20HQH', '21HTA',
    #     '21HTB', '21HTC', '20GNV',
    #     '21HTT', '21HTU', '21HTV',
    #     '21HUA', '21HUB', '21HUC',
    #     '21HUT', '21HUU', '21HUV',
    #     '21HVA', '21HVB', '21HVT',
    #     '21HVU', '21HVV'
    # ]

    process_tiles(countries, output_folder, skip_processed, debug, spark)
